#!/bin/bash -e

if [ -z "$CC" ] ; then
	CC=gcc
fi
if [ -z "$CXX" ] ; then
	CXX=g++
fi

usage() {
	echo
	echo "Usage: BuildVersions.sh /path/to/install/dir shellp_version grale_version graleshell_version gamods_version graleeditor_version"
	echo
	echo "   or"
	echo
	echo "Usage: BuildVersions.sh /path/to/install/dir latest"
	echo
	exit -1
}

testCompile() {
	F=`mktemp`
	echo -e "$@" > "${F}.c"
	if $CC -c -o ${F}.o ${F}.c ; then
		echo "yes"
	else
		echo "no"
	fi
	rm -f ${F} ${F}.c ${F}.o
}

test_cfitsio() {
	echo "Testing if CFITSIO can already be found" >&2
	echo `testCompile "#include <fitsio.h>\nint main(void) { return 0; }"`
}

test_gsl() {
	echo "Testing if GSL can already be found" >&2
	echo `testCompile "#include <gsl/gsl_rng.h>\nint main(void) { return 0; }"`
}

yesNoAnswer() {
	while true ; do
		read X
		if [ "$X" = "y" ] || [ "$X" = "n" ] ; then
			echo $X
			break
		fi
		echo "Please enter 'y' or 'n'" >&2
	done
}

getVersion () {
	ID="$1"
	HTML="$2"

	echo "$HTML" | grep "$ID" | grep "tar.gz" | sed -e "s/.*$ID//" | sed -e "s/.tar.gz.*//"
}

getLatestVersions() {
	echo "Downloading HTML pages to extract latest versions" >&2
	PAGEINFO=`curl -o - http://research.edm.uhasselt.be/jori/page/Physics/Grale.html 2>/dev/null ; curl -o - http://research.edm.uhasselt.be/jori/page/Misc/Shellp.html 2>/dev/null`
	echo `getVersion shellp- "$PAGEINFO"` \
	    `getVersion grale- "$PAGEINFO"` \
	    `getVersion graleshell- "$PAGEINFO"` \
	    `getVersion grale_gamodules- "$PAGEINFO"` \
	    `getVersion graleeditor- "$PAGEINFO"` 
}

CURL=`which curl 2>/dev/null | cat`
if [ -z "$CURL" ] ; then
	echo "Need curl to be able to download. Aborting" >&2
	exit -1
fi

if [ "$#" = 6 ] ; then
	V_SHELLP="$2"
	V_GRALE="$3"
	V_SHELL="$4"
	V_MODS="$5"
	V_EDITOR="$6"
else
	if [ "$#" = "2" ] && [ "$2" = "latest" ] ; then
		VERSIONS=`getLatestVersions`
		V_SHELLP=`echo $VERSIONS | cut -f 1 -d " "`
		V_GRALE=`echo $VERSIONS | cut -f 2 -d " "`
		V_SHELL=`echo $VERSIONS | cut -f 3 -d " "`
		V_MODS=`echo $VERSIONS | cut -f 4 -d " "`
		V_EDITOR=`echo $VERSIONS | cut -f 5 -d " "`
		echo "Detected versions: "
		echo "  shellp: $V_SHELLP"
		echo "  grale: $V_GRALE"
		echo "  graleshell: $V_SHELL"
		echo "  grale_gamodules: $V_MODS"
		echo "  graleeditor: $V_EDITOR"
	else
		usage
	fi
fi

INSTDIR="$1"

STARTDIR=`pwd`
if ! [ -e "$INSTDIR" ] || ! [ -d "$INSTDIR" ] ; then
	echo "ERROR: Specified installation directory $INSTDIR does not exist, or is not a directory"
	exit -1
fi
cd $INSTDIR
INSTDIR=`pwd` # Get full path
echo "Installation base: $INSTDIR"

SKIPGRALEEDITOR="no"
QT5=`qmake -version 2>&1 | grep "Qt version 5" |cat`
if [ -z "$QT5" ] ; then
	echo "WARNING: qmake for Qt version 5 is not found"
	echo "Continue anyway? (y/n)"
	if [ `yesNoAnswer` != "y" ] ; then
		echo "Aborting"
		exit -1
	fi

	echo "Skip graleeditor? (y/n)"
	if [ `yesNoAnswer` = "y" ] ; then
		SKIPGRALEEDITOR="yes"
	fi
else
	echo "Qt5 detected"
fi

MPICC=`which mpicc 2>/dev/null |cat`
if [ -z "$MPICC" ] ; then
	echo "WARNING: couldn't detect MPI"
	echo "Continue anyway? (y/n)"
	if [ `yesNoAnswer` != "y" ] ; then
		echo "Aborting"
		exit -1
	fi
else
	echo "MPI detected"
fi

INSTALLFILES=""
if [ `test_cfitsio` != "yes" ] ; then
	INSTALLFILES="$INSTALLFILES cfitsio_latest.tar.gz:cfitsio,http://heasarc.gsfc.nasa.gov/FTP/software/fitsio/c/"
	echo "Will download and compile CFITSIO"
else
	echo "CFITSIO already installed"
fi

if [ `test_gsl` != "yes" ] ; then
	INSTALLFILES="$INSTALLFILES gsl-1.16.tar.gz,http://gnu.xl-mirror.nl/gsl/"
	echo "Will download and compile GSL"
else
	echo "GSL already installed"
fi

INSTALLFILES="$INSTALLFILES jthread-1.3.1.tar.gz,http://research.edm.uhasselt.be/jori/jthread/"
INSTALLFILES="$INSTALLFILES errut-1.0.1.tar.gz,http://research.edm.uhasselt.be/jori/errut/"
INSTALLFILES="$INSTALLFILES enut-1.1.2.tar.gz,http://research.edm.uhasselt.be/enut/"
INSTALLFILES="$INSTALLFILES serut-1.0.1.tar.gz,http://research.edm.uhasselt.be/jori/serut/"
INSTALLFILES="$INSTALLFILES mogal-0.11.0.tar.gz,http://research.edm.uhasselt.be/jori/mogal/"
INSTALLFILES="$INSTALLFILES shellp-${V_SHELLP}.tar.gz,http://research.edm.uhasselt.be/jori/shellp/"
INSTALLFILES="$INSTALLFILES grale-${V_GRALE}.tar.gz,http://research.edm.uhasselt.be/jori/grale/"
INSTALLFILES="$INSTALLFILES graleshell-${V_SHELL}.tar.gz,http://research.edm.uhasselt.be/jori/grale/"
INSTALLFILES="$INSTALLFILES grale_gamodules-${V_MODS}.tar.gz,http://research.edm.uhasselt.be/jori/grale/"

if [ "$SKIPGRALEEDITOR" = "no" ] ; then
	INSTALLFILES="$INSTALLFILES graleeditor-${V_EDITOR}.tar.gz,http://research.edm.uhasselt.be/jori/grale/"
fi

B=`pwd`/build
if ! [ -e "$B" ] ; then
	mkdir $B
fi

P=`pwd`/gralepackage
rm -rf $P

for l in $INSTALLFILES ; do

	cd $B

	N=`echo $l|cut -f 1 -d ,|cut -f 1 -d ":"`
	U=`echo $l|cut -f 2 -d ,`
	OPTS=`echo $l|cut -f 3 -d ,`
	echo "Processing $N"
	if ! [ -e $N ] ; then
		echo "Downloading"
		curl -o $N ${U}${N}
	fi

	X=`echo $l|cut -f 1 -d ,|grep ":"|cat`
	if [ -z "$X" ] ; then
		D=${N:0:${#N}-7}
	else
		D=`echo $l|cut -f 1 -d ,| cut -f 2 -d ":"`
	fi
	echo "Extracting"
	if ! [ -e $D ] ; then
		tar xfz $N
	fi

	if [ -e $D/configure ] ; then
		echo "Running configure"
		cd $D
		if ! [ -e Makefile ] ; then
			CXXFLAGS=-O2 CFLAGS=-O2 ./configure --prefix="$P" $OPTS
		fi
		make -j 4
		make install
	else
		echo "Running cmake"
		if ! [ -e $D/build ] ; then
			mkdir $D/build
		fi

		cd $D/build
		CPPFLAGS="-I$P/include" cmake .. -DCMAKE_BUILD_TYPE=release -DCMAKE_INSTALL_PREFIX="$P" \
			-DJThread_DIR="$P/lib/cmake/Shellp" \
			-DErrUt_DIR="$P/lib/cmake/Shellp" \
			-DSerUt_DIR="$P/lib/cmake/Shellp" \
			-DENUt_DIR="$P/lib/cmake/Shellp" \
			-DMOGAL_DIR="$P/lib/cmake/Shellp" \
			-DGRALE_DIR="$P/lib/cmake/GRALE" \
			-DShellp_DIR="$P/lib/cmake/Shellp" \
			-DADDITIONAL_LIBRARIES=$ADDITIONAL_LIBRARIES
		make -j 4

		echo "${D:0:15}"
		if [ "${D:0:15}" = "grale_gamodules" ] ; then
			cp -v libgalens* "$P/lib/"
		else
			make install

			# It seems that cmake's rpath code doesn't allow Qt5 to be found after
			# a make install. As a quick fix, we'll just copy the build executables
			COPYEXES="graleshell graleshellmpi graleeditor"
			for i in $COPYEXES ; do
				if [ -e "$i" ] ; then
					cp -v $i "$P/bin"
				fi
			done
		fi
	fi
done

cd $P
cat > bin/activate.sh << EOF
if [ -z "$LD_LIBRARY_PATH" ] ; then
	export LD_LIBRARY_PATH="$P/lib"
else
	export LD_LIBRARY_PATH="$P/lib:$LD_LIBRARY_PATH"
fi
export GRALE_MODULEPATH="$P/lib"
export PATH="$P/bin:$PATH"
EOF
chmod u+rx bin/activate.sh

cat << EOF

Build finished!

Run

    source $P/bin/activate.sh

to use this version of grale

EOF
