/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "cmdgrid.h"
#include "graleshellinstance.h"
#include <shellp/iosystem.h>
#include <grale/grid.h>
#include <grale/real2dlensfunction.h>
#include <grale/gravitationallens.h>
#include <grale/compositelens.h>
#include <grale/gridfunction.h>
#include <stdio.h>

#include "debugnew.h"

using namespace shellp;
using namespace std;

CmdGridNewUniform::CmdGridNewUniform(const std::string &cmdname) : ShellCommand(cmdname)
{
	centerxarg = new ShellCmdRealArg("x");
	centerxarg->setDescription("X coordinate of the center of the square-shaped grid.");
	centerxarg->setUnitType(UNIT_ANGULAR);
	addArgument(centerxarg);
	centeryarg = new ShellCmdRealArg("y");
	centeryarg->setDescription("Y coordinate of the center of the square-shaped grid.");
	centeryarg->setUnitType(UNIT_ANGULAR);
	addArgument(centeryarg);
	sizearg = new ShellCmdRealArg("size");
	sizearg->setDescription("Length of a side of the square-shaped grid.");
	sizearg->setUnitType(UNIT_ANGULAR);
	addArgument(sizearg);
	subdivarg = new ShellCmdIntArg("subdiv");
	subdivarg->setDefault("10");
	subdivarg->setDescription("Specifies the number of grid squares on a side.");
	addArgument(subdivarg);
	m_pGridName = new ShellCmdStringArg("gridname");
	m_pGridName->setDefault("*");
	m_pGridName->setDescription("Specify a name for the grid, is useful when combining multiple grids");
	addArgument(m_pGridName);

	setDescription("Creates a uniform grid which can be used in a lens inversion procedure.");
}

CmdGridNewUniform::~CmdGridNewUniform()
{
	delete centerxarg;
	delete centeryarg;
	delete sizearg;
	delete subdivarg;
	delete m_pGridName;
}

bool CmdGridNewUniform::execute()
{
	grale::Grid *g = new grale::Grid();

	if (!g->buildUniform(grale::Vector2D<double>(centerxarg->getValue(),centeryarg->getValue()),sizearg->getValue(),subdivarg->getValue()))
	{
		setErrorString(std::string("Couldn't build uniform grid: ") + g->getErrorString());
		delete g;
		return false;
	}
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	
	string name = m_pGridName->getValue();
	if (name.length() == 0)
	{
		setErrorString("No grid name specified");
		return false;
	}

	if (name == "*") // the default
		name = "";

	gsinst->SetGrid(g, name);
	return true;
}

CmdGridNewLens::CmdGridNewLens(const std::string &cmdname) : ShellCommand(cmdname)
{
	centerxarg = new ShellCmdRealArg("x");
	centerxarg->setDescription("X coordinate of the center of the square-shaped grid.");
	centerxarg->setUnitType(UNIT_ANGULAR);
	addArgument(centerxarg);
	centeryarg = new ShellCmdRealArg("y");
	centeryarg->setDescription("Y coordinate of the center of the square-shaped grid.");
	centeryarg->setUnitType(UNIT_ANGULAR);
	addArgument(centeryarg);
	sizearg = new ShellCmdRealArg("size");
	sizearg->setDescription("Length of a side of the square-shaped grid.");
	sizearg->setUnitType(UNIT_ANGULAR);
	addArgument(sizearg);
	minsquaresarg = new ShellCmdIntArg("minsquares");
	minsquaresarg->setDescription("Minimun number of squares in the resulting grid.");
	minsquaresarg->setMin(10);
	addArgument(minsquaresarg);
	maxsquaresarg = new ShellCmdIntArg("maxsquares");
	maxsquaresarg->setDescription("Maximun number of squares in the resulting grid.");
	maxsquaresarg->setMin(10);
	addArgument(maxsquaresarg);

	xintarg = new ShellCmdIntArg("xintsteps");
	xintarg->setDescription("The number of parts in the x-direction in which the mass density should "
			        "be subdivided when integrating.");
	xintarg->setMin(16);
	xintarg->setMax(8192);
	xintarg->setDefault("128");
	addArgument(xintarg);
	yintarg = new ShellCmdIntArg("yintsteps");
	yintarg->setDescription("The number of parts in the y-direction in which the mass density should "
			        "be subdivided when integrating.");
	yintarg->setMin(16);
	yintarg->setMax(8192);
	yintarg->setDefault("128");
	addArgument(yintarg);
	
	m_pIgnoreSheet = new ShellCmdBoolArg("ignoresheet");
	m_pIgnoreSheet->setDefault("yes");
	m_pIgnoreSheet->setDescription("If set to true, a mass sheet lens present in a composite lens will be ignored.");
	addArgument(m_pIgnoreSheet);
	
	m_pKeepLargerArg = new ShellCmdBoolArg("keeplargersquares");
	m_pKeepLargerArg->setDefault("no");
	m_pKeepLargerArg->setDescription("If true, when a grid cell is subdivided, the original grid cell will also be kept.");
	addArgument(m_pKeepLargerArg);

	m_pGridName = new ShellCmdStringArg("gridname");
	m_pGridName->setDefault("*");
	m_pGridName->setDescription("Specify a name for the grid, is useful when combining multiple grids");
	addArgument(m_pGridName);

	m_pInitialSubDivArg = new ShellCmdIntArg("initialuniformsubdiv");
	m_pInitialSubDivArg->setDefault("1");
	m_pInitialSubDivArg->setDescription("If this is N, the grid is first organized into NxN squares and then subdivided further based on the contents of each cell.");
	addArgument(m_pInitialSubDivArg);

	setDescription("Creates a grid based on the mass distribution of the current lens, which can be "
		       "used in a lens inversion procedure.");
}

CmdGridNewLens::~CmdGridNewLens()
{
	delete centerxarg;
	delete centeryarg;
	delete sizearg;
	delete minsquaresarg;
	delete maxsquaresarg;
	delete xintarg;
	delete yintarg;
	delete m_pIgnoreSheet;
	delete m_pKeepLargerArg;
	delete m_pGridName;
	delete m_pInitialSubDivArg;
}

bool CmdGridNewLens::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *pCurrentLens, *pGridLens;
	grale::CompositeLens compositeLens;
	
	if ((pCurrentLens = gsinst->GetLens()) == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}

	if (m_pIgnoreSheet->getValue() == true)
	{
		if (pCurrentLens->getLensType() == grale::GravitationalLens::Composite) // it's a composite lens
		{
			const grale::CompositeLens *pCompLens = (const grale::CompositeLens *)pCurrentLens;
			bool found = false;
			int numSubLenses = pCompLens->getNumberOfSubLenses();

			for (int i = 0 ; !found && i < numSubLenses ; i++)
			{
				if (pCompLens->getSubLens(i)->getLensType() == grale::GravitationalLens::MassSheet)
					found = true;
			}

			if (!found)
				pGridLens = pCurrentLens;
			else
			{
				// TODO: also search sublevels in case a sublens is a composite lens itself 

				grale::CompositeLensParams newParams;
				int count = 0;

				for (int i = 0 ; i < numSubLenses ; i++)
				{
					const grale::GravitationalLens *pSubLens = pCompLens->getSubLens(i);

					if (pSubLens->getLensType() == grale::GravitationalLens::MassSheet)
						continue;

					grale::Vector2D<double> position = pCompLens->getSubLensPosition(i);
					double angle = pCompLens->getSubLensAngle(i);
					double factor = pCompLens->getSubLensFactor(i);

					if (!newParams.addLens(factor, position, angle, *pSubLens))
					{
						setErrorString(std::string("Coudln't create a new temporary lens without mass sheet: ") + newParams.getErrorString());
						return false;
					}

					count++;
				}

				if (count == 0)
				{
					setErrorString("Without mass-sheet sublenses, the composite lens doesn't contain any other lenses");
					return false;
				}

				if (!compositeLens.init(pCurrentLens->getLensDistance(), &newParams))
				{
					setErrorString(std::string("Couldn't initialize the temporary lens without the mass-sheet lenses: ") + compositeLens.getErrorString());
					return false;
				}

				// ok, initialization was successful

				pGridLens = &compositeLens;
			}

		}
		else
			pGridLens = pCurrentLens;
	}
	else
		pGridLens = pCurrentLens;

	grale::Vector2D<double> center(centerxarg->getValue(), centeryarg->getValue());
	grale::Vector2D<double> diff(sizearg->getValue(), sizearg->getValue());

	diff *= 1.01;
	
	grale::Real2DLensFunction lensfunction(*pGridLens);
	grale::GridFunction f(lensfunction, center-diff, center+diff, xintarg->getValue(), yintarg->getValue(), false);

	grale::Grid *g = new grale::Grid();


	if (!g->buildDensityMinMax(f, center, sizearg->getValue(), minsquaresarg->getValue(), maxsquaresarg->getValue(), 
				         xintarg->getValue(), yintarg->getValue(), false, m_pKeepLargerArg->getValue(),
			             m_pInitialSubDivArg->getValue()))
	{
		setErrorString(std::string("Couldn't build density based grid: ") + g->getErrorString());
		delete g;
		return false;
	}
	
	std::list<grale::GridSquare> squares;
	char buf[256];

	g->getSquares(squares);

	sprintf(buf, "Built grid with %d squares", (int)squares.size());
	gsinst->getIOSystem()->writeOutputLine(buf);
	
	string name = m_pGridName->getValue();
	if (name.length() == 0)
	{
		setErrorString("No grid name specified");
		return false;
	}

	if (name == "*") // the default
		name = "";

	gsinst->SetGrid(g, name);
	return true;
}

CmdGridPlot::CmdGridPlot(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pFileNameArg = new ShellCmdStringArg("filename");
	m_pFileNameArg->setDescription("Name of the file to which the grid data will be written.");
	addArgument(m_pFileNameArg);

	setDescription("Write the grid data to a file which can be plotted with gnuplot.");
}

CmdGridPlot::~CmdGridPlot()
{
	delete m_pFileNameArg;
}

bool CmdGridPlot::plotGrid(FILE *pFile, const string &name, const grale::Grid *pGrid)
{
	string gridName = name;

	if (gridName == "")
		gridName = "[default]";

	if (!pGrid)
	{
		setErrorString("Internal error: grid for name '" + name + "' is null");
		return false;
	}

	fprintf(pFile, "# Squares for grid %s\n", gridName.c_str());

	std::list<grale::GridSquare> squares;
	pGrid->getSquares(squares);

	for (auto it = squares.begin() ; it != squares.end() ; it++)
	{
		double x = (*it).getCenter().getX();
		double y = (*it).getCenter().getY();
		double s = (*it).getSize()/2.0;
		
		fprintf(pFile, "%g %g\n",x-s,y-s);
		fprintf(pFile, "%g %g\n",x-s,y+s);
		fprintf(pFile, "%g %g\n",x+s,y+s);
		fprintf(pFile, "%g %g\n",x+s,y-s);
		fprintf(pFile, "%g %g\n\n",x-s,y-s);
	}
	fprintf(pFile, "\n\n");

	return true;
}

bool CmdGridPlot::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	vector<string> gridNames;

	pInstance->GetGridNames(gridNames);

	grale::Grid *pGrid = pInstance->GetGrid(""); // Default grid
	if (pGrid == 0 && gridNames.size() == 0)
	{
		setErrorString("No grid has been created yet");
		return false;
	}

	FILE *pFile = fopen(m_pFileNameArg->getValue().c_str(), "wt");
	if (pFile == 0)
	{
		setErrorString("Couldn't open file for writing");
		return false;
	}
	
	// First process the default grid
	if (pGrid)
	{
		if (!plotGrid(pFile, "", pGrid))
		{
			fclose(pFile);
			return false; // error string was already set
		}
	}

	// Then process the rest of them
	for (size_t i = 0 ; i < gridNames.size() ; i++)
	{
		grale::Grid *pGrid = pInstance->GetGrid(gridNames[i]);
		if (!plotGrid(pFile, gridNames[i], pGrid))
		{
			fclose(pFile);
			return false; // error string was already set
		}
	}

	fclose(pFile);
	return true;
}

CmdGridList::CmdGridList(const string &cmdName) : ShellCommand(cmdName)
{
	setDescription("List the names of the grids that were defined");
}

CmdGridList::~CmdGridList()
{
}

bool CmdGridList::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInstance->getIOSystem();
	vector<string> gridNames;

	pInstance->GetGridNames(gridNames);

	char buf[256];
	grale::Grid *pGrid = pInstance->GetGrid(""); // Default grid
	if (pGrid == 0)
		sprintf(buf, "  No default grid was defined");
	else
	{
		std::list<grale::GridSquare> squares;
		pGrid->getSquares(squares);
		sprintf(buf, "  Default grid has %d squares", (int)squares.size());
	}
	
	pIOSys->writeOutputLine(buf);

	if (gridNames.size() == 0)
		pIOSys->writeOutputLine("  No named grids are defined");
	else
	{
		pIOSys->writeOutputLine("  Named grids:");
		for (size_t i = 0 ; i < gridNames.size() ; i++)
		{
			pGrid = pInstance->GetGrid(gridNames[i]);
			if (pGrid == 0)
			{
				setErrorString("Internal error: no grid for name '" + gridNames[i] + "'");
				return false;
			}

			std::list<grale::GridSquare> squares;
			pGrid->getSquares(squares);
			
			sprintf(buf, "%d squares", (int)squares.size());

			pIOSys->writeOutputLine("    '" + gridNames[i] + "': " + string(buf));
		}
	}

	return true;
}

CmdGridClear::CmdGridClear(const string &cmdName) : ShellCommand(cmdName)
{
	m_pName = new ShellCmdStringArg("gridname");
	m_pName->setDefault("*");
	m_pName->setDescription("Name of the grid to clear (use the default value for the default grid");
	addArgument(m_pName);

	m_pSureArg = new ShellCmdBoolArg("sure");
	m_pSureArg->setDefault("no");
	m_pSureArg->setDescription("To avoid accidentally clearing a grid, set this value to 'yes' if you're sure");
	addArgument(m_pSureArg);

	setDescription("Clear a grid that was previously set.");
}

CmdGridClear::~CmdGridClear()
{
	delete m_pName;
	delete m_pSureArg;
}

bool CmdGridClear::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();

	string name = m_pName->getValue();
	if (name.length() == 0)
	{
		setErrorString("No grid name specified");
		return false;
	}

	if (name == "*") // the default
		name = "";

	if (!m_pSureArg->getValue())
	{
		setErrorString("You need to specify 'yes' as the value for the last parameter if you're sure");
		return false;
	}

	if (!pInstance->GetGrid(name))
	{
		setErrorString("Specified grid was not found");
		return false;
	}

	pInstance->ClearGrid(name);

	return true;
}

CmdGridClearAll::CmdGridClearAll(const string &cmdName) : ShellCommand(cmdName)
{
	m_pSureArg = new ShellCmdBoolArg("sure");
	m_pSureArg->setDefault("no");
	m_pSureArg->setDescription("To avoid accidentally clearing the grids, set this value to 'yes' if you're sure");
	addArgument(m_pSureArg);

	setDescription("Clears all grids that were previously set.");
}

CmdGridClearAll::~CmdGridClearAll()
{
	delete m_pSureArg;
}

bool CmdGridClearAll::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();

	if (!m_pSureArg->getValue())
	{
		setErrorString("You need to specify 'yes' as the value for the parameter if you're sure");
		return false;
	}

	pInstance->ClearGrids();

	return true;
}

