/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "cmdimagesdata.h"
#include "cmdinvert.h"
#include "graleshellinstance.h"
#include <shellp/fileiosystem.h>
#include <grale/imagesdataextended.h>
#include <grale/constants.h>
#include <grale/gravitationallens.h>
#include <grale/imagesdata.h>
#include <grale/imagesbackprojector.h>
#include <grale/polygon2d.h>
#include <grale/galensmodule.h>
#include <grale/lensfitnessobject.h>
#include <grale/triangulation.h>
#include <grale/gridfunction.h>
#include <grale/real2dfunctionfitsplotter.h>
#include <grale/sourceplane.h>
#include <grale/gaussiandistribution.h>
#include <grale/randomnumbergenerator.h>
#include <grale/configurationparameters.h>
#include <assert.h>
#include <memory>
#include <map>

#include "debugnew.h"

using namespace shellp;
using namespace std;
using namespace grale;

CmdImagesDataLoadSave::CmdImagesDataLoadSave(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Name of the images data file.");
	addArgument(fnamearg);
}

CmdImagesDataLoadSave::~CmdImagesDataLoadSave()
{
	delete fnamearg;
}

CmdImagesDataLoad::CmdImagesDataLoad(const std::string &cmdname) : CmdImagesDataLoadSave(cmdname)
{
	setDescription("Load a file containing the images of a specific source.");
}

CmdImagesDataLoad::~CmdImagesDataLoad()
{
}

bool CmdImagesDataLoad::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat = new grale::ImagesDataExtended();

	if (!dat->load(GetFileName()))
	{
		setErrorString(std::string("Can't load the image data file: ") + dat->getErrorString());
		delete dat;
		return false;
	}

	gsinst->SetImagesData(dat);
	return true;
}

CmdImagesDataSave::CmdImagesDataSave(const std::string &cmdname) : CmdImagesDataLoadSave(cmdname)
{
	setDescription("Save the images data in memory to a file.");
}

CmdImagesDataSave::~CmdImagesDataSave()
{
}

bool CmdImagesDataSave::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat;

	if ((dat = gsinst->GetImagesData()) == 0)
	{
		setErrorString("No images data are loaded into memory.");
		return false;
	}
	
	if (!dat->save(GetFileName()))
	{
		setErrorString(std::string("Can't save the image data file: ") + dat->getErrorString());
		return false;
	}

	return true;
}

CmdImagesDataTriangulationCreate::CmdImagesDataTriangulationCreate(const std::string &cmdname) : ShellCommand(cmdname)
{
	setDescription("Creates a triangulation for the currently active images data");
}

CmdImagesDataTriangulationCreate::~CmdImagesDataTriangulationCreate()
{
}

bool CmdImagesDataTriangulationCreate::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pGSInst->getIOSystem();
	grale::ImagesDataExtended *pDat;

	if ((pDat = pGSInst->GetImagesData()) == 0)
	{
		setErrorString("No images data are loaded into memory.");
		return false;
	}

	pIOSys->writeOutputLine("Creating triangulation (this can take some time)");	
	
	if (!pDat->createTriangulation())
	{
		setErrorString(std::string("Unable to create triangulation: ") + pDat->getErrorString());
		return false;
	}
	
	return true;
}

CmdImagesDataTriangulationClear::CmdImagesDataTriangulationClear(const std::string &cmdname) : ShellCommand(cmdname)
{
	setDescription("Clears the triangulation data for the currently active images data");
}

CmdImagesDataTriangulationClear::~CmdImagesDataTriangulationClear()
{
}

bool CmdImagesDataTriangulationClear::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pGSInst->getIOSystem();
	grale::ImagesDataExtended *pDat;

	if ((pDat = pGSInst->GetImagesData()) == 0)
	{
		setErrorString("No images data are loaded into memory.");
		return false;
	}

	pDat->clearTriangulation();
	
	return true;
}

CmdImagesDataImport::CmdImagesDataImport(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Name of the file to import images data from.");
	addArgument(fnamearg);
	masknamearg = new ShellCmdStringArg("maskfile");
	masknamearg->setDescription("If set, this specifies a PNG file which will be used as a mask on the real data.");
	masknamearg->setDefault("*");
	addArgument(masknamearg);
	x1arg = new ShellCmdRealArg("leftX");
	x1arg->setUnitType(UNIT_ANGULAR);
	x1arg->setDescription("X coordinate of the left side of the images file.");
	addArgument(x1arg);
	y1arg = new ShellCmdRealArg("bottomY");
	y1arg->setUnitType(UNIT_ANGULAR);
	y1arg->setDescription("Y coordinate of the bottom of the images file.");
	addArgument(y1arg);
	x2arg = new ShellCmdRealArg("rightX");
	x2arg->setUnitType(UNIT_ANGULAR);
	x2arg->setDescription("X coordinate of the right side of the images file.");
	addArgument(x2arg);
	y2arg = new ShellCmdRealArg("topY");
	y2arg->setUnitType(UNIT_ANGULAR);
	y2arg->setDescription("Y coordinate of the left side of the images file.");
	addArgument(y2arg);
	nullarg = new ShellCmdBoolArg("useNullSpace");
	nullarg->setDefault("no");
	nullarg->setDescription("If enabled, the empty pixels right next to the images will be imported.");
	addArgument(nullarg);
	m_pTriangArg = new ShellCmdBoolArg("triangulate");
	m_pTriangArg->setDefault("yes");
	m_pTriangArg->setDescription("Flag indicating if we should try to create a triangulation based on the pixels in the image.");
	addArgument(m_pTriangArg);
}

CmdImagesDataImport::~CmdImagesDataImport()
{
	delete fnamearg;
	delete masknamearg;
	delete x1arg;
	delete x2arg;
	delete y1arg;
	delete y2arg;
	delete nullarg;
	delete m_pTriangArg;
}

CmdImagesDataImportPNG::CmdImagesDataImportPNG(const std::string &cmdname) : CmdImagesDataImport(cmdname)
{
	setDescription("Import data about the images of a specific source from a PNG file.");
}

CmdImagesDataImportPNG::~CmdImagesDataImportPNG()
{
}

bool CmdImagesDataImportPNG::execute()
{
#ifdef GRALECONFIG_SUPPORT_QT4
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat = new grale::ImagesDataExtended();

	if (!dat->loadPNG(GetFileName(),grale::Vector2D<double>(GetX1(),GetY1()),grale::Vector2D<double>(GetX2(),GetY2()),UseNullSpace(),GetMaskFileName(), getTriangulationFlag()))
	{
		setErrorString(std::string("Can't import the image data file: ") + dat->getErrorString());
		delete dat;
		return false;
	}

	gsinst->SetImagesData(dat);
	return true;
#else
	setErrorString("No PNG support was compiled in.");
	return false;
#endif // GRALECONFIG_SUPPORT_QT4
}

CmdImagesDataImportFITS::CmdImagesDataImportFITS(const std::string &cmdname) : CmdImagesDataImport(cmdname)
{
	setDescription("Import data about the images of a specific source from a FITS file.");
}

CmdImagesDataImportFITS::~CmdImagesDataImportFITS()
{
}

bool CmdImagesDataImportFITS::execute()
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat = new grale::ImagesDataExtended();

	if (!dat->loadFITS(GetFileName(),grale::Vector2D<double>(GetX1(),GetY1()),grale::Vector2D<double>(GetX2(),GetY2()),UseNullSpace(),GetMaskFileName(), getTriangulationFlag()))
	{
		setErrorString(std::string("Can't import the image data file: ") + dat->getErrorString());
		delete dat;
		return false;
	}

	gsinst->SetImagesData(dat);
	return true;
#else
	setErrorString("No cfitsio support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

CmdImagesDataImportFITSAuto::CmdImagesDataImportFITSAuto(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Name of the file to import images data from.");
	addArgument(fnamearg);
	masknamearg = new ShellCmdStringArg("maskfile");
	masknamearg->setDescription("If set, this specifies a PNG file which will be used as a mask on the real data.");
	masknamearg->setDefault("*");
	addArgument(masknamearg);
	unitmultarg = new ShellCmdRealArg("unitmultiplier");
	unitmultarg->setDescription("Multiplier for the base unit.");
	addArgument(unitmultarg);
	unitbasearg = new ShellCmdStringArg("unitbase");
	unitbasearg->setDescription("Base unit. Can be one of 'arcsec','arcmin','deg','rad'.");
	addArgument(unitbasearg);
	nullarg = new ShellCmdBoolArg("useNullSpace");
	nullarg->setDefault("no");
	nullarg->setDescription("If enabled, the empty pixels right next to the images will be imported");
	addArgument(nullarg);
	m_pTriangArg = new ShellCmdBoolArg("triangulate");
	m_pTriangArg->setDefault("yes");
	m_pTriangArg->setDescription("Flag indicating if we should try to create a triangulation based on the pixels in the image.");
	addArgument(m_pTriangArg);
	
	setDescription("Import data about the images of a source from a FITS file using coordinate info stored in the file. Primarily intended to be used on FITS files created by GRALE.");
}

CmdImagesDataImportFITSAuto::~CmdImagesDataImportFITSAuto()
{
	delete fnamearg;
	delete masknamearg;
	delete unitmultarg;
	delete unitbasearg;
	delete nullarg;
	delete m_pTriangArg;
}

bool CmdImagesDataImportFITSAuto::execute()
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	double baseunit = 0;
	std::string unitbasestr = unitbasearg->getValue();
	
	if (unitbasestr == std::string("arcsec"))
		baseunit = ANGLE_ARCSEC;
	else if (unitbasestr == std::string("arcmin"))
		baseunit = ANGLE_ARCMIN;
	else if (unitbasestr == std::string("deg"))
		baseunit = ANGLE_DEGREE;
	else if (unitbasestr == std::string("rad"))
		baseunit = 1.0;
	else
	{
		setErrorString("Invalid base unit (expecting 'arcsec', 'arcmin', 'deg' or 'rad')");
		return false;
	}

	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat = new grale::ImagesDataExtended();
	std::string maskname("");
	
	if (masknamearg->getValue() != std::string("*"))
		maskname = masknamearg->getValue();
	
	if (!dat->loadFITS(fnamearg->getValue(),unitmultarg->getValue()*baseunit,nullarg->getValue(),maskname, m_pTriangArg->getValue()))
	{
		setErrorString(std::string("Can't import the image data file: ") + dat->getErrorString());
		delete dat;
		return false;
	}

	gsinst->SetImagesData(dat);
	return true;
#else
	setErrorString("No cfitsio support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

CmdImagesDataImportFITSWCS::CmdImagesDataImportFITSWCS(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Name of the file to import images data from.");
	addArgument(fnamearg);
	masknamearg = new ShellCmdStringArg("maskfile");
	masknamearg->setDescription("If set, this specifies a PNG file which will be used as a mask on the real data.");
	masknamearg->setDefault("*");
	addArgument(masknamearg);
	nullarg = new ShellCmdBoolArg("useNullSpace");
	nullarg->setDefault("no");
	nullarg->setDescription("If enabled, the empty pixels right next to the images will be imported");
	addArgument(nullarg);
	m_pTriangArg = new ShellCmdBoolArg("triangulate");
	m_pTriangArg->setDefault("yes");
	m_pTriangArg->setDescription("Flag indicating if we should try to create a triangulation based on the pixels in the image.");
	addArgument(m_pTriangArg);
	
	setDescription("Import data about the images of a source from a FITS file using coordinate info stored in the file. The WCSTOOLS library is used to interpret pixel coordinates.");
}

CmdImagesDataImportFITSWCS::~CmdImagesDataImportFITSWCS()
{
	delete fnamearg;
	delete masknamearg;
	delete nullarg;
	delete m_pTriangArg;
}

bool CmdImagesDataImportFITSWCS::execute()
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *dat = new grale::ImagesDataExtended();
	std::string maskname("");
	
	if (masknamearg->getValue() != std::string("*"))
		maskname = masknamearg->getValue();
	
	if (!dat->loadFITS(fnamearg->getValue(),nullarg->getValue(),maskname, m_pTriangArg->getValue()))
	{
		setErrorString(std::string("Can't import the image data file: ") + dat->getErrorString());
		delete dat;
		return false;
	}

	gsinst->SetImagesData(dat);
	return true;
#else
	setErrorString("No cfitsio support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}


CmdImagesDataSourceImage::CmdImagesDataSourceImage(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pImageNumberArg = new ShellCmdIntArg("imgnum");
	m_pImageNumberArg->setDescription("Index of the image which is used in the reconstruction. You can set this to -1 to select all images.");
	m_pImageNumberArg->setMin(-1);
	m_pImageNumberArg->setDefault("-1");
	addArgument(m_pImageNumberArg);
	m_pDsArg = new ShellCmdRealArg("D_s");
	m_pDsArg->setDescription("Angular diameter distance between this source and the observer.");
	m_pDsArg->setUnitType(UNIT_DISTANCE);
	m_pDsArg->setMin(0);
	addArgument(m_pDsArg);
	m_pDdsArg = new ShellCmdRealArg("D_ds");
	m_pDdsArg->setDescription("Angular diameter distance between this source and the lens.");
	m_pDdsArg->setUnitType(UNIT_DISTANCE);
	m_pDdsArg->setMin(0);
	addArgument(m_pDdsArg);
	m_pFileArg = new ShellCmdStringArg("filename");
	m_pFileArg->setDescription("Name of the FITS file to store the source image in");
	addArgument(m_pFileArg);
	m_pNumXArg = new ShellCmdIntArg("numx");
	m_pNumXArg->setDescription("Width (pixels) of the resulting FITS file");
	m_pNumXArg->setDefault("100");
	m_pNumXArg->setMin(5);
	addArgument(m_pNumXArg);
	m_pNumYArg = new ShellCmdIntArg("numy");
	m_pNumYArg->setDescription("Height (pixels) of the resulting FITS file");
	m_pNumYArg->setDefault("100");
	m_pNumYArg->setMin(5);
	addArgument(m_pNumYArg);
	m_pFastArg = new ShellCmdBoolArg("fast");
	m_pFastArg->setDescription("Flag indicating whether to use the fast but less accurate algorithm");
	m_pFastArg->setDefault("yes");
	addArgument(m_pFastArg);
	m_pResampleArg = new ShellCmdIntArg("resamplesize");
	m_pResampleArg->setDescription("The points will first be calculated on a grid 'resamplesize' times "
			               "larger in each direction, and will then be resized to the requested "
				       "size. This may yield better results if the fast algorithm is used.");
	m_pResampleArg->setDefault("1");
	m_pResampleArg->setMin(1);
	addArgument(m_pResampleArg);
	m_pUseMagnifications = new ShellCmdBoolArg("demagnify");
	m_pUseMagnifications->setDefault("no");
	m_pUseMagnifications->setDescription("Use the calculated magnification info to demagnify the value of each point in the images data. This "
			                     "should only be done if the data describes point sources.");
	addArgument(m_pUseMagnifications);
	setDescription("Uses the current lens to project the images data onto the sourceplane and saves the "
		       "resulting source image as a FITS file");
}

CmdImagesDataSourceImage::~CmdImagesDataSourceImage()
{
	delete m_pImageNumberArg;
	delete m_pDsArg;
	delete m_pDdsArg;
	delete m_pFileArg;
	delete m_pNumXArg;
	delete m_pNumYArg;
	delete m_pFastArg;
	delete m_pResampleArg;
	delete m_pUseMagnifications;
}

bool CmdImagesDataSourceImage::execute()
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pGSInst->getIOSystem();
	grale::GravitationalLens *pLens = pGSInst->GetLens();
	grale::ImagesDataExtended *pImgDat;
	int numImgs;

	if ((pImgDat = pGSInst->GetImagesData()) == 0)
	{
		setErrorString("No images data are loaded into memory");
		return false;
	}

	if (pLens == 0)
	{
		setErrorString("No gravitational lens has been created yet");
		return false;
	}

	if ((numImgs = pImgDat->getNumberOfImages()) <= 0)
	{
		setErrorString("Current images data doesn't contain any images");
		return false;
	}

	int imgNum = m_pImageNumberArg->getValue();

	if (imgNum >= 0 && imgNum >= numImgs)
	{
		char str[256];

		sprintf(str,"%d",numImgs-1);
		setErrorString(std::string("The image index must lie in the range 0..")  + std::string(str));
		return false;
	}
	// Retrieve the triangulation data

	std::vector<std::list<grale::TriangleIndices> > triangulations(numImgs);

	if (!pImgDat->hasTriangulation())
	{
		setErrorString("Current images data doesn't have a triangulation yet");
		return false;
	}

	for (int i = 0 ; i < numImgs ; i++)
	{
		if (imgNum == -1 || imgNum == i)
		{
			if (!pImgDat->getTriangles(i, triangulations[i]))
			{
				setErrorString(std::string("Unable to retrieve triangle data: ") + pImgDat->getErrorString());
				return false;
			}
		}
	}

	// Project the images back onto the source plane

	pIOSys->writeOutputLine("Projecting the image points back onto the source plane");

	std::list<grale::ImagesDataExtended *> imagesDataList;
	
	pImgDat->setDs(m_pDsArg->getValue());
	pImgDat->setDds(m_pDdsArg->getValue());

	imagesDataList.push_back(pImgDat);

	grale::ImagesBackProjector bp(*pLens, imagesDataList, 0, false); // z_d doesn't matter here
	
	// Create triangle list and store min and max positions
	
	std::list<grale::Triangle2DPlus<double> > allTriangles;
	double minX = 1e30;
	double maxX = -1e30;
	double minY = 1e30;
	double maxY = -1e30;

	for (int i = 0 ; i < numImgs ; i++)
	{
		const grale::Vector2D<float> *pBetas = bp.getBetas(0, i); // only one source
		const float *pIntensities = bp.getOriginalIntensities(0, i);
		const float *pInvMag = bp.getInverseMagnifications(0, i);
		std::list<grale::TriangleIndices>::const_iterator it, endIt;

		it = triangulations[i].begin();
		endIt = triangulations[i].end();
		
		while (it != endIt)
		{
			grale::TriangleIndices indices = (*it);
			grale::Vector2D<float> fp1 = pBetas[indices.getIndex(0)];
			grale::Vector2D<float> fp2 = pBetas[indices.getIndex(1)];
			grale::Vector2D<float> fp3 = pBetas[indices.getIndex(2)];
			grale::Vector2D<double> p1 = grale::Vector2D<double>(fp1.getX(), fp1.getY());
			grale::Vector2D<double> p2 = grale::Vector2D<double>(fp2.getX(), fp2.getY());
			grale::Vector2D<double> p3 = grale::Vector2D<double>(fp3.getX(), fp3.getY());
			double h1 = pIntensities[indices.getIndex(0)];
			double h2 = pIntensities[indices.getIndex(1)];
			double h3 = pIntensities[indices.getIndex(2)];

			if (m_pUseMagnifications->getValue())
			{
				h1 *= (double)ABS(pInvMag[indices.getIndex(0)]);
				h2 *= (double)ABS(pInvMag[indices.getIndex(1)]);
				h3 *= (double)ABS(pInvMag[indices.getIndex(2)]);
			}
			
			minX = MIN(minX, p1.getX());
			minX = MIN(minX, p2.getX());
			minX = MIN(minX, p3.getX());
			minY = MIN(minY, p1.getY());
			minY = MIN(minY, p2.getY());
			minY = MIN(minY, p3.getY());
			maxX = MAX(maxX, p1.getX());
			maxX = MAX(maxX, p2.getX());
			maxX = MAX(maxX, p3.getX());
			maxY = MAX(maxY, p1.getY());
			maxY = MAX(maxY, p2.getY());
			maxY = MAX(maxY, p3.getY());

			allTriangles.push_back(grale::Triangle2DPlus<double>(p1, p2, p3, h1, h2, h3));

			it++;
		}
	}

	// Create gridfunction and plot it

	pIOSys->writeOutputLine("Calculating source image points");
	
	grale::GridFunction gridFunction(allTriangles, grale::Vector2D<double>(minX, minY), grale::Vector2D<double>(maxX, maxY), m_pNumXArg->getValue(), m_pNumYArg->getValue(), false, m_pFastArg->getValue(), m_pResampleArg->getValue());

	pIOSys->writeOutputLine("Creating plot file");

	grale::Real2DFunctionFITSPlotter plotter(m_pFileArg->getValue(), pGSInst->GetAngularUnit()/bp.getAngularScale(), pGSInst->GetAngularUnit()/bp.getAngularScale(), 1.0/bp.getIntensityScale()); // TODO: scales!

	if (!plotter.plot(gridFunction, grale::Vector2D<double>(minX, minY), grale::Vector2D<double>(maxX, maxY), m_pNumXArg->getValue(), m_pNumYArg->getValue()))
	{
		setErrorString(std::string("Couldn't plot source image: ") + plotter.getErrorString());
		return false;
	}
	return true;
#else
	setErrorString("No CFITSIO support was available at compile time");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO

}

CmdImagesDataListAdd::CmdImagesDataListAdd(const std::string &cmdname) : ShellCommand(cmdname)
{
	Dsarg = new ShellCmdRealArg("D_s");
	Dsarg->setUnitType(UNIT_DISTANCE);
	Dsarg->setDescription("The angular diameter distance between source and observer for these images.");
	addArgument(Dsarg);
	Ddsarg = new ShellCmdRealArg("D_ds");
	Ddsarg->setUnitType(UNIT_DISTANCE);
	Ddsarg->setDescription("The angular diameter distance between source and lens for these images.");
	addArgument(Ddsarg);
	extraarg = new ShellCmdBoolArg("extraparameters");
	extraarg->setDefault("no");
	extraarg->setDescription("Flag indicating if extra parameters should be assigned to the images data set.");
	addArgument(extraarg);
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDefault("*");
	fnamearg->setDescription("Specifies the file which contains the images data. By default, the "
			         "currently loaded images data are used.");
	addArgument(fnamearg);

	m_pTypeArg = new ShellCmdStringArg("type");
	m_pTypeArg->setDefault("*");
	m_pTypeArg->setDescription("For some lens inversion modules, each images data set needs to have a type name. "
			                   "This can be specified as this argument - possible values will depend on the "
							   "specific module.");
	addArgument(m_pTypeArg);
	
	setDescription("Adds images data to the list of images data. Can be used when using the inversion "
		       "procedure.");

	valarg = new ShellCmdRealArg("value");
	imgargs.push_back(valarg);

	m_pKeyNameArg = new ShellCmdStringArg("key");
	
#define VALUETYPE_BOOL	0
#define VALUETYPE_INT	1
#define VALUETYPE_REAL	2
#define VALUETYPE_STR	3

	m_pValueTypeArg = new ShellCmdChoiceArg("valuetype");
	m_pValueTypeArg->registerChoice("bool", "A boolean value", VALUETYPE_BOOL);
	m_pValueTypeArg->registerChoice("int", "An integer value", VALUETYPE_INT);
	m_pValueTypeArg->registerChoice("real", " A real value", VALUETYPE_REAL);
	m_pValueTypeArg->registerChoice("str", "A character string", VALUETYPE_STR);

	m_pValueArgStr = new ShellCmdStringArg("value");
	m_pValueArgBool = new ShellCmdBoolArg("value");
	m_pValueArgInt = new ShellCmdIntArg("value");
	m_pValueArgReal = new ShellCmdRealArg("value");

	m_typedImageArgsStr.push_back(m_pKeyNameArg);
	m_typedImageArgsStr.push_back(m_pValueTypeArg);
	m_typedImageArgsStr.push_back(m_pValueArgStr);

	m_typedImageArgsBool.push_back(m_pKeyNameArg);
	m_typedImageArgsBool.push_back(m_pValueTypeArg);
	m_typedImageArgsBool.push_back(m_pValueArgBool);

	m_typedImageArgsInt.push_back(m_pKeyNameArg);
	m_typedImageArgsInt.push_back(m_pValueTypeArg);
	m_typedImageArgsInt.push_back(m_pValueArgInt);

	m_typedImageArgsReal.push_back(m_pKeyNameArg);
	m_typedImageArgsReal.push_back(m_pValueTypeArg);
	m_typedImageArgsReal.push_back(m_pValueArgReal);
}

CmdImagesDataListAdd::~CmdImagesDataListAdd()
{
	delete Dsarg;
	delete Ddsarg;
	delete extraarg;
	delete fnamearg;
	delete valarg;
	delete m_pTypeArg;
	delete m_pKeyNameArg;
	delete m_pValueTypeArg;
	delete m_pValueArgStr;
	delete m_pValueArgBool;
	delete m_pValueArgInt;
	delete m_pValueArgReal;
}

bool CmdImagesDataListAdd::executeOld(grale::ImagesDataExtended *imgdata)
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *iosys = gsinst->getIOSystem();

	iosys->writeOutputLine("Enter extra parameters below. Type 'end' to end input.");
	
	bool done = false;
	bool error = false;
	int count = 0;
	int idx = 0;
	while (!done)
	{
		std::string line;
		char str[1024];

		sprintf(str,"Extra parameter %d: ",count++);
		std::string prefix(str);

		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;

			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = args[0];

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(imgargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing extra parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid

					char numStr[256];
					sprintf(numStr, "%d", idx);
					idx++;

					double val = valarg->getValue();
					imgdata->setExtraParameter(numStr, val);
				}
			}
		}
	}

	if (error)
		return false;

	return true;
}

bool CmdImagesDataListAdd::executeNew(grale::ImagesDataExtended *pImgDat)
{
	GraleShellInstance *pInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	pIOSys->writeOutputLine("Enter the extra data below. Type 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_typedImageArgsStr));

	std::string prefix("Extra info: ");
	bool done = false;
	bool error = false;

	while (!done)
	{
		std::string line;
		
		if (!pIOSys->readInputLine(prefix, line))
		{
			done = true;
			break;
		}

		if (pIOSys->echo())
			pIOSys->print("%s%s", prefix.c_str(), line.c_str());

		std::vector<std::string> arguments;

		IOSystem::splitLine(line, arguments);
		
		if (arguments.size() == 1)
		{
			std::string str = arguments[0];

			if (str == std::string("end"))
			{
				done = true;
				break;
			}
		}
		
		std::string errorString;

		// First parse value as string, if that works see if the valuetype parameter
		// says it should be parsed as somethings else
		if (!ShellCmdArg::parseArgumentList(m_typedImageArgsStr, arguments, errorString))
		{
			error = true;
			done = true;
			setErrorString(std::string("Error parsing extra arguments: ") + errorString);
			break;
		}

		string key = m_pKeyNameArg->getValue();
		const int valueType = m_pValueTypeArg->getValue();
		if (valueType == VALUETYPE_STR)
		{
			// Already interpreted as a string

			pImgDat->setExtraParameter(key, m_pValueArgStr->getValue());
			//pIOSys->print("Added string %s %s", key.c_str(), m_pValueArgStr->getValue().c_str());
		}
		else if (valueType == VALUETYPE_BOOL)
		{
			if (!ShellCmdArg::parseArgumentList(m_typedImageArgsBool, arguments, errorString))
			{
				error = true;
				done = true;
				setErrorString(std::string("Error parsing extra arguments: ") + errorString);
				break;
			}

			pImgDat->setExtraParameter(key, m_pValueArgBool->getValue());
			//pIOSys->print("Added bool %s %d", key.c_str(), (int)m_pValueArgBool->getValue());
		}
		else if (valueType == VALUETYPE_INT)
		{
			if (!ShellCmdArg::parseArgumentList(m_typedImageArgsInt, arguments, errorString))
			{
				error = true;
				done = true;
				setErrorString(std::string("Error parsing extra arguments: ") + errorString);
				break;
			}

			pImgDat->setExtraParameter(key, m_pValueArgInt->getValue());
			//pIOSys->print("Added int %s %d", key.c_str(), m_pValueArgInt->getValue());
		}
		else if (valueType == VALUETYPE_REAL)
		{
			if (!ShellCmdArg::parseArgumentList(m_typedImageArgsReal, arguments, errorString))
			{
				error = true;
				done = true;
				setErrorString(std::string("Error parsing extra arguments: ") + errorString);
				break;
			}

			pImgDat->setExtraParameter(key, m_pValueArgReal->getValue());
			//pIOSys->print("Added real %s %g", key.c_str(), m_pValueArgReal->getValue());
		}
		else
		{
			error = true;
			done = true;
			setErrorString("Internal error: unknown value type in extra parameters");
			break;
		}
	}

	if (error)
		return false;

	return true;
}

bool CmdImagesDataListAdd::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *iosys = gsinst->getIOSystem();
	grale::ImagesDataExtended *imgdata = 0;
	
	if (fnamearg->getValue() != std::string("*")) // load file
	{
		imgdata = new grale::ImagesDataExtended(Dsarg->getValue(), Ddsarg->getValue());
		imgdata->clearExtraParameters();

		if (!imgdata->load(fnamearg->getValue()))
		{
			setErrorString(std::string("Couldn't load specified file: ") + imgdata->getErrorString());
			delete imgdata;
			return false;
		}
	}
	else // add currently loaded data
	{
		grale::ImagesDataExtended *curimgdata;
		
		if ((curimgdata = gsinst->GetImagesData()) == 0)
		{
			setErrorString("No currently loaded images data available");
			return false;
		}

		imgdata = new grale::ImagesDataExtended(*curimgdata);
		imgdata->setDs(Dsarg->getValue());
		imgdata->setDds(Ddsarg->getValue());
		imgdata->clearExtraParameters();
	}

	if (m_pTypeArg->getValue() == "*") // Nothing specified as type name, use old method to keep scripts working
	{
		if (extraarg->getValue())
		{
			if (!executeOld(imgdata)) 
			{
				// Error has already been set
				delete imgdata;
				return false;
			}
		}
	}
	else // Use new method
	{
		if (extraarg->getValue())
		{
			if (!executeNew(imgdata))
			{
				// Error has already been set
				delete imgdata;
				return false;
			}
		}
		// Do this after executeNew, so that it overrides a 'type' key that was set there
		imgdata->setExtraParameter("type", m_pTypeArg->getValue());
	}
	
	gsinst->AddToImagesDataList(imgdata);

	double distanceFraction = Ddsarg->getValue()/Dsarg->getValue();
	char str[1024];

	sprintf(str,"Distance fraction: %10.10g", distanceFraction);

	gsinst->getIOSystem()->writeOutputLine(std::string(str));

	return true;
}

CmdImagesDataListClear::CmdImagesDataListClear(const std::string &cmdname) : ShellCommand(cmdname)
{
	setDescription("Clears the list containing images data.");
}

CmdImagesDataListClear::~CmdImagesDataListClear()
{
}

bool CmdImagesDataListClear::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	gsinst->ClearImagesDataList();
	return true;
}

CmdImagesDataBackProjectAll::CmdImagesDataBackProjectAll(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("File name to which the back projected images data should be stored.");
	addArgument(fnamearg);

	m_pUseTrianglesArg = new ShellCmdBoolArg("usetriangles");
	m_pUseTrianglesArg->setDefault("false");
	m_pUseTrianglesArg->setDescription("If set, and if an image data set in the list contains a triangulation "
			                   "(e.g. for null space data), the output file will contain those back-projected " 
					   "triangles instead of just back-projected points.");
	addArgument(m_pUseTrianglesArg);

	setDescription("Using the current lens, all the images data in the list will be projected back "
		       "onto the their source planes.");
}

CmdImagesDataBackProjectAll::~CmdImagesDataBackProjectAll()
{
	delete fnamearg;
	delete m_pUseTrianglesArg;
}

bool CmdImagesDataBackProjectAll::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens = gsinst->GetLens();
	std::list<grale::ImagesDataExtended *> imagesdata;
	double angularunit = gsinst->GetAngularUnit();
	
	if (lens == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}

	gsinst->GetImagesDataList(imagesdata);

	if (imagesdata.empty())
	{
		setErrorString("No images data in list.");
		return false;
	}

	// TODO: set a better angular distance scale
	lens->setDerivativeAngularDistanceScale(angularunit);

	grale::ImagesBackProjector bp(*lens, imagesdata, 0, false); // z_d doesn't matter here
	
	if (!m_pUseTrianglesArg)
	{
		if (!bp.write(fnamearg->getValue(),angularunit, false))
		{
			setErrorString(bp.getErrorString());
			return false;
		}
	}
	else
	{
		FILE *pFile = fopen(fnamearg->getValue().c_str(), "wt");
		if (pFile == 0)
		{
			setErrorString("Unable to open the specified output file for writing");
			return false;
		}

		int numSources = bp.getNumberOfSources();
		assert(numSources == imagesdata.size());
		std::list<grale::ImagesDataExtended *>::const_iterator it = imagesdata.begin();

		for (int s = 0 ; s < numSources ; s++, ++it)
		{
			const grale::ImagesDataExtended *pImgData = *it;

			if (pImgData->hasTriangulation())
			{
				int numImages = pImgData->getNumberOfImages();
				assert(numImages == bp.getNumberOfImages(s));

				for (int img = 0 ; img < numImages ; img++)
				{
					std::list<grale::TriangleIndices> triangles;

					if (!pImgData->getTriangles(img, triangles))
					{
						setErrorString("Unable to get triangulation for an image: " + pImgData->getErrorString());
						return false;
					}

					std::list<grale::TriangleIndices>::const_iterator trIt = triangles.begin();
					for ( ; trIt != triangles.end() ; ++trIt)
					{
						grale::TriangleIndices indices = *trIt;
						for (int tp = 0 ; tp < 4 ; tp++) // we use 4 to get the starting point again
						{
							int pointIdx = indices.getIndex(tp%3);
							assert(pointIdx < bp.getNumberOfImagePoints(s, img));

							grale::Vector2D<float> beta = bp.getBetas(s, img)[pointIdx];
							beta *= (bp.getAngularScale()/angularunit);

							double height = 0;
							
							if (bp.hasOriginalIntensities(s)) 
							{
								double intens = bp.getOriginalIntensities(s, img)[pointIdx];

								height = intens*bp.getIntensityScale();
							}

							fprintf(pFile, "%g %g %g\n", (double)beta.getX(), (double)beta.getY(), (double)height);
						}
						fprintf(pFile, "\n");
					}
				}
			}
			else
			{
				// TODO: this code is a copy-paste from code in imagesbackprojector.cpp, it
				//       should actually be a subroutine there that we can use
				int numImages = bp.getNumberOfImages(s);

				for (int i = 0 ; i < numImages ; i++)
				{
					int numPoints = bp.getNumberOfImagePoints(s,i);

					for (int p = 0 ; p < numPoints ; p++)
					{
						grale::Vector2D<float> beta = bp.getBetas(s, i)[p];
						beta *= (bp.getAngularScale()/angularunit);

						double height = 0;

						if (bp.hasOriginalIntensities(s))
						{
							double intens = bp.getOriginalIntensities(s, i)[p];

							height = intens*bp.getIntensityScale();
						}

						fprintf(pFile, "%g %g %g\n", (double)beta.getX(), (double)beta.getY(), (double)height);
					}
					fprintf(pFile, "\n");
				}
			}
			

			fprintf(pFile, "\n\n");
		}

		fclose(pFile);
	}
		
	return true;
}

CmdImagesDataSelect::CmdImagesDataSelect(const std::string &cmdname) : ShellCommand(cmdname)
{
	numarg = new ShellCmdIntArg("index");
	numarg->setMin(0);
	numarg->setDescription("Index of the images data to select (starting with 0).");
	addArgument(numarg);
	
	setDescription("Selects an images data instance from the images data list.");
}

CmdImagesDataSelect::~CmdImagesDataSelect()
{
	delete numarg;
}

bool CmdImagesDataSelect::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	std::list<grale::ImagesDataExtended *> imgslist;

	gsinst->GetImagesDataList(imgslist);
	if (imgslist.empty())
	{
		setErrorString("No images data instances in the list");
		return false;
	}

	int idx = numarg->getValue();
	if (idx >= imgslist.size())
	{
		char buf[256];

		sprintf(buf,"The highest allowed index value is currently %d", (int)(imgslist.size()-1));
		setErrorString(buf);
		return false;
	}

	std::list<grale::ImagesDataExtended *>::const_iterator it = imgslist.begin();
	for (int i = 0 ; i < idx ; i++)
		it++;

	grale::ImagesDataExtended *newdata = new grale::ImagesDataExtended(*(*it));
	newdata->setDs(0);
	newdata->setDds(0);
	newdata->clearExtraParameters();
	gsinst->SetImagesData(newdata);
	return true;
}

CmdImagesDataFitness::CmdImagesDataFitness(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("algorithm");
	fnamearg->setDescription("The name of the inversion algorithm from which the fitness function should be used");
	addArgument(fnamearg);
	m_pZArg = new ShellCmdRealArg("z_lens");
	m_pZArg->setDescription("The redshift of the lens (should only matter if time-delay data is used).");
	addArgument(m_pZArg);

	setDescription("Projects the images data in the list onto the source plane and calculates the fitness value of this situation using a specific genetic algorithm implementation.");
}

CmdImagesDataFitness::~CmdImagesDataFitness()
{
	delete fnamearg;
	delete m_pZArg;
}

bool CmdImagesDataFitness::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens = gsinst->GetLens();
	std::list<grale::ImagesDataExtended *> imagesdata;
	double angularunit = gsinst->GetAngularUnit();
	
	if (lens == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}

	gsinst->GetImagesDataList(imagesdata);

	if (imagesdata.empty())
	{
		setErrorString("No images data in list.");
		return false;
	}

	std::string basedir, modulename, errStr;
	if (!CmdInvertGridBase::GetModuleDirectory(basedir, errStr))
	{
		setErrorString("Can't get module directory: " + errStr);
		return false;
	}
	CmdInvertGridBase::GetModuleName(fnamearg->getValue(), modulename);

	grale::GALensModule module;
	if (!module.open(basedir, modulename))
	{
		setErrorString(std::string("Couldn't load factory module '") + modulename + std::string("': ") + module.getErrorString());
		return false;
	}

	unique_ptr<LensFitnessObject> fitnessobject(module.createFitnessObject());
	if (fitnessobject == 0)
	{
		setErrorString("The CreateFitnessObject function was not able to create a valid fitness object");
		return false;
	}

	std::list<grale::ImagesDataExtended *> dummy;
	double z_d = m_pZArg->getValue();

	ConfigurationParameters params;
	if (!CmdInvertGridBase::getModuleParameters(gsinst, fnamearg->getValue(), params, errStr))
	{
		setErrorString(errStr);
		return false;
	}

	if (!fitnessobject->init(z_d, imagesdata, dummy, &params)) // The imagesdata list can be modified: data which does not need to be projected back can be removed
	{
		setErrorString(std::string("Unable to initialize the fitness object: ") + fitnessobject->getErrorString());
		return false;
	}

	int numcomp = fitnessobject->getNumberOfFitnessComponents();
	float *fitnessvalues = new float[numcomp];
	grale::ImagesBackProjector bp(*lens, imagesdata, z_d, false); // uses the possibly modified imagesdata

	fitnessobject->postInit(imagesdata, dummy, bp.getAngularScale());
	fitnessobject->calculateOverallFitness(bp, fitnessvalues);

	char str[1024];
	sprintf(str,"Fitness: ");
	char *p = str+strlen(str);

	for (int i = 0 ; i < numcomp ; i++)
	{
		sprintf(p,"%g ",(double)fitnessvalues[i]);
		p += strlen(p);
	}

	delete [] fitnessvalues;

	gsinst->getIOSystem()->writeOutputLine(std::string(str));

	return true;
}

CmdImagesDataRegularBase::CmdImagesDataRegularBase(const std::string &cmdname) : ShellCommand(cmdname)
{
	x1arg = new ShellCmdRealArg("leftX");
	x1arg->setDescription("X coordinate of the left side of the area.");
	x1arg->setUnitType(UNIT_ANGULAR);
	addArgument(x1arg);

	y1arg = new ShellCmdRealArg("bottomY");
	y1arg->setDescription("Y coordinate of the bottom of the area.");
	y1arg->setUnitType(UNIT_ANGULAR);
	addArgument(y1arg);

	x2arg = new ShellCmdRealArg("rightX");
	x2arg->setDescription("X coordinate of the right side of the area.");
	x2arg->setUnitType(UNIT_ANGULAR);
	addArgument(x2arg);

	y2arg = new ShellCmdRealArg("topY");
	y2arg->setDescription("Y coordinate of the top of the area.");
	y2arg->setUnitType(UNIT_ANGULAR);
	addArgument(y2arg);

	numxarg = new ShellCmdIntArg("numX","32");
	numxarg->setMin(2);
	numxarg->setDescription("Number of points in the X direction.");
	addArgument(numxarg);

	numyarg = new ShellCmdIntArg("numY","32");
	numyarg->setMin(2);
	numyarg->setDescription("Number of points in the Y direction.");
	addArgument(numyarg);

	m_pNeedTriangArg = new ShellCmdBoolArg("needtriangles");
	m_pNeedTriangArg->setDefault("yes");
	m_pNeedTriangArg->setDescription("Flag indicating if the procedure should exit if one of the images in the "
			"images data list does not contain a triangulation.");
	addArgument(m_pNeedTriangArg);

	m_pTriangulateArg = new ShellCmdBoolArg("triangulate");
	m_pTriangulateArg->setDefault("yes");
	m_pTriangulateArg->setDescription("Flag indicating if a triangulation should be created.");
	addArgument(m_pTriangulateArg);
}


CmdImagesDataRegularBase::~CmdImagesDataRegularBase()
{
	delete x1arg;
	delete y1arg;
	delete x2arg;
	delete y2arg;
	delete numxarg;
	delete numyarg;
	delete m_pNeedTriangArg;
	delete m_pTriangulateArg;
}

bool CmdImagesDataRegularBase::getAvoidTriangles(std::list<grale::Triangle2D<double> > &triangles)
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	std::list<grale::ImagesDataExtended *> imagesdata;
	std::list<grale::ImagesDataExtended *>::const_iterator it;

	gsinst->GetImagesDataList(imagesdata);
	triangles.clear();

	for (it = imagesdata.begin() ; it != imagesdata.end() ; it++)
	{
		grale::ImagesDataExtended *pImages = (*it);

		if (!pImages->hasTriangulation())
		{
			if (m_pNeedTriangArg->getValue() == true)
			{
				setErrorString("An images data set in the list does not contain a triangulation");
				return false;
			}
		}

		int numImages = pImages->getNumberOfImages();

		for (int i = 0 ; i < numImages ; i++)
		{
			std::list<grale::TriangleIndices> indices;
			std::list<grale::TriangleIndices>::const_iterator it;

			if (!pImages->getTriangles(i, indices))
			{
				setErrorString(std::string("Error retrieving triangles from an image: ") + pImages->getErrorString());
				return false;
			}

			for (it	= indices.begin() ; it != indices.end() ; it++)
			{
				grale::TriangleIndices idx = (*it);
				grale::Vector2D<double> p1 = pImages->getImagePointPosition(i, idx.getIndex(0));
				grale::Vector2D<double> p2 = pImages->getImagePointPosition(i, idx.getIndex(1));
				grale::Vector2D<double> p3 = pImages->getImagePointPosition(i, idx.getIndex(2));

				triangles.push_back(grale::Triangle2D<double>(p1,p2,p3));
			}
		}
	}

	return true;
}

CmdImagesDataNewRegular::CmdImagesDataNewRegular(const std::string &cmdname) : CmdImagesDataRegularBase(cmdname)
{
	m_pIntensFlagArg = new ShellCmdBoolArg("useintens");
	m_pIntensFlagArg->setDefault("no");
	m_pIntensFlagArg->setDescription("Flag indicating if zero intensities should be set for the image points.");
	addArgument(m_pIntensFlagArg);
	m_pCutArg = new ShellCmdBoolArg("cut");
	m_pCutArg->setDefault("no");
	m_pCutArg->setDescription("If true, a finer cut will be made for the images which should be excluded. In this "
			          "case, the images do need triangulation.");
	addArgument(m_pCutArg);
	m_pMinDistArg = new ShellCmdBoolArg("mindist");
	m_pMinDistArg->setDefault("no");
	m_pMinDistArg->setDescription("Flag indicating if instead of intensity information, for each point the minimal distance to an excluded region should be calculated. Only used when the 'cut' parameter is true. ");
	addArgument(m_pMinDistArg);

	setDescription("Creates a regular set of points and excludes the points in the convex hulls of the images data already present in the list. "
		       "Some inversion systems may use this data as extra null-space information.");
}

CmdImagesDataNewRegular::~CmdImagesDataNewRegular()
{
	delete m_pIntensFlagArg;
	delete m_pCutArg;
	delete m_pMinDistArg;
}

bool CmdImagesDataNewRegular::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	std::list<grale::Triangle2D<double> > triangles;
	
	if (m_pCutArg->getValue() == false)
	{
		getAvoidTriangles(triangles);
	
		bool ret;
		grale::ImagesDataExtended *newimg;;

		newimg = new grale::ImagesDataExtended();
		
		ret = newimg->createGridImage(grale::Vector2D<double>(x1arg->getValue(),y1arg->getValue()),
			                   grale::Vector2D<double>(x2arg->getValue(),y2arg->getValue()),
					   numxarg->getValue(),
					   numyarg->getValue(), triangles, m_pTriangulateArg->getValue(),
					   m_pIntensFlagArg->getValue());

		if (!ret)
		{
			setErrorString(std::string("Unable to create new images data: ") + newimg->getErrorString());
			delete newimg;
			return false;
		}

		pGSInst->SetImagesData(newimg);
	}
	else
	{
		std::list<grale::ImagesDataExtended *> imagesData;
		std::list<grale::ImagesDataExtended *>::const_iterator it;
		std::list<grale::ImagesData *> excludeImages;

		pGSInst->GetImagesDataList(imagesData);

		for (it = imagesData.begin() ; it != imagesData.end() ; it++)
		{
			grale::ImagesDataExtended *pImages = *it;

			if (!pImages->hasTriangulation())
			{
				setErrorString("An images data set in the list does not contain a triangulation (is always needed when using cut)");
				return false;
			}

			excludeImages.push_back(pImages);
		}

		grale::ImagesDataExtended *pNewImg = new grale::ImagesDataExtended();

		if (!pNewImg->createGridImage(grale::Vector2D<double>(x1arg->getValue(),y1arg->getValue()),
					      grale::Vector2D<double>(x2arg->getValue(),y2arg->getValue()),
					      numxarg->getValue(), numyarg->getValue(), excludeImages, 
					      m_pIntensFlagArg->getValue(),
				              m_pMinDistArg->getValue()))
		{
			setErrorString(std::string("Unable to create new images data: ") + pNewImg->getErrorString());
			delete pNewImg;
			return false;
		}

		if (!m_pTriangulateArg->getValue())
			pNewImg->clearTriangulation();

		pGSInst->SetImagesData(pNewImg);
	}

	return true;
}

CmdImagesDataAddRegular::CmdImagesDataAddRegular(const std::string &cmdname) : CmdImagesDataRegularBase(cmdname)
{
	setDescription("Creates a regular set of points and excludes the points in the convex hulls of the images data already present in the list. "
		       "Some inversion systems may use this data as extra null-space information.");
}

CmdImagesDataAddRegular::~CmdImagesDataAddRegular()
{
}

bool CmdImagesDataAddRegular::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	std::list<grale::Triangle2D<double> > triangles;
	
	getAvoidTriangles(triangles);

	bool ret;
	grale::ImagesDataExtended *newimg;
	grale::ImagesDataExtended *pOldImg = gsinst->GetImagesData();

	if (pOldImg == 0)
	{
		setErrorString("No images data set currently exists");
		return false;
	}
		
	newimg = new grale::ImagesDataExtended(*pOldImg);

	ret = newimg->addGridImage(grale::Vector2D<double>(x1arg->getValue(),y1arg->getValue()),
			grale::Vector2D<double>(x2arg->getValue(),y2arg->getValue()),
			numxarg->getValue(), numyarg->getValue(), triangles, m_pTriangulateArg->getValue());

	if (!ret)
	{
		setErrorString(std::string("Unable to create new images data: ") + newimg->getErrorString());
		delete newimg;
		return false;
	}

	gsinst->SetImagesData(newimg);
	return true;
}

CmdImagesDataSubtractPosition::CmdImagesDataSubtractPosition(const std::string &cmdname) : ShellCommand(cmdname)
{
	xarg = new ShellCmdRealArg("x");
	xarg->setDescription("X-coordinate value to subtract from each position.");
	xarg->setUnitType(UNIT_ANGULAR);
	addArgument(xarg);
	yarg = new ShellCmdRealArg("y");
	yarg->setDescription("Y-coordinate value to subtract from each position.");
	yarg->setUnitType(UNIT_ANGULAR);
	addArgument(yarg);

	setDescription("Subtract a coordinate vector from each position in the current images data set.");
}

CmdImagesDataSubtractPosition::~CmdImagesDataSubtractPosition()
{
	delete xarg;
	delete yarg;
}

bool CmdImagesDataSubtractPosition::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *curimgdata;

	if ((curimgdata = gsinst->GetImagesData()) == 0)
	{
		setErrorString("No currently loaded images data available");
		return false;
	}

	curimgdata->subtractPosition(grale::Vector2D<double>(xarg->getValue(),yarg->getValue()));

	return true;
}

CmdImagesDataSubtractIntensity::CmdImagesDataSubtractIntensity(const std::string &cmdname) : ShellCommand(cmdname)
{
	varg = new ShellCmdRealArg("value");
	varg->setDescription("Subtract an intensity value.");
	addArgument(varg);

	setDescription("Subtract an intensity value from each point in the current images data set.");
}

CmdImagesDataSubtractIntensity::~CmdImagesDataSubtractIntensity()
{
	delete varg;
}

bool CmdImagesDataSubtractIntensity::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *curimgdata;

	if ((curimgdata = gsinst->GetImagesData()) == 0)
	{
		setErrorString("No currently loaded images data available");
		return false;
	}

	curimgdata->subtractIntensity(varg->getValue());

	return true;
}

CmdImagesDataNonDominatedSet::CmdImagesDataNonDominatedSet(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("algorithm");
	fnamearg->setDescription("The name of the inversion algorithm from which the fitness function should be used");
	addArgument(fnamearg);
	m_pZArg = new ShellCmdRealArg("z_lens");
	m_pZArg->setDescription("The redshift of the lens (should only matter if time-delay data is used).");
	addArgument(m_pZArg);

	setDescription("Projects the images data in the list onto the source plane and calculates the fitness value for all lenses. It then calculates the non-dominated set of all these lenses.");

	fnamearg2 = new ShellCmdStringArg("filename");
	fileargs.push_back(fnamearg2);
}

CmdImagesDataNonDominatedSet::~CmdImagesDataNonDominatedSet()
{
	delete fnamearg;
	delete fnamearg2;
	delete m_pZArg;
}

bool CmdImagesDataNonDominatedSet::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	std::list<grale::ImagesDataExtended *> imagesdata;
	double angularunit = gsinst->GetAngularUnit();
	IOSystem *iosys = gsinst->getIOSystem();

	gsinst->GetImagesDataList(imagesdata);

	if (imagesdata.empty())
	{
		setErrorString("No images data in list.");
		return false;
	}

	std::string basedir, modulename, errStr;
	if (!CmdInvertGridBase::GetModuleDirectory(basedir, errStr))
	{
		setErrorString("Can't get module directory: " + errStr);
		return false;
	}
	CmdInvertGridBase::GetModuleName(fnamearg->getValue(), modulename);

	grale::GALensModule module;
	if (!module.open(basedir, modulename))
	{
		setErrorString(std::string("Couldn't load factory module '") + modulename + std::string("': ") + module.getErrorString());
		return false;
	}

	unique_ptr<LensFitnessObject> fitnessobject(module.createFitnessObject());
	if (fitnessobject == 0)
	{
		setErrorString("The CreateFitnessObject function was not able to create a valid fitness object");
		return false;
	}

	std::list<grale::ImagesDataExtended *> dummy;
	double z_d = m_pZArg->getValue();

	ConfigurationParameters params;
	if (!CmdInvertGridBase::getModuleParameters(gsinst, fnamearg->getValue(), params, errStr))
	{
		setErrorString(errStr);
		return false;
	}

	if (!fitnessobject->init(z_d, imagesdata, dummy, &params))
	{
		setErrorString(std::string("Unable to initialize the fitness object: ") + fitnessobject->getErrorString());
		return false;
	}

	int numcomp = fitnessobject->getNumberOfFitnessComponents();
	std::list<FitnessValues *> fitnessvalues;

	iosys->writeOutputLine("Enter the file names the lenses for which you want to calculate the non dominated set. Type 'end' to stop.");

	std::string prefix("Lens file: ");
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;
		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;
			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = args[0];

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(fileargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing filename: ") + errstr);
				}
				else
				{
					// ok, load lens

					grale::GravitationalLens *l = 0;
					std::string errstr;

					if (!grale::GravitationalLens::load(fnamearg2->getValue(), &l, errstr))
					{
						setErrorString(std::string("Couldn't load specified lens: ") + errstr);
						error = true;
						done = true;
					}
					else
					{
						grale::ImagesBackProjector bp(*l, imagesdata, z_d, false);
						FitnessValues *fv = new FitnessValues(numcomp,fnamearg2->getValue());

						fitnessobject->postInit(imagesdata, dummy, bp.getAngularScale());
						
						fitnessobject->calculateOverallFitness(bp, fv->GetValues());
						delete l;

						fitnessvalues.push_back(fv);
					}
				}
			}
		}
	}	

	if (!error)
	{
		// calculate non dominated set

		std::list<FitnessValues *> populationleft;
		std::list<FitnessValues *> nondominatedset;
		std::list<FitnessValues *>::const_iterator it1,it2;

		for (it1 = fitnessvalues.begin() ; it1 != fitnessvalues.end() ; it1++)
			populationleft.push_back(*it1);

		for (it1 = populationleft.begin() ; it1 != populationleft.end() ; it1++)
		{
			FitnessValues *fv = (*it1);
			bool dominated = false;

			for (it2 = populationleft.begin() ; !dominated && it2 != populationleft.end() ; it2++)
			{
				if (it2 == it1)
					continue;

				FitnessValues *fv2 = (*it2);
				int count1 = 0;
				int count2 = 0;

				for (int i = 0 ; i < numcomp ; i++)
				{
					if (fv2->GetValues()[i] < fv->GetValues()[i])
					{
						count1++;
						count2++;
					}
					else
					{
						if (!(fv->GetValues()[i] < fv2->GetValues()[i]))
							count1++;
					}
				}

				if (count1 == numcomp && count2 > 0)
					dominated = true;
			}

			if (!dominated)
				nondominatedset.push_back(*it1);
		}

		iosys->writeOutputLine("");
		iosys->writeOutputLine("The non dominated set consists of:");
		iosys->writeOutputLine("");

		for (it2 = nondominatedset.begin() ; it2 != nondominatedset.end() ; it2++)
		{
			FitnessValues *fv = (*it2);

			std::string fname = std::string(" + ") + fv->GetFilename();
			iosys->writeOutputLine(fname);
		}

		iosys->writeOutputLine("");
	}

	while (!fitnessvalues.empty())
	{
		delete *(fitnessvalues.begin());
		fitnessvalues.pop_front();
	}

	if (error)
		return false;

	return true;
}

CmdImagesDataNonDominatedSet::FitnessValues::FitnessValues(int n,const std::string f)
{ 
	values = new float[n]; 
	fname = f; 
}

CmdImagesDataNonDominatedSet::FitnessValues::~FitnessValues()
{
	delete [] values;
}

CmdImagesDataNewPoints::CmdImagesDataNewPoints(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pFileArg = new ShellCmdStringArg("filename");
	m_pFileArg->setDescription("Filename to read image points from. Leave to default value to read from standard input.");
	m_pFileArg->setDefault("*");
	addArgument(m_pFileArg);

	setDescription("Create an images data instance containing a single point for each image.");

	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_coordArgs.push_back(m_pXArg);
	m_coordArgs.push_back(m_pYArg);
}

CmdImagesDataNewPoints::~CmdImagesDataNewPoints()
{
	delete m_pFileArg;
	delete m_pXArg;
	delete m_pYArg;
}

bool CmdImagesDataNewPoints::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	std::string fileName = m_pFileArg->getValue();
	FileIOSystem fileIOSys;
	IOSystem *pIOSys;

	if (fileName == std::string("*"))
		pIOSys = pGSInst->getIOSystem();
	else
	{
		if (!fileIOSys.init(fileName))
		{
			setErrorString(std::string("Can't open file '") + fileName + std::string("' to read point data from: ") + fileIOSys.getErrorString());
			return false;
		}
		pIOSys = &fileIOSys;
	}

	pIOSys->writeOutputLine("Enter the point data for each image below. Type 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_coordArgs));
	
	std::list<grale::Vector2D<double> > points;
	std::string prefix("Coordinates: ");
	bool done = false;
	bool error = false;

	while (!done)
	{
		std::string line;
		
		if (!pIOSys->readInputLine(prefix, line))
			done = true;
		else
		{
			if (pIOSys->echo())
				pIOSys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> arguments;

			IOSystem::splitLine(line, arguments);
			
			if (arguments.size() == 1)
			{
				std::string str = arguments[0];

				if (str == std::string("end"))
					done = true;
			}
			
			if (!done)
			{
				std::string errorString;

				if (!ShellCmdArg::parseArgumentList(m_coordArgs, arguments, errorString))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing coordinates: ") + errorString);
				}
				else
				{
					// ok, arguments are valid
					
					double x = m_pXArg->getValue();
					double y = m_pYArg->getValue();

					points.push_back(grale::Vector2D<double>(x, y));
				}
			}
		}
	}

	if (error)
		return false;

	grale::ImagesDataExtended *pNewImgData = new grale::ImagesDataExtended();

	if (!pNewImgData->create(points.size(), false, false))
	{
		setErrorString(std::string("Couldn't create new images data instance: ") + pNewImgData->getErrorString());
		delete pNewImgData;
		return false;
	}
	
	std::list<grale::Vector2D<double> >::const_iterator it;
	int index = 0;

	for (it = points.begin() ; it != points.end() ; it++, index++)
	{
		if (pNewImgData->addPoint(index, *it) < 0)
		{
			setErrorString(std::string("Error adding point to images data: ") +  pNewImgData->getErrorString());
			delete pNewImgData;
			return false;
		}
	}

	pGSInst->SetImagesData(pNewImgData);
	return true;
}

CmdImagesDataNew::CmdImagesDataNew(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pNumImagesArg = new ShellCmdIntArg("numimages");
	m_pNumImagesArg->setDescription("Number of images which will be contained in this data set.");
	m_pNumImagesArg->setMin(1);
	addArgument(m_pNumImagesArg);

	m_pIntensFlagArg = new ShellCmdBoolArg("intensities");
	m_pIntensFlagArg->setDefault("no");
	m_pIntensFlagArg->setDescription("Flag indicating if intensity information will be present in the image points.");
	addArgument(m_pIntensFlagArg);

	m_pShearFlagArg = new ShellCmdBoolArg("shearinfo");
	m_pShearFlagArg->setDefault("no");
	m_pShearFlagArg->setDescription("Flag indicating if shear information will be present in the image points.");
	addArgument(m_pShearFlagArg);

	setDescription("Creates an empty images data instance which can contain information about a specific number of images.");
}

CmdImagesDataNew::~CmdImagesDataNew()
{
	delete m_pNumImagesArg;
	delete m_pIntensFlagArg;
	delete m_pShearFlagArg;
}

bool CmdImagesDataNew::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *pNewImgData = new grale::ImagesDataExtended();

	if (!pNewImgData->create(m_pNumImagesArg->getValue(), m_pIntensFlagArg->getValue(), m_pShearFlagArg->getValue()))
	{
		setErrorString(std::string("Unable to initialize the new images data instance: ") + pNewImgData->getErrorString());
		delete pNewImgData;
		return false;
	}

	pGSInst->SetImagesData(pNewImgData);	
	return true;
}

CmdImagesDataAddPoints::CmdImagesDataAddPoints(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pImageNumberArg = new ShellCmdIntArg("imageindex");
	m_pImageNumberArg->setMin(0);
	m_pImageNumberArg->setDescription("Index of the image for which points will be added.");
	addArgument(m_pImageNumberArg);

	m_pFileArg = new ShellCmdStringArg("filename");
	m_pFileArg->setDescription("Filename to read points data from. Leave to default value to read from standard input.");
	m_pFileArg->setDefault("*");
	addArgument(m_pFileArg);

	setDescription("Add points to the current images data instance.");

	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pIntensArg = new ShellCmdRealArg("intensity");
	m_pShear1 = new ShellCmdRealArg("shearcomponent1");
	m_pShear2 = new ShellCmdRealArg("shearcomponent2");
}

CmdImagesDataAddPoints::~CmdImagesDataAddPoints()
{
	delete m_pImageNumberArg;
	delete m_pFileArg;
	delete m_pXArg;
	delete m_pYArg;
	delete m_pIntensArg;
	delete m_pShear1;
	delete m_pShear2;
}

bool CmdImagesDataAddPoints::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *pImgDat;

	if ((pImgDat = pGSInst->GetImagesData()) == 0)
	{
		setErrorString("Currently no images data instance is available");
		return false;
	}
	if (m_pImageNumberArg->getValue() >= pImgDat->getNumberOfImages())
	{
		setErrorString("The specified image index is too high");
		return false;
	}

	bool gotIntens = pImgDat->hasIntensities();
	bool gotShear = pImgDat->hasShearInfo();

	m_pointInfoArgs.clear();
	m_pointInfoArgs.push_back(m_pXArg);
	m_pointInfoArgs.push_back(m_pYArg);

	if (gotIntens)
		m_pointInfoArgs.push_back(m_pIntensArg);
	if (gotShear)
	{
		m_pointInfoArgs.push_back(m_pShear1);
		m_pointInfoArgs.push_back(m_pShear2);
	}

	std::string fileName = m_pFileArg->getValue();
	FileIOSystem fileIOSys;
	IOSystem *pIOSys;

	if (fileName == std::string("*"))
		pIOSys = pGSInst->getIOSystem();
	else
	{
		if (!fileIOSys.init(fileName))
		{
			setErrorString(std::string("Can't open file '") + fileName + std::string("' to read point data from: ") + fileIOSys.getErrorString());
			return false;
		}
		pIOSys = &fileIOSys;
	}

	pIOSys->writeOutputLine("Enter the data for each point below. Type 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_pointInfoArgs));
	
	std::list<grale::Vector2D<double> > points;
	std::list<double> intensities;
	std::list<double> shear1;
	std::list<double> shear2;
	std::string prefix("Point info: ");
	bool done = false;
	bool error = false;

	while (!done)
	{
		std::string line;
		
		if (!pIOSys->readInputLine(prefix, line))
			done = true;
		else
		{
			if (pIOSys->echo())
				pIOSys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> arguments;

			IOSystem::splitLine(line, arguments);
			
			if (arguments.size() == 1)
			{
				std::string str = arguments[0];

				if (str == std::string("end"))
					done = true;
			}
			
			if (!done)
			{
				std::string errorString;

				if (!ShellCmdArg::parseArgumentList(m_pointInfoArgs, arguments, errorString))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing coordinates: ") + errorString);
				}
				else
				{
					// ok, arguments are valid
					
					double x = m_pXArg->getValue();
					double y = m_pYArg->getValue();

					points.push_back(grale::Vector2D<double>(x, y));

					if (gotIntens)
						intensities.push_back(m_pIntensArg->getValue());
					if (gotShear)
					{
						shear1.push_back(m_pShear1->getValue());
						shear2.push_back(m_pShear2->getValue());
					}
				}
			}
		}
	}

	if (error)
		return false;

	std::list<grale::Vector2D<double> >::const_iterator pointIt = points.begin();
	std::list<double>::const_iterator intensIt = intensities.begin();
	std::list<double>::const_iterator shear1It = shear1.begin();
	std::list<double>::const_iterator shear2It = shear2.begin();

	while (pointIt != points.end())
	{
		if (!gotIntens)
		{
			if (!gotShear)
				error = (pImgDat->addPoint(m_pImageNumberArg->getValue(), *pointIt) < 0);
			else
				error = (pImgDat->addPoint(m_pImageNumberArg->getValue(), *pointIt, *shear1It, *shear2It) < 0);
		}
		else
		{
			if (!gotShear)
				error = (pImgDat->addPoint(m_pImageNumberArg->getValue(), *pointIt, *intensIt) < 0);
			else
				error = (pImgDat->addPoint(m_pImageNumberArg->getValue(), *pointIt, *intensIt, *shear1It, *shear2It) < 0);
		}

		if (error)
		{
			setErrorString(std::string("Error adding point: ") + pImgDat->getErrorString());
			return false;
		}

		pointIt++;
		intensIt++;
		shear1It++;
		shear2It++;
	}

	return true;
}

CmdImagesDataAddShearInfo::CmdImagesDataAddShearInfo(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pReducedArg = new ShellCmdBoolArg("reduced");
	m_pReducedArg->setDefault("no");
	m_pReducedArg->setDescription("Flag indicating if the reduced shear should be calculated instead of the normal shear.");
	addArgument(m_pReducedArg);

	m_pSigmaX = new ShellCmdRealArg("sigmax");
	m_pSigmaX->setDefault("0");
	m_pSigmaX->setDescription("Standard deviation of the gaussian noise that should be added to the x-value of the shear.");
	addArgument(m_pSigmaX);

	m_pSigmaY = new ShellCmdRealArg("sigmay");
	m_pSigmaY->setDefault("0");
	m_pSigmaY->setDescription("Standard deviation of the gaussian noise that should be added to the y-value of the shear.");
	addArgument(m_pSigmaY);

	setDescription("For all the image points in the current images data instance, this command "
		       "calculates and stores the shear info (regular or reduced) using the current lens "
			   "and current source plane. Gaussian noise of specified standard deviation can be "
			   "added to the components.");
}

CmdImagesDataAddShearInfo::~CmdImagesDataAddShearInfo()
{
	delete m_pReducedArg;
	delete m_pSigmaX;
	delete m_pSigmaY;
}

bool CmdImagesDataAddShearInfo::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagesDataExtended *pImgDat;
	grale::GravitationalLens *pLens;
	grale::SourcePlane *pSrcPlane;

	if ((pImgDat = pGSInst->GetImagesData()) == 0)
	{
		setErrorString("Currently no images data instance is available");
		return false;
	}

	if ((pLens = pGSInst->GetLens()) == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}

	if ((pSrcPlane = pGSInst->GetSourcePlane()) == 0)
	{
		setErrorString("No source plane has been created yet");
		return false;
	}

	grale::ImagesDataExtended *pNewImg = new grale::ImagesDataExtended();
	int numImages = pImgDat->getNumberOfImages();
	double D_s = pSrcPlane->getD_s();
	double D_ds = pSrcPlane->getD_ds();
	bool error;

	if (!pNewImg->create(numImages, false,  true))
	{
		setErrorString(std::string("Unable to create a new images data instance: ") + pNewImg->getErrorString());
		delete pNewImg;
		return false;
	}

	grale::GaussianDistribution xGaussDist(0, m_pSigmaX->getValue(), pGSInst->GetRandomNumberGenerator());
	grale::GaussianDistribution yGaussDist(0, m_pSigmaY->getValue(), pGSInst->GetRandomNumberGenerator());

	for (int i = 0 ; i < numImages ; i++)
	{
		int numPoints = pImgDat->getNumberOfImagePoints(i);

		for (int p = 0 ; p < numPoints ; p++)
		{
			grale::Vector2D<double> pos = pImgDat->getImagePointPosition(i, p);
			double axx, ayy, axy;
			double shearAngle, shearSize;

			if (!pLens->getAlphaVectorDerivatives(pos, axx, ayy, axy))
			{
				delete pNewImg;
				setErrorString(std::string("Unable to calculate the deflection angle derivatives: ") + pLens->getErrorString());
				return false;
			}

			double gamma1 = 0.5*(D_ds/D_s)*(axx-ayy);
			double gamma2 = (D_ds/D_s)*axy;

			if (m_pReducedArg->getValue())
			{
				double kappa = 0.5*(D_ds/D_s)*(axx+ayy);

				gamma1 /= (1.0-kappa);
				gamma2 /= (1.0-kappa);
			}

			gamma1 += xGaussDist.pickNumber();
			gamma2 += yGaussDist.pickNumber();
			
			if (pNewImg->addPoint(i, pos, gamma1, gamma2) < 0)
			{
				setErrorString(std::string("Unable to add point to new images data: ") + pNewImg->getErrorString());
				delete pNewImg;
				return false;
			}
		}
	}

	pGSInst->SetImagesData(pNewImg);

	return true;
}

CmdImagesDataNewGroups::CmdImagesDataNewGroups(const string &cmdName) : ShellCommand(cmdName)
{
	m_pFileArg = new ShellCmdStringArg("filename");
	m_pFileArg->setDefault("*");
	m_pFileArg->setDescription("Filename to read points data from. Leave to default value to read from standard input.");
	addArgument(m_pFileArg);

	setDescription("Create image data set with different images where point groups can be identified");
	
	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pGroupNameArg = new ShellCmdStringArg("groupname");
	m_pGroupNameArg->setDefault("*");
	
	m_pointInfoArgs.push_back(m_pXArg);
	m_pointInfoArgs.push_back(m_pYArg);
	m_pointInfoArgs.push_back(m_pGroupNameArg);
}

CmdImagesDataNewGroups::~CmdImagesDataNewGroups()
{
	delete m_pXArg;
	delete m_pYArg;
	delete m_pGroupNameArg;
}

bool CmdImagesDataNewGroups::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	std::string fileName = m_pFileArg->getValue();
	FileIOSystem fileIOSys;
	IOSystem *pIOSys;

	if (fileName == std::string("*"))
		pIOSys = pGSInst->getIOSystem();
	else
	{
		if (!fileIOSys.init(fileName))
		{
			setErrorString(std::string("Can't open file '") + fileName + std::string("' to read point data from: ") + fileIOSys.getErrorString());
			return false;
		}
		pIOSys = &fileIOSys;
	}

	ImagesDataExtended *pImgDat = new ImagesDataExtended();
	if (!pImgDat->create(1, false, false))
	{
		setErrorString("Unable to create start images data instance: " + pImgDat->getErrorString());
		delete pImgDat;
		return false;
	}

	pIOSys->writeOutputLine("Enter the data for each point below. Type 'new' to start new image, 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_pointInfoArgs));
	
	map<string, vector< pair<int, int> > > groupInfo;

	std::string prefix("Point info: ");
	bool done = false;
	bool error = false;

	int curImg = 0;

	while (!done)
	{
		std::string line;
		bool isNew = false;
		
		if (!pIOSys->readInputLine(prefix, line))
			done = true;
		else
		{
			if (pIOSys->echo())
				pIOSys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> arguments;

			IOSystem::splitLine(line, arguments);
			
			if (arguments.size() == 1)
			{
				std::string str = arguments[0];

				if (str == "end")
					done = true;
				else if (str == "new")
					isNew = true;
			}
			
			if (!done)
			{
				if (isNew)
				{
					curImg = pImgDat->addImage();
					if (curImg < 0)
					{
						error = true;
						done = true;
						setErrorString("Unable to add image: " + pImgDat->getErrorString());
					}
				}
				else
				{
					std::string errorString;

					if (!ShellCmdArg::parseArgumentList(m_pointInfoArgs, arguments, errorString))
					{
						error = true;
						done = true;
						setErrorString(std::string("Error parsing coordinates: ") + errorString);
					}
					else
					{
						// ok, arguments are valid
						
						double x = m_pXArg->getValue();
						double y = m_pYArg->getValue();
						string group = m_pGroupNameArg->getValue();

						int pointNr = pImgDat->addPoint(curImg, Vector2D<double>(x, y));

						if (group != "*")
						{
							auto it = groupInfo.find(group);
							if (it != groupInfo.end())
							{
								it->second.push_back(pair<int,int>(curImg, pointNr));
							}
							else
							{
								groupInfo[group] = { { curImg, pointNr } };
							}
						}
					}
				}
			}
		}
	}

	if (error)
	{
		delete pImgDat;
		return false;
	}

	// Validate image: no empty images!
	for (int i = 0 ; i < pImgDat->getNumberOfImages() ; i++)
	{
		if (pImgDat->getNumberOfImagePoints(i) <= 0)
		{
			setErrorString("A specified image does not contain any points");
			delete pImgDat;
			return false;
		}
	}

	// Create groups
	for (auto it = groupInfo.begin() ; it != groupInfo.end() ; ++it)
	{
		int groupIdx = pImgDat->addGroup();
		if (groupIdx < 0)
		{
			setErrorString("Unable to add group to images data: " + pImgDat->getErrorString());
			delete pImgDat;
			return false;
		}

		for (auto imgAndPoint : it->second)
		{
			if (!pImgDat->addGroupPoint(groupIdx, imgAndPoint.first, imgAndPoint.second))
			{
				setErrorString("Unable to add group point info to images data: " + pImgDat->getErrorString());
				delete pImgDat;
				return false;
			}
		}
	}

	// Store images data instance
	pGSInst->SetImagesData(pImgDat);

	return true;
}
