/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include <mpi.h>
#include "cmdinvertmpi.h"
#include "mainmpi.h"
#include "graleshellinstance.h"
#include "graleshellmpigeneticalgorithm.h"
#include "graleshellgaparams.h"
#include <shellp/iosystem.h>
#include <mogal/gamodule.h>
#include <grale/lensinversiongenomebase.h>
#include <grale/gridlensinversiongafactoryparams.h>
#include <grale/gridlensinversiongafactoryparamsnew.h>
#include <grale/arealensinversiongafactoryparams.h>
#include <grale/treelensinversiongafactoryparams.h>
#include <grale/imagesdata.h>
#include <grale/constants.h>
#include <grale/grid.h>
#include <grale/galensmodule.h>
#include <grale/gravitationallens.h>
#include <serut/dummyserializer.h>
#include <serut/memoryserializer.h>
#include <stdlib.h>
#include <stdio.h>

#include <iostream>

#include "debugnew.h"

using namespace shellp;

CmdInvertGridMPI::CmdInvertGridMPI(const std::string &cmdname) : CmdInvertGridBase(cmdname)
{
	AddDefaultArguments();
	setDescription("Use a genetic algorithm which is executed in a distributed way using MPI to invert a lens. The images in the images data list are used and Plummer basis functions are placed corresponding to the current grid layout.");
}

CmdInvertGridMPI::~CmdInvertGridMPI()
{
}

bool CmdInvertGridMPI::execute()
{
	std::string basedir;
	std::string modulename;

	if (!GetModuleDirectory(basedir))
		return false;
	GetModuleName(modulename);
	
	mogal::GAFactoryParams *params = GetFactoryParameters();
	if (params == 0)
		return false;

	mogal::MPIGeneticAlgorithm *ga = GetMPIGeneticAlgorithm();
	if (ga == 0)
	{
		delete params;
		return false;
	}

	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	mogal::GAFactory *pFactory;
	mogal::GAModule module;
	
	if (!module.open(basedir, modulename))
	{
		setErrorString("Couldn't open module: " + module.getErrorString());
		delete params;
		delete ga;
		return false;
	}

	pFactory = module.createFactoryInstance();
	if (pFactory == 0)
	{
		setErrorString("Couldn't create GA factory from module: " + module.getErrorString());
		delete params;
		delete ga;
		return false;
	}

	if (!pFactory->init(params))
	{
		setErrorString("Couldn't init GA factory: " + pFactory->getErrorString());
		delete params;
		delete ga;
		delete pFactory;
		return false;
	}

	serut::DummySerializer dumSer;

	dumSer.writeString(basedir);
	dumSer.writeString(modulename);
	if (!params->write(dumSer))
	{
		setErrorString("Unable to count the number of bytes to serialize: " + params->getErrorString());
		delete params;
		delete ga;
		delete pFactory;
		return false;
	}

	int paramSize = dumSer.getBytesWritten();
	
	std::vector<uint8_t> buffer(paramSize);
	serut::MemorySerializer mSer(0, 0, &(buffer[0]), buffer.size());

	mSer.writeString(basedir);
	mSer.writeString(modulename);
	if (!params->write(mSer))
	{
		setErrorString("Unable to serialize the genetic algorithm factory parameters to memory: " + params->getErrorString());
		delete params;
		delete ga;
		delete pFactory;
		return false;
	}

	// Send the command and transmit the parameters to the other ranks
	int command = GRALESHELL_MPICOMMAND_GA;

	MPI_Bcast(&command, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&paramSize, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&(buffer[0]), paramSize, MPI_BYTE, 0, MPI_COMM_WORLD);

	if (!ga->runMPI(*pFactory, GetPopulationSize(), gsinst->GetGAParams()))
	{
		setErrorString(std::string("Error running genetic algorithm: ") + ga->getErrorString());
		delete params;
		delete ga;
		delete pFactory;
		return false;
	}
	delete params;

	if (!ProcessSolution(*ga))
	{
		delete ga;
		delete pFactory;
		return false;
	}
	delete ga;
	delete pFactory;
	
	return true;
}

mogal::MPIGeneticAlgorithm *CmdInvertGridMPI::GetMPIGeneticAlgorithm()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	GraleShellGAParams *params = gsinst->GetGAParams();

	if (params->SaveIntermediate())
		return new GraleShellMPIGeneticAlgorithm(params->GetIntermediatePrefix());
	return new GraleShellMPIGeneticAlgorithm();
}
