/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "cmdlenses.h"
#include "graleshellinstance.h"
#include <shellp/fileiosystem.h>
#include <grale/pointmasslens.h>
#include <grale/plummerlens.h>
#include <grale/multipleplummerlens.h>
#include <grale/gausslens.h>
#include <grale/nsielens.h>
#include <grale/nsislens.h>
#include <grale/sielens.h>
#include <grale/sislens.h>
#include <grale/squarelens.h>
#include <grale/multiplesquarelens.h>
#include <grale/multiplegausslens.h>
#include <grale/masssheetlens.h>
#include <grale/massdisklens.h>
#include <grale/compositelens.h>
#include <grale/profilelens.h>
#include <grale/polynomialmassprofilelens.h>
#include <grale/multiplewendlandlens.h>
#include <grale/imagesdataextended.h>
#include <grale/real2dlensfunction.h>
#include <grale/real2dfunctiongnuplotplotter.h>
#include <grale/real2dfunctionfitsplotter.h>
#include <grale/circularintegrator.h>
#include <grale/constants.h>
#include <grale/fitsimage.h>
#include <grale/lensplane.h>
#include <grale/imageplane.h>
#include <grale/sourceplane.h>
#include <grale/testlens.h>
#include <grale/nfwlens.h>
#include <grale/ellipticnfwlens.h>
#include <grale/sersiclens.h>
#include <grale/ellipticsersiclens.h>
#include <grale/rectangularintegrator.h>
#include <stdio.h>
#include <assert.h>
#include <vector>
#include <limits>

#include "debugnew.h"

using namespace shellp;

// Base class: all lenses need the D_d parameter

CmdLens::CmdLens(const std::string &cmdname) : ShellCommand(cmdname)
{
	distarg = new ShellCmdRealArg("D_d");
	distarg->setDescription("Angular diameter distance between lens and observer.");
	distarg->setUnitType(UNIT_DISTANCE);
	addArgument(distarg);
}

CmdLens::~CmdLens()
{
	delete distarg;
}

// Point mass lens

CmdLensPointmass::CmdLensPointmass(const std::string &cmdname) : CmdLens(cmdname)
{
	massarg = new ShellCmdRealArg("mass");
	massarg->setDescription("Mass of the lens.");
	massarg->setUnitType(UNIT_MASS);
	addArgument(massarg);	
	setDescription("Create a point mass lens.");
}

CmdLensPointmass::~CmdLensPointmass()
{
	delete massarg;
}

bool CmdLensPointmass::execute()
{
	double D_d = GetD_d();
	double mass = massarg->getValue();
	
	grale::PointmassLens *l = new grale::PointmassLens();
	grale::PointmassLensParams p(mass);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the point mass lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);
	
	return true;
}

// Plummer lens

CmdLensPlummer::CmdLensPlummer(const std::string &cmdname) : CmdLens(cmdname)
{
	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	massarg->setDescription("Total mass of the Plummer lens.");
	addArgument(massarg);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	widtharg->setDescription("Width parameter of the Plummer distribution.");
	addArgument(widtharg);
	setDescription("Create a Plummer lens.");
}

CmdLensPlummer::~CmdLensPlummer()
{
	delete massarg;
	delete widtharg;
}

bool CmdLensPlummer::execute()
{
	double D_d = GetD_d();
	double mass = massarg->getValue();
	double width = widtharg->getValue();
	
	grale::PlummerLens *l = new grale::PlummerLens();
	grale::PlummerLensParams p(mass,width);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the Plummer lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Gaussian lens

CmdLensGauss::CmdLensGauss(const std::string &cmdname) : CmdLens(cmdname)
{
	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	massarg->setDescription("Total mass of the Gaussian lens.");
	addArgument(massarg);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	widtharg->setDescription("Width parameter of the Gaussian distribution.");
	addArgument(widtharg);
	setDescription("Create a Gaussian lens.");
}

CmdLensGauss::~CmdLensGauss()
{
	delete massarg;
	delete widtharg;
}

bool CmdLensGauss::execute()
{
	double D_d = GetD_d();
	double mass = massarg->getValue();
	double width = widtharg->getValue();
	
	grale::GaussLens *l = new grale::GaussLens();
	grale::GaussLensParams p(mass,width);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the Gaussian lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Multiple plummer lens

CmdLensMultiplePlummers::CmdLensMultiplePlummers(const std::string &cmdname) : CmdLens(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename","*");
	fnamearg->setDescription("File name to read Plummer info from or '*' to read from current input.");
	addArgument(fnamearg);
	setDescription("Create a lens consisting of multiple Plummer distributions.");

	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	xposarg = new ShellCmdRealArg("x");
	xposarg->setUnitType(UNIT_ANGULAR);
	yposarg = new ShellCmdRealArg("y");
	yposarg->setUnitType(UNIT_ANGULAR);
	plumargs.push_back(massarg);
	plumargs.push_back(widtharg);
	plumargs.push_back(xposarg);
	plumargs.push_back(yposarg);
}

CmdLensMultiplePlummers::~CmdLensMultiplePlummers()
{
	delete fnamearg;
	delete massarg;
	delete widtharg;
	delete xposarg;
	delete yposarg;
}

bool CmdLensMultiplePlummers::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	double D_d = GetD_d();
	std::string fname = fnamearg->getValue();
	IOSystem *iosys;
	FileIOSystem fiosys;
	std::string prefix("Plummer info: ");
	
	if (fname == std::string("*")) // use current input
		iosys = gsinst->getIOSystem();
	else
	{
		if (!fiosys.init(fname))
		{
			setErrorString(std::string("Can't open file '") + fname + std::string("' to read Plummer data from: ") + fiosys.getErrorString());
			return false;
		}
		iosys = &fiosys;
	}
	
	iosys->writeOutputLine("Enter individual Plummer parameters below. Type 'end' to end input.");
	iosys->writeOutputLine("Argument order is: ");
	iosys->writeOutputLine(ShellCmdArg::getArgumentNames(plumargs));
	
	grale::MultiplePlummerLensParams p;
	
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;

		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;
			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = args[0];

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(plumargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing individual Plummer parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid
					
					double mass = massarg->getValue();
					double width = widtharg->getValue();
					double x = xposarg->getValue();
					double y = yposarg->getValue();

					p.addLensInfo(grale::PlummerLensInfo(mass,width,grale::Vector2D<double>(x,y)));
				}
			}
		}
	}

	if (error)
		return false;

	// parameters are filled in, create the lens
	
	grale::MultiplePlummerLens *l = new grale::MultiplePlummerLens();

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the multiple Plummer lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	gsinst->SetLens(l);

	return true;
}

// Non-singular isothermal ellipse

CmdLensNSIE::CmdLensNSIE(const std::string &cmdname) : CmdLens(cmdname)
{
	Varg = new ShellCmdRealArg("sigma");
	Varg->setDescription("Dispersion velocity (in km/s).");
	addArgument(Varg);
	Farg = new ShellCmdRealArg("ellipticity");
	Farg->setMin(0.0);
	Farg->setMax(1.0);
	Farg->setDescription("Parameter indicating the ellipticity. Must lie between 0 and 1.");
	addArgument(Farg);
	corearg = new ShellCmdRealArg("coreradius");
	corearg->setUnitType(UNIT_ANGULAR);
	corearg->setDescription("Core radius of the NSIE.");
	addArgument(corearg);
	setDescription("Create a non-singular isothermal ellipse (NSIE) lens.");
}

CmdLensNSIE::~CmdLensNSIE()
{
	delete Varg;
	delete Farg;
	delete corearg;
}

bool CmdLensNSIE::execute()
{
	double D_d = GetD_d();
	double V = Varg->getValue() * 1000.0; // was given in kilometers per second
	double F = Farg->getValue();
	double corerad = corearg->getValue();
	
	grale::NSIELens *l = new grale::NSIELens();
	grale::NSIELensParams p(V,F,corerad);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the NSIE lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Non-singular isothermal sphere

CmdLensNSIS::CmdLensNSIS(const std::string &cmdname) : CmdLens(cmdname)
{
	Varg = new ShellCmdRealArg("sigma");
	Varg->setDescription("Dispersion velocity (in km/s).");
	addArgument(Varg);
	corearg = new ShellCmdRealArg("coreradius");
	corearg->setUnitType(UNIT_ANGULAR);
	corearg->setDescription("Core radius of the NSIS.");
	addArgument(corearg);
	setDescription("Create a non-singular isothermal sphere (NSIS) lens.");
}

CmdLensNSIS::~CmdLensNSIS()
{
	delete corearg;
	delete Varg;
}

bool CmdLensNSIS::execute()
{
	double D_d = GetD_d();
	double V = Varg->getValue() * 1000.0; // was given in kilometers per second
	double corerad = corearg->getValue();
	
	grale::NSISLens *l = new grale::NSISLens();
	grale::NSISLensParams p(V,corerad);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the NSIS lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Singular isothermal ellipse

CmdLensSIE::CmdLensSIE(const std::string &cmdname) : CmdLens(cmdname)
{
	Varg = new ShellCmdRealArg("sigma");
	Varg->setDescription("Dispersion velocity (in km/s).");
	addArgument(Varg);
	Farg = new ShellCmdRealArg("ellipticity");
	Farg->setMin(0.0);
	Farg->setMax(1.0);
	Farg->setDescription("Parameter indicating the ellipticity. Must lie between 0 and 1.");
	addArgument(Farg);
	setDescription("Create a singular isothermal ellipse (SIE) lens.");
}

CmdLensSIE::~CmdLensSIE()
{
	delete Varg;
	delete Farg;
}

bool CmdLensSIE::execute()
{
	double D_d = GetD_d();
	double V = Varg->getValue() * 1000.0; // was given in kilometers per second
	double F = Farg->getValue();
	
	grale::SIELens *l = new grale::SIELens();
	grale::SIELensParams p(V,F);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the SIE lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Singular isothermal sphere

CmdLensSIS::CmdLensSIS(const std::string &cmdname)  : CmdLens(cmdname)
{
	Varg = new ShellCmdRealArg("sigma");
	Varg->setDescription("Dispersion velocity (in km/s).");
	addArgument(Varg);
	setDescription("Create a singular isothermal sphere (SIS) lens.");
}

CmdLensSIS::~CmdLensSIS()
{
	delete Varg;
}

bool CmdLensSIS::execute()
{
	double D_d = GetD_d();
	double V = Varg->getValue() * 1000.0; // was given in kilometers per second
	
	grale::SISLens *l = new grale::SISLens();
	grale::SISLensParams p(V);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the SIS lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

CmdLensMassSheetDegeneracy::CmdLensMassSheetDegeneracy(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pDsArg = new ShellCmdRealArg("D_s");
	m_pDsArg->setDescription("Angular diameter distance from observer to source.");
	m_pDsArg->setMin(0);
	m_pDsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDsArg);
	
	m_pDdsArg = new ShellCmdRealArg("D_ds");
	m_pDdsArg->setDescription("Angular diameter distance from lens to source.");
	m_pDdsArg->setMin(0);
	m_pDdsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDdsArg);

	m_pLambdaArg = new ShellCmdRealArg("lambda");
	m_pLambdaArg->setDescription("Parameter with which the current mass will be scaled and which "
			             "describes the density (in units of the critical density) of the "
				     "mass sheet which will be added.");
	addArgument(m_pLambdaArg);

	setDescription("Takes the current lens and creates a degenerate version by scaling it and "
		       "adding a sheet of mass with density (1-lambda) times the critical density "
		       "(for D_s and D_ds).");
}

CmdLensMassSheetDegeneracy::~CmdLensMassSheetDegeneracy()
{
	delete m_pDsArg;
	delete m_pDdsArg;
	delete m_pLambdaArg;
}

bool CmdLensMassSheetDegeneracy::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *pLens = pInstance->GetLens();
	
	if (pLens == 0)
	{
		setErrorString("No lens has been created yet.");
		return false;
	}

	grale::MassSheetLensParams sheetParams(pLens->getLensDistance(), m_pDsArg->getValue(), m_pDdsArg->getValue());
	grale::MassSheetLens sheetLens;

	if (!sheetLens.init(pLens->getLensDistance(), &sheetParams))
	{
		setErrorString(std::string("Couldn't initialize the constant mass sheet lens: ") + sheetLens.getErrorString());
		return false;
	}

	grale::CompositeLensParams compositeLensParams;

	if (!compositeLensParams.addLens(m_pLambdaArg->getValue(), grale::Vector2D<double>(0,0), 0, *pLens))
	{
		setErrorString(compositeLensParams.getErrorString());
		return false;
	}

	if (!compositeLensParams.addLens((1.0-m_pLambdaArg->getValue()), grale::Vector2D<double>(0,0), 0, sheetLens))
	{
		setErrorString(compositeLensParams.getErrorString());
		return false;
	}
	
	grale::CompositeLens *pNewLens = new grale::CompositeLens();
	
	if (!pNewLens->init(pLens->getLensDistance(), &compositeLensParams))
	{
		setErrorString(std::string("Couldn't initialize composite lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}

	pInstance->SetLens(pNewLens);
	
	return true;
}

CmdLensMassDiskDegeneracy::CmdLensMassDiskDegeneracy(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pDsArg = new ShellCmdRealArg("D_s");
	m_pDsArg->setDescription("Angular diameter distance from observer to source.");
	m_pDsArg->setMin(0);
	m_pDsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDsArg);
	
	m_pDdsArg = new ShellCmdRealArg("D_ds");
	m_pDdsArg->setDescription("Angular diameter distance from lens to source.");
	m_pDdsArg->setMin(0);
	m_pDdsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDdsArg);

	m_pLambdaArg = new ShellCmdRealArg("lambda");
	m_pLambdaArg->setDescription("Parameter with which the current mass will be scaled and which "
			             "describes the density (in units of the critical density) of the "
				     "mass disk which will be added.");
	addArgument(m_pLambdaArg);

	m_pCenterXArg = new ShellCmdRealArg("x");
	m_pCenterXArg->setDefault("0");
	m_pCenterXArg->setUnitType(UNIT_ANGULAR);
	m_pCenterXArg->setDescription("X-coordinate of the center of the disk.");
	addArgument(m_pCenterXArg);
	
	m_pCenterYArg = new ShellCmdRealArg("y");
	m_pCenterYArg->setDefault("0");
	m_pCenterYArg->setUnitType(UNIT_ANGULAR);
	m_pCenterYArg->setDescription("Y-coordinate of the center of the disk.");
	addArgument(m_pCenterYArg);
	
	m_pAngularRadiusArg = new ShellCmdRealArg("radius");
	m_pAngularRadiusArg->setMin(0);
	m_pAngularRadiusArg->setUnitType(UNIT_ANGULAR);
	m_pAngularRadiusArg->setDescription("Radius of the constant density disk. The disk must "
			                    "encompass all images to create a degenerate version of "
					    "the lens.");
	addArgument(m_pAngularRadiusArg);
	
	setDescription("Takes the current lens and creates a degenerate version by scaling it and "
		       "adding a disk of mass with density (1-lambda) times the critical density "
		       "(for D_s and D_ds). The disk must encompass all the images to create a degenerate "
		       "version of the lens.");
}

CmdLensMassDiskDegeneracy::~CmdLensMassDiskDegeneracy()
{
	delete m_pDsArg;
	delete m_pDdsArg;
	delete m_pLambdaArg;
	delete m_pCenterXArg;
	delete m_pCenterYArg;
	delete m_pAngularRadiusArg;
}

bool CmdLensMassDiskDegeneracy::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *pLens = pInstance->GetLens();
	
	if (pLens == 0)
	{
		setErrorString("No lens has been created yet.");
		return false;
	}

	grale::MassDiskLensParams diskParams(pLens->getLensDistance(), m_pDsArg->getValue(), m_pDdsArg->getValue(), m_pAngularRadiusArg->getValue());
	grale::MassDiskLens diskLens;

	if (!diskLens.init(pLens->getLensDistance(), &diskParams))
	{
		setErrorString(std::string("Couldn't initialize the constant mass disk lens: ") + diskLens.getErrorString());
		return false;
	}

	grale::CompositeLensParams compositeLensParams;

	if (!compositeLensParams.addLens(m_pLambdaArg->getValue(), grale::Vector2D<double>(0,0), 0, *pLens))
	{
		setErrorString(compositeLensParams.getErrorString());
		return false;
	}

	if (!compositeLensParams.addLens((1.0-m_pLambdaArg->getValue()), grale::Vector2D<double>(m_pCenterXArg->getValue(), m_pCenterYArg->getValue()), 0, diskLens))
	{
		setErrorString(compositeLensParams.getErrorString());
		return false;
	}
	
	grale::CompositeLens *pNewLens = new grale::CompositeLens();
	
	if (!pNewLens->init(pLens->getLensDistance(), &compositeLensParams))
	{
		setErrorString(std::string("Couldn't initialize composite lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}

	pInstance->SetLens(pNewLens);
	
	return true;
}

CmdLensMassSheet::CmdLensMassSheet(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDsArg = new ShellCmdRealArg("D_s");
	m_pDsArg->setDescription("Angular diameter distance from observer to source.");
	m_pDsArg->setMin(0);
	m_pDsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDsArg);
	
	m_pDdsArg = new ShellCmdRealArg("D_ds");
	m_pDdsArg->setDescription("Angular diameter distance from lens to source.");
	m_pDdsArg->setMin(0);
	m_pDdsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDdsArg);

	setDescription("Creates a mass-sheet of which the density corresponds to the critical density for the "
		       "specified angular diameter distances.");
}

CmdLensMassSheet::~CmdLensMassSheet()
{
	delete m_pDsArg;
	delete m_pDdsArg;
}

bool CmdLensMassSheet::execute()
{
	double D_d = GetD_d();
	double D_ds = m_pDdsArg->getValue();
	double D_s = m_pDsArg->getValue();
	
	grale::MassSheetLens *pLens = new grale::MassSheetLens();
	grale::MassSheetLensParams params(D_d, D_s, D_ds);

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the mass-sheet lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);
	return true;
}

CmdLensMassDisk::CmdLensMassDisk(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDsArg = new ShellCmdRealArg("D_s");
	m_pDsArg->setDescription("Angular diameter distance from observer to source.");
	m_pDsArg->setMin(0);
	m_pDsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDsArg);
	
	m_pDdsArg = new ShellCmdRealArg("D_ds");
	m_pDdsArg->setDescription("Angular diameter distance from lens to source.");
	m_pDdsArg->setMin(0);
	m_pDdsArg->setUnitType(UNIT_DISTANCE);
	addArgument(m_pDdsArg);

	m_pAngularRadiusArg = new ShellCmdRealArg("radius");
	m_pAngularRadiusArg->setMin(0);
	m_pAngularRadiusArg->setUnitType(UNIT_ANGULAR);
	m_pAngularRadiusArg->setDescription("Radius of the constant density disk.");
	addArgument(m_pAngularRadiusArg);
	
	setDescription("Creates a disk of constant density of which the density corresponds to the critical density "
		       "for the specified angular diameter distances.");
}

CmdLensMassDisk::~CmdLensMassDisk()
{
	delete m_pDsArg;
	delete m_pDdsArg;
	delete m_pAngularRadiusArg;
}

bool CmdLensMassDisk::execute()
{
	double D_d = GetD_d();
	double D_ds = m_pDdsArg->getValue();
	double D_s = m_pDsArg->getValue();
	double radius = m_pAngularRadiusArg->getValue();
	
	grale::MassDiskLens *pLens = new grale::MassDiskLens();
	grale::MassDiskLensParams params(D_d, D_s, D_ds, radius);

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the mass-disk lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}

CmdLensLoadSave::CmdLensLoadSave(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Filename of the lens.");
	addArgument(fnamearg);
}

CmdLensLoadSave::~CmdLensLoadSave()
{
	delete fnamearg;
}

CmdLensLoad::CmdLensLoad(const std::string &cmdname) : CmdLensLoadSave(cmdname)
{
	setDescription("Load lens from the specified file.");
}

CmdLensLoad::~CmdLensLoad()
{
}

bool CmdLensLoad::execute()
{
	std::string fname = GetFileName();
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	std::string errstr;
	grale::GravitationalLens *l;

	if (!grale::GravitationalLens::load(fname,&l,errstr))
	{
		setErrorString(std::string("Couldn't load lens: ") + errstr);
		return false;
	}

	gsinst->SetLens(l);

	return true;
}

CmdLensSave::CmdLensSave(const std::string &cmdname) : CmdLensLoadSave(cmdname)
{
	setDescription("Save lens to the specified file.");
}

CmdLensSave::~CmdLensSave()
{
}

bool CmdLensSave::execute()
{
	std::string fname = GetFileName();
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *l = gsinst->GetLens();

	if (l == 0)
	{
		setErrorString("No lens has been created yet.");
		return false;
	}
	
	if (!l->save(fname))
	{
		setErrorString(std::string("Couldn't save lens: ") + l->getErrorString());
		return false;
	}
	
	return true;
}

CmdLensPlotMassBase::CmdLensPlotMassBase(const std::string &cmdName) : ShellCommand(cmdName)
{
}

CmdLensPlotMassBase::~CmdLensPlotMassBase()
{
}

grale::Real2DFunction *CmdLensPlotMassBase::getPlotFunction(grale::GravitationalLens &lens) const
{
	return new grale::Real2DLensFunction(lens);
}

CmdLensPlot::CmdLensPlot(const std::string &cmdname) : CmdLensPlotMassBase(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Output file name.");
	addArgument(fnamearg);
	x1arg = new ShellCmdRealArg("leftX");
	x1arg->setUnitType(UNIT_ANGULAR);
	x1arg->setDescription("X coordinate of the left side of the plot area.");
	addArgument(x1arg);
	y1arg = new ShellCmdRealArg("bottomY");
	y1arg->setUnitType(UNIT_ANGULAR);
	y1arg->setDescription("Y coordinate of the bottom of the plot area.");
	addArgument(y1arg);
	x2arg = new ShellCmdRealArg("rightX");
	x2arg->setUnitType(UNIT_ANGULAR);
	x2arg->setDescription("X coordinate of the right side of the plot area.");
	addArgument(x2arg);
	y2arg = new ShellCmdRealArg("topY");
	y2arg->setUnitType(UNIT_ANGULAR);
	y2arg->setDescription("Y coordinate of the left side of the plot area.");
	addArgument(y2arg);
	numxarg = new ShellCmdIntArg("numX");
	numxarg->setMin(16);
	numxarg->setDescription("Number of points to sample the plot area in the X direction.");
	addArgument(numxarg);
	numyarg = new ShellCmdIntArg("numY");
	numyarg->setMin(16);
	numyarg->setDescription("Number of points to sample the plot area in the Y direction.");
	addArgument(numyarg);
}

CmdLensPlot::~CmdLensPlot()
{
	delete fnamearg;
	delete x1arg;
	delete x2arg;
	delete y1arg;
	delete y2arg;
	delete numxarg;
	delete numyarg;
}

CmdLensPlotGnuplot::CmdLensPlotGnuplot(const std::string &cmdname) : CmdLensPlot(cmdname)
{
	numxarg->setDefault("75");
	numyarg->setDefault("75");
	
	arg3d = new ShellCmdBoolArg("3d","yes");
	arg3d->setDescription("Flag indicating if a 3D plot should be generated.");
	addArgument(arg3d);
	colorarg = new ShellCmdBoolArg("color","yes");
	colorarg->setDescription("Flag indicating if color output should be generated.");
	addArgument(colorarg);
	xlabelarg = new ShellCmdStringArg("Xlabel","X");
	xlabelarg->setDescription("Label of the X axis.");
	addArgument(xlabelarg);
	ylabelarg = new ShellCmdStringArg("Ylabel","Y");
	ylabelarg->setDescription("Label of the Y axis.");
	addArgument(ylabelarg);
	setDescription("Create a gnuplot file of the mass density of the lens.");
}

CmdLensPlotGnuplot::~CmdLensPlotGnuplot()
{
	delete arg3d;
	delete colorarg;
	delete xlabelarg;
	delete ylabelarg;
}

bool CmdLensPlotGnuplot::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens = gsinst->GetLens();
	
	if (lens == 0)
	{
		setErrorString("A lens must have been created first");
		return false;
	}

	grale::Real2DFunction *pFunction = getPlotFunction(*lens);
	assert(pFunction != 0);

	grale::Real2DFunctionGnuPlotPlotter p(GetFileName(),gsinst->GetAngularUnit(),gsinst->GetAngularUnit());
	
	p.setColor(colorarg->getValue());
	p.setXLabel(xlabelarg->getValue());
	p.setYLabel(ylabelarg->getValue());
	p.set3D(arg3d->getValue());
	
	grale::Vector2D<double> bottomleft(GetX1(),GetY1());
	grale::Vector2D<double> topright(GetX2(),GetY2());
	
	if (!p.plot(*pFunction,bottomleft,topright,GetNumX(),GetNumY()))
	{
		delete pFunction;
		setErrorString(std::string("Error creating plot: ") + p.getErrorString());
		return false;
	}
	
	delete pFunction;
	return true;
}

CmdLensPlotFITS::CmdLensPlotFITS(const std::string &cmdname) : CmdLensPlot(cmdname)
{
	numxarg->setDefault("512");
	numyarg->setDefault("512");
	
	setDescription("Create a FITS plot of the mass density of the lens.");
}

CmdLensPlotFITS::~CmdLensPlotFITS()
{
}

bool CmdLensPlotFITS::execute()
{
#ifdef GRALECONFIG_SUPPORT_CFITSIO
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens = gsinst->GetLens();
	
	if (lens == 0)
	{
		setErrorString("A lens must have been created first");
		return false;
	}

	grale::Real2DFunction *pFunction = getPlotFunction(*lens);
	assert(pFunction != 0);

	grale::Real2DFunctionFITSPlotter p(GetFileName(),gsinst->GetAngularUnit(),gsinst->GetAngularUnit());
	
	grale::Vector2D<double> bottomleft(GetX1(),GetY1());
	grale::Vector2D<double> topright(GetX2(),GetY2());
	
	if (!p.plot(*pFunction,bottomleft,topright,GetNumX(),GetNumY()))
	{
		delete pFunction;
		setErrorString(std::string("Error creating plot: ") + p.getErrorString());
		return false;
	}

	delete pFunction;
	return true;
#else
	setErrorString("No cfitsio support was compiled in");
	return false;
#endif // GRALECONFIG_SUPPORT_CFITSIO
}

CmdLensPlotProfile::CmdLensPlotProfile(const std::string &cmdname) : ShellCommand(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename");
	fnamearg->setDescription("Name of the file in which the profile data should be stored.");
	addArgument(fnamearg);
	radiusarg = new ShellCmdRealArg("radius");
	radiusarg->setUnitType(UNIT_ANGULAR);
	radiusarg->setMin(0);
	radiusarg->setDescription("Radius of the area in which the profile should be calculated.");
	addArgument(radiusarg);
	xarg = new ShellCmdRealArg("x");
	xarg->setUnitType(UNIT_ANGULAR);
	xarg->setDescription("X coordinate of the center of the area for which the profile is calculated.");
	xarg->setDefault("0");
	addArgument(xarg);
	yarg = new ShellCmdRealArg("y");
	yarg->setUnitType(UNIT_ANGULAR);
	yarg->setDescription("Y coordinate of the center of the area for which the profile is calculated.");
	yarg->setDefault("0");
	addArgument(yarg);
	rsteparg = new ShellCmdIntArg("rsteps");
	rsteparg->setMin(10);
	rsteparg->setDefault("100");
	rsteparg->setDescription("Number of integration steps along the radial direction.");
	addArgument(rsteparg);
	asteparg = new ShellCmdIntArg("asteps");
	asteparg->setMin(10);
	asteparg->setDefault("100");
	asteparg->setDescription("Number of integration steps in the angular direction.");
	addArgument(asteparg);
	flagarg = new ShellCmdBoolArg("densityflag");
	flagarg->setDefault("yes");
	flagarg->setDescription("Flag indicating if the density profile of the mass profile should be calculated.");
	addArgument(flagarg);

	setDescription("Calculates the circularly averaged mass density profile of the current lens.");
}

CmdLensPlotProfile::~CmdLensPlotProfile()
{
	delete fnamearg;
	delete radiusarg;
	delete xarg;
	delete yarg;
	delete rsteparg;
	delete asteparg;
	delete flagarg;
}

bool CmdLensPlotProfile::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *lens = gsinst->GetLens();
	IOSystem *iosys = gsinst->getIOSystem();

	if (lens == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}

	FILE *f = fopen(fnamearg->getValue().c_str(),"wt");
	if (f == 0)
	{
		setErrorString("Could not open specified file");
		return false;
	}
	
	grale::Vector2D<double> center(xarg->getValue(), yarg->getValue());
	double r = radiusarg->getValue();
	int numr = rsteparg->getValue();
	int numa = asteparg->getValue();
	grale::Real2DLensFunction dens(*lens);

	double rstep = r/(double)numr;
	double startr = 0;
	bool usedens = flagarg->getValue();
	int prevpct = -1;
	double massunit = gsinst->GetMassUnit();
	double D_d = lens->getLensDistance();
	
	if (usedens)
	{
		for (int i = 0 ; i < numr ; i++, startr += rstep)
		{
			grale::CircularIntegrator integrator(center, startr, startr+rstep, 1, numa);
			double area = CONST_PI*(2.0*startr*rstep+rstep*rstep);
			double sum = integrator.integrate(dens);
			
			sum /= area;

			fprintf(f,"%g %g\n",(double)((startr+rstep/2.0)/gsinst->GetAngularUnit()),(double)(sum));
			int pct = (int)((((double)(i+1)/(double)(numr))*100.0)+0.5);
			if (pct != prevpct)
			{
				iosys->setPercentageFeedback(pct);
				prevpct = pct;
			}
		}
	}
	else
	{
		double totalsum = 0;
		
		for (int i = 0 ; i < numr ; i++, startr += rstep)
		{
			grale::CircularIntegrator integrator(center, startr, startr+rstep, 1, numa);
			double area = CONST_PI*(2.0*startr*rstep+rstep*rstep);
			double sum = integrator.integrate(dens)*D_d*D_d/massunit;
			
			totalsum += sum;

			fprintf(f,"%g %g\n",(double)((startr+rstep/2.0)/gsinst->GetAngularUnit()),(double)(totalsum));
			int pct = (int)((((double)(i+1)/(double)(numr))*100.0)+0.5);
			if (pct != prevpct)
			{
				iosys->setPercentageFeedback(pct);
				prevpct = pct;
			}
		}
	}
	
	iosys->writeOutputLine(" 100%");
	iosys->writeOutputLine("Finished");

	fclose(f);
	
	return true;
}

CmdLensAverage::CmdLensAverage(const std::string &cmdname) : ShellCommand(cmdname)
{
	checkdistarg = new ShellCmdBoolArg("checkdistances");
	checkdistarg->setDefault("yes");
	checkdistarg->setDescription("Flag indicating if a check should be performed to verify that the angular diameter distances of all the lenses are the same.");
	addArgument(checkdistarg);

	setDescription("Create an average multiple plummer lens, multiple square lens or multiple gaussian lens from a number of input lenses. "
		       "If 'checkdistances' is false, the distance of the first input lens will be used. ");

	fnamearg = new ShellCmdStringArg("filename");
	fileargs.push_back(fnamearg);
}

CmdLensAverage::~CmdLensAverage()
{
	delete checkdistarg;
	delete fnamearg;
}

bool CmdLensAverage::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *iosys = gsinst->getIOSystem();
	std::string prefix("Lens file: ");
	double D_d = -1;
	
	iosys->writeOutputLine("Enter the file names of the multiple plummer lenses, multiple square lenses or multiple gaussian lenses below. Type 'end' to end input.");

	std::list<grale::PlummerLensInfo> plummerinfo;
	std::list<grale::SquareLensInfo> squareinfo;
	std::list<grale::GaussLensInfo> gaussinfo;

#define LENSTYPE_UNKNOWN -1
#define LENSTYPE_PLUMMER 0
#define LENSTYPE_SQUARE	 1
#define LENSTYPE_GAUSS	 2
	
	int lenstype = LENSTYPE_UNKNOWN;
	int lenscount = 0;
	
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;
		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;

			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = args[0];

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(fileargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing filename: ") + errstr);
				}
				else
				{
					// ok, load lens and add distribution to plummers or squares
					lenscount++;

					grale::GravitationalLens *l = 0;
					std::string errstr;

					if (!grale::GravitationalLens::load(fnamearg->getValue(), &l, errstr))
					{
						setErrorString(std::string("Couldn't load specified lens: ") + errstr);
						error = true;
						done = true;
					}
					else
					{
						if (D_d < 0)
							D_d = l->getLensDistance();
						else
						{
							if (checkdistarg->getValue())
							{
								if (D_d != l->getLensDistance())
								{
									setErrorString("Specified lens has a different angular diameter distance than previous lenses");
									error = true;
									done = true;
								}
							}
						}

						if (!error)
						{
							const grale::GravitationalLensParams *lensparams = l->getLensParameters();
							if (lensparams != 0)
							{
								int currenttype = LENSTYPE_UNKNOWN;
								
								const grale::MultiplePlummerLensParams *mparams = dynamic_cast<const grale::MultiplePlummerLensParams *>(lensparams);
								const grale::MultipleSquareLensParams *sparams = dynamic_cast<const grale::MultipleSquareLensParams *>(lensparams);
								const grale::MultipleGaussLensParams *gparams = dynamic_cast<const grale::MultipleGaussLensParams *>(lensparams);

								if (mparams == 0 && sparams == 0 && gparams == 0)
								{
									setErrorString("Specified lens is not a multiple plummer lens, multiple square lens or multiple gaussian lens");
									error = true;
									done = true;
								}
								else
								{
									if (sparams != 0)
										currenttype = LENSTYPE_SQUARE;
									else if (mparams != 0)
										currenttype = LENSTYPE_PLUMMER;
									else if (gparams != 0)
										currenttype = LENSTYPE_GAUSS;

									if (lenstype == LENSTYPE_UNKNOWN)
										lenstype = currenttype;

									if (lenstype != currenttype)
									{
										setErrorString("Current lens is incompatible with the previous lenses");
										error = true;
										done = true;
									}
									else
									{
										if (lenstype == LENSTYPE_PLUMMER)
										{
											// got the parameters, append the plummer info
											const std::list<grale::PlummerLensInfo> &info = mparams->getLensInfo();
											std::list<grale::PlummerLensInfo>::const_iterator it;

											for (it = info.begin() ; it != info.end() ; it++)
												plummerinfo.push_back(*it);
										}
										else if (lenstype == LENSTYPE_GAUSS)
										{
											// got the parameters, append the gaussian info
											const std::list<grale::GaussLensInfo> &info = gparams->getLensInfo();
											std::list<grale::GaussLensInfo>::const_iterator it;

											for (it = info.begin() ; it != info.end() ; it++)
												gaussinfo.push_back(*it);
					
										}
										else // square
										{
											// got the parameters, append the plummer info
											const std::list<grale::SquareLensInfo> &info = sparams->getLensInfo();
											std::list<grale::SquareLensInfo>::const_iterator it;

											for (it = info.begin() ; it != info.end() ; it++)
												squareinfo.push_back(*it);
										}
									}
								}
							}
						}
						delete l;
					}
				}
			}
		}
	}

	if (error)
		return false;

	// check if any lenses were loaded
	if (lenscount == 0)
	{
		setErrorString("No lens files were specified");
		return false;
	}
	
	if (plummerinfo.size() == 0 && squareinfo.size() == 0)
	{
		setErrorString("No mass distributions were found");
		return false;
	}
	
	// parameters are filled in, create the lens
	
	if (lenstype == LENSTYPE_PLUMMER)
	{
		std::list<grale::PlummerLensInfo>::iterator it;

		for (it = plummerinfo.begin() ; it != plummerinfo.end() ; it++)
		{
			double mass = (*it).getMass();
			mass /= (double)lenscount;
			(*it).setMass(mass);
		}
		
		grale::MultiplePlummerLens *l = new grale::MultiplePlummerLens();
		grale::MultiplePlummerLensParams p(plummerinfo);

		if (!l->init(D_d,&p))
		{
			setErrorString(std::string("Error initializing the multiple Plummer lens: ") + l->getErrorString());
			delete l;
			return false;
		}

		// Ok, lens created and initialized, now set it as the active lens
		
		gsinst->SetLens(l);
	}
	else if (lenstype == LENSTYPE_GAUSS)
	{
		std::list<grale::GaussLensInfo>::iterator it;

		for (it = gaussinfo.begin() ; it != gaussinfo.end() ; it++)
		{
			double mass = (*it).getMass();
			mass /= (double)lenscount;
			(*it).setMass(mass);
		}
		
		grale::MultipleGaussLens *l = new grale::MultipleGaussLens();
		grale::MultipleGaussLensParams p(gaussinfo);

		if (!l->init(D_d,&p))
		{
			setErrorString(std::string("Error initializing the multiple gaussian lens: ") + l->getErrorString());
			delete l;
			return false;
		}

		// Ok, lens created and initialized, now set it as the active lens
		
		gsinst->SetLens(l);
		
	}
	else // square
	{
		std::list<grale::SquareLensInfo>::iterator it;

		for (it = squareinfo.begin() ; it != squareinfo.end() ; it++)
		{
			double mass = (*it).getMass();
			mass /= (double)lenscount;
			(*it).setMass(mass);
		}
		
		grale::MultipleSquareLens *l = new grale::MultipleSquareLens();
		grale::MultipleSquareLensParams p(squareinfo);

		if (!l->init(D_d,&p))
		{
			setErrorString(std::string("Error initializing the multiple square lens: ") + l->getErrorString());
			delete l;
			return false;
		}

		// Ok, lens created and initialized, now set it as the active lens
		
		gsinst->SetLens(l);
	}
		
	return true;
}

// Square lens

CmdLensSquare::CmdLensSquare(const std::string &cmdname) : CmdLens(cmdname)
{
	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	massarg->setDescription("Total mass of the square lens.");
	addArgument(massarg);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	widtharg->setDescription("Width of the square shaped region.");
	addArgument(widtharg);
	setDescription("Create a lens with a square mass distribution.");
}

CmdLensSquare::~CmdLensSquare()
{
	delete massarg;
	delete widtharg;
}

bool CmdLensSquare::execute()
{
	double D_d = GetD_d();
	double mass = massarg->getValue();
	double width = widtharg->getValue();
	
	grale::SquareLens *l = new grale::SquareLens();
	grale::SquareLensParams p(mass,width);

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the square lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();

	gsinst->SetLens(l);

	return true;
}

// Multiple squares lens

CmdLensMultipleSquares::CmdLensMultipleSquares(const std::string &cmdname) : CmdLens(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename","*");
	fnamearg->setDescription("File name to read square info from or '*' to read from current input.");
	addArgument(fnamearg);
	setDescription("Create a lens consisting of multiple square-shaped distributions.");

	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	xposarg = new ShellCmdRealArg("x");
	xposarg->setUnitType(UNIT_ANGULAR);
	yposarg = new ShellCmdRealArg("y");
	yposarg->setUnitType(UNIT_ANGULAR);
	squareargs.push_back(massarg);
	squareargs.push_back(widtharg);
	squareargs.push_back(xposarg);
	squareargs.push_back(yposarg);
}

CmdLensMultipleSquares::~CmdLensMultipleSquares()
{
	delete fnamearg;
	delete massarg;
	delete widtharg;
	delete xposarg;
	delete yposarg;
}

bool CmdLensMultipleSquares::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	double D_d = GetD_d();
	std::string fname = fnamearg->getValue();
	IOSystem *iosys;
	FileIOSystem fiosys;
	std::string prefix("Square info: ");
	
	if (fname == std::string("*")) // use current input
		iosys = gsinst->getIOSystem();
	else
	{
		if (!fiosys.init(fname))
		{
			setErrorString(std::string("Can't open file '") + fname + std::string("' to read square data from: ") + fiosys.getErrorString());
			return false;
		}
		iosys = &fiosys;
	}
	
	iosys->writeOutputLine("Enter individual square parameters below. Type 'end' to end input.");
	iosys->writeOutputLine("Argument order is: ");
	iosys->writeOutputLine(ShellCmdArg::getArgumentNames(squareargs));
	
	grale::MultipleSquareLensParams p;
	
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;
		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;

			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = *(args.begin());

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(squareargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing individual square parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid
					
					double mass = massarg->getValue();
					double width = widtharg->getValue();
					double x = xposarg->getValue();
					double y = yposarg->getValue();

					p.addLensInfo(grale::SquareLensInfo(mass,width,grale::Vector2D<double>(x,y)));
				}
			}
		}
	}

	if (error)
		return false;

	// parameters are filled in, create the lens
	
	grale::MultipleSquareLens *l = new grale::MultipleSquareLens();

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the multiple square lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	gsinst->SetLens(l);

	return true;
}

// Multiple squares lens, based on a FITS file

CmdLensMultipleSquaresFits::CmdLensMultipleSquaresFits(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pFileArg = new ShellCmdStringArg("filename");
	m_pFileArg->setDescription("Name of the FITS file from which the mass density should be read.");
	addArgument(m_pFileArg);
	m_pWidthArg = new ShellCmdRealArg("width");
	m_pWidthArg->setDescription("Width of the data in the FITS file. Will be used to calculate the size of a pixel in the data file.");
	m_pWidthArg->setUnitType(UNIT_ANGULAR);
	m_pWidthArg->setMin(0);
	addArgument(m_pWidthArg);
	m_pTotalMassArg = new ShellCmdRealArg("mass");
	m_pTotalMassArg->setDescription("Total mass of the lens. The pixel values will be scaled to obtain this value as the total mass of the lens. Set to a negative value to disable this scaling. ");
	m_pTotalMassArg->setUnitType(UNIT_MASS);
	addArgument(m_pTotalMassArg);

	setDescription("Create a lens consisting of multiple squares, of which the values are based on the data in a FITS file.");
}

CmdLensMultipleSquaresFits::~CmdLensMultipleSquaresFits()
{
	delete m_pFileArg;
	delete m_pWidthArg;
	delete m_pTotalMassArg;
}

bool CmdLensMultipleSquaresFits::execute()
{
	double D_d = GetD_d();
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::FITSImage fitsImage;

	if (!fitsImage.loadNoCoordinates(m_pFileArg->getValue()))
	{
		setErrorString(std::string("Couldn't open specified file: ") + fitsImage.getErrorString());
		return false;
	}

	int width = fitsImage.getWidth();
	int height = fitsImage.getHeight();
	double pixelSize = m_pWidthArg->getValue()/(double)width;
	double xOffset = -m_pWidthArg->getValue()/2.0;
	double yOffset = -((m_pWidthArg->getValue()*(double)height)/(double)width)/2.0;
	double pixelSum = 0;
	double scaleValue = 1.0;
	
	if (m_pTotalMassArg->getValue() > 0)
	{
		for (int i = 0 ; i < width ; i++)
			for (int j = 0 ; j < height ; j++)
				pixelSum += fitsImage.getPixel(i, j);

		scaleValue = m_pTotalMassArg->getValue()/pixelSum;
	}
	else
	{
		scaleValue = pixelSize*pixelSize*D_d*D_d;
	}

	grale::MultipleSquareLensParams pParams;
	
	for (int i = 0 ; i < width ; i++)
	{
		for (int j = 0 ; j < height ; j++)
		{
			double unscaledMass = fitsImage.getPixel(i, j);
			double x = pixelSize*(double)i + xOffset;
			double y = pixelSize*(double)j + yOffset;

			if (unscaledMass != 0) // allows us to mask out some pixels
				pParams.addLensInfo(grale::SquareLensInfo(unscaledMass*scaleValue, pixelSize, grale::Vector2D<double>(x,y)));
		}
	}

	grale::MultipleSquareLens *pLens = new grale::MultipleSquareLens();

	if (!pLens->init(D_d, &pParams))
	{
		setErrorString(std::string("Error initializing the multiple square lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	pGSInst->SetLens(pLens);

	return true;
}

// Multiple gaussian lens

CmdLensMultipleGaussians::CmdLensMultipleGaussians(const std::string &cmdname) : CmdLens(cmdname)
{
	fnamearg = new ShellCmdStringArg("filename","*");
	fnamearg->setDescription("File name to read gaussian distribution info from or '*' to read from current input.");
	addArgument(fnamearg);
	setDescription("Create a lens consisting of multiple gaussian distributions.");

	massarg = new ShellCmdRealArg("mass");
	massarg->setUnitType(UNIT_MASS);
	widtharg = new ShellCmdRealArg("width");
	widtharg->setUnitType(UNIT_ANGULAR);
	xposarg = new ShellCmdRealArg("x");
	xposarg->setUnitType(UNIT_ANGULAR);
	yposarg = new ShellCmdRealArg("y");
	yposarg->setUnitType(UNIT_ANGULAR);
	squareargs.push_back(massarg);
	squareargs.push_back(widtharg);
	squareargs.push_back(xposarg);
	squareargs.push_back(yposarg);
}

CmdLensMultipleGaussians::~CmdLensMultipleGaussians()
{
	delete fnamearg;
	delete massarg;
	delete widtharg;
	delete xposarg;
	delete yposarg;
}

bool CmdLensMultipleGaussians::execute()
{
	GraleShellInstance *gsinst = (GraleShellInstance *)ShellInstance::getInstance();
	double D_d = GetD_d();
	std::string fname = fnamearg->getValue();
	IOSystem *iosys;
	FileIOSystem fiosys;
	std::string prefix("Gaussian distribution info: ");
	
	if (fname == std::string("*")) // use current input
		iosys = gsinst->getIOSystem();
	else
	{
		if (!fiosys.init(fname))
		{
			setErrorString(std::string("Can't open file '") + fname + std::string("' to read gaussian distribution data from: ") + fiosys.getErrorString());
			return false;
		}
		iosys = &fiosys;
	}
	
	iosys->writeOutputLine("Enter individual gaussian distribution parameters below. Type 'end' to end input.");
	iosys->writeOutputLine("Argument order is: ");
	iosys->writeOutputLine(ShellCmdArg::getArgumentNames(squareargs));
	
	grale::MultipleGaussLensParams p;
	
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;
		if (!iosys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (iosys->echo())
				iosys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;

			IOSystem::splitLine(line,args);
			if (args.size() == 1)
			{
				std::string str = *(args.begin());

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(squareargs,args,errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing individual gaussian distribution parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid
					
					double mass = massarg->getValue();
					double width = widtharg->getValue();
					double x = xposarg->getValue();
					double y = yposarg->getValue();

					p.addLensInfo(grale::GaussLensInfo(mass,width,grale::Vector2D<double>(x,y)));
				}
			}
		}
	}

	if (error)
		return false;

	// parameters are filled in, create the lens
	
	grale::MultipleGaussLens *l = new grale::MultipleGaussLens();

	if (!l->init(D_d,&p))
	{
		setErrorString(std::string("Error initializing the multiple gauss lens: ") + l->getErrorString());
		delete l;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	gsinst->SetLens(l);

	return true;
}

CmdLensComposite::CmdLensComposite(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pCenterXArg = new ShellCmdRealArg("xcenter");
	m_pCenterXArg->setUnitType(UNIT_ANGULAR);
		
	m_pCenterYArg = new ShellCmdRealArg("ycenter");
	m_pCenterYArg->setUnitType(UNIT_ANGULAR);
	
	m_pAngleArg = new ShellCmdRealArg("angle");
	m_pScaleArg = new ShellCmdRealArg("scalefactor");
	m_pFileNameArg = new ShellCmdStringArg("filename");

	m_lensArgs.push_back(m_pCenterXArg);
	m_lensArgs.push_back(m_pCenterYArg);
	m_lensArgs.push_back(m_pAngleArg);
	m_lensArgs.push_back(m_pScaleArg);
	m_lensArgs.push_back(m_pFileNameArg);

	setDescription("Creates a gravitational lens composed of several sublenses. Note "
		       "that currently no check is performed to verify that all sublenses "
		       "are at the same angular diameter distance. ");
}

CmdLensComposite::~CmdLensComposite()
{
	delete m_pCenterXArg;
	delete m_pCenterYArg;
	delete m_pAngleArg;
	delete m_pScaleArg;
	delete m_pFileNameArg;
}

bool CmdLensComposite::execute()
{
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInstance->getIOSystem();
	std::string prefix("Component info: ");
	
	pIOSys->writeOutputLine("Enter component parameters below. Type 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_lensArgs));
	
	grale::CompositeLensParams lensParams;
	
	bool done = false;
	bool error = false;
	bool gotLens = false;
	double D_d = 0;
	while (!done)
	{
		std::string line;
		
		if (!pIOSys->readInputLine(prefix, line))
			done = true;
		else
		{
			if (pIOSys->echo())
				pIOSys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;
			
			IOSystem::splitLine(line, args);
			if (args.size() == 1)
			{
				std::string str = *(args.begin());

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(m_lensArgs, args, errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing individual component parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid
					
					double xCenter = m_pCenterXArg->getValue();
					double yCenter = m_pCenterYArg->getValue();
					double angle = m_pAngleArg->getValue();
					double scale = m_pScaleArg->getValue();
					std::string fileName = m_pFileNameArg->getValue();
					grale::GravitationalLens *pLens;
					
					if (!grale::GravitationalLens::load(fileName, &pLens, errstr))
					{
						error = true;
						done = true;
						setErrorString(std::string("Couldn't load '") + fileName + std::string("': ") + errstr);
					}
					else
					{
						if (!gotLens)
						{
							gotLens = true;
							D_d = pLens->getLensDistance();
						}
						
						if (!lensParams.addLens(scale, grale::Vector2D<double>(xCenter, yCenter), angle, *pLens))
						{
							done = true;
							error = true;
							setErrorString(std::string("Couldn't add lens to the composite lens:") + lensParams.getErrorString());
						}
						delete pLens;							
					}
				}
			}
		}
	}

	if (error)
		return false;

	if (!gotLens)
	{
		setErrorString("No sublenses were specified");
		return false;
	}
	
	// parameters are filled in, create the lens
	
	grale::CompositeLens *pNewLens = new grale::CompositeLens();

	if (!pNewLens->init(D_d, &lensParams))
	{
		setErrorString(std::string("Error initializing the composite lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	pInstance->SetLens(pNewLens);

	return true;
}

CmdLensScaleDegeneracy::CmdLensScaleDegeneracy(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pCenterXArg = new ShellCmdRealArg("xcenter");
	m_pCenterXArg->setUnitType(UNIT_ANGULAR);
	m_pCenterXArg->setDescription("X-coordinate of the center of the scaling degeneracy.");
	addArgument(m_pCenterXArg);

	m_pCenterYArg = new ShellCmdRealArg("ycenter");
	m_pCenterYArg->setUnitType(UNIT_ANGULAR);
	m_pCenterYArg->setDescription("Y-coordinate of the center of the scaling degeneracy.");
	addArgument(m_pCenterYArg);

	setDescription("Look for a general source plane scaling degenracy.");
}

CmdLensScaleDegeneracy::~CmdLensScaleDegeneracy()
{
	delete m_pCenterXArg;
	delete m_pCenterYArg;
}

bool CmdLensScaleDegeneracy::execute()
{
	// TODO: make several things configurable
	
	std::list<grale::ImagesDataExtended *> imagesData;
	std::list<grale::ImagesDataExtended *>::const_iterator it;
	
	GraleShellInstance *pInstance = (GraleShellInstance *)ShellInstance::getInstance();
	
	pInstance->GetImagesDataList(imagesData);
	
	if (imagesData.size() == 0)
	{
		setErrorString("No data in the images data list");
		return false;
	}

	double D_d = GetD_d();
	int numPoints = 100000;
	double radius = 0;
	double radialStep = 0;

	grale::Vector2D<double> center(m_pCenterXArg->getValue(), m_pCenterYArg->getValue());

	for (it = imagesData.begin() ; it != imagesData.end() ; it++)
	{
		grale::ImagesDataExtended *pImg = (*it);

		for (int i = 0 ; i < pImg->getNumberOfImages() ; i++)
		{
			for (int p = 0 ; p < pImg->getNumberOfImagePoints(i) ; p++)
			{
				grale::Vector2D<double> point = pImg->getImagePointPosition(i, p) - center;

				double distance = point.getLength();

				if (distance > radius)
					radius = distance;
			}
		}
	}
	
	radius *= 1.01; // make it slightly larger
	radialStep = radius/((double)(numPoints-1));
	
	std::vector<int> sourceNumber(numPoints);
	std::vector<double> criticalDensity(imagesData.size());
	int s;
	int totalImages = 0;
	
	for (int i = 0 ; i < numPoints ; i++)
		sourceNumber[i] = -1;
	
	for (it = imagesData.begin(), s = 0 ; it != imagesData.end() ; it++, s++)
	{
		grale::ImagesDataExtended *pImg = (*it);

		// The critical density
		criticalDensity[s] = (SPEED_C*SPEED_C/(4.0*CONST_PI*CONST_G*D_d))*pImg->getDs()/pImg->getDds();
		
		for (int i = 0 ; i < pImg->getNumberOfImages() ; i++)
		{
			double minDist = -1;
			double maxDist = -1;

			for (int p = 0 ; p < pImg->getNumberOfImagePoints(i) ; p++)
			{
				grale::Vector2D<double> point = pImg->getImagePointPosition(i, p) - center;

				double distance = point.getLength();

				if (minDist < 0 || distance < minDist)
					minDist = distance;
				if (maxDist < 0 || distance > maxDist)
					maxDist = distance;
			}

			int maxIndex = (int)(maxDist/radialStep) + 1;
			int minIndex = (int)(minDist/radialStep) - 1;

			if (maxIndex > numPoints-1)
				maxIndex = numPoints-1;
			if (minIndex < 0)
				minIndex = 0;

			for (int p = minIndex ; p <= maxIndex ; p++)
			{
				if (!(sourceNumber[p] == -1 || sourceNumber[p] == s))
				{
					setErrorString("Overlapping image regions; can't construct a scale degeneracy");
					return false;
				}
				sourceNumber[p] = s;
			}

			totalImages++;
		}
	}

	std::vector<double> startRadii(totalImages);
	std::vector<double> endRadii(totalImages);
	std::vector<double> criticalMasses(totalImages);
	bool done = false;
	int index = 0;
	int curImage = 0;

	while (!done)
	{
		while (sourceNumber[index] < 0 && index < numPoints) // (*)
			index++;

		if (index >= numPoints)
			done = true;
		else // found the start of an image
		{
			int currentSourceNumber = sourceNumber[index];
			startRadii[curImage] = (double)index * radialStep;
			criticalMasses[curImage] = criticalDensity[currentSourceNumber];

			// look for the end of the image

			while (sourceNumber[index] == currentSourceNumber && index < numPoints)
			{
				sourceNumber[index] = -1; // we need to set this to -1, otherwise (*)
				                          // might fail if two regions touch each other
				index++;
			}
			
			if (index >= numPoints)
				done = true;

			index--;

			endRadii[curImage] = (double)index * radialStep;

			curImage++;
		}
	}

	startRadii.resize(curImage);
	endRadii.resize(curImage);
	criticalMasses.resize(curImage);

	int numAnnuli = curImage;
	grale::PolynomialMassProfileLensParams newLensParams;	
	
	for (int i = 0 ; i < numAnnuli ; i++)
	{
		if (i == 0)
		{
			std::vector<double> coefficients(3);

			coefficients[0] = 0;
			coefficients[1] = 0;
			coefficients[2] = 1;
			
			grale::PolynomialPart polynomialPart(0, 0, 
					                     endRadii[i], 
							     CONST_PI*D_d*D_d*criticalMasses[i]*endRadii[i]*endRadii[i], 
							     endRadii[i], 
							     coefficients);

			newLensParams.addPolynomialPart(polynomialPart);
		}
		else
		{
			// Interpolated part
			
			{
				double startRadius = endRadii[i-1];
				double endRadius = startRadii[i];
				double startMass = CONST_PI*D_d*D_d*startRadius*startRadius*criticalMasses[i-1];
				double endMass =  CONST_PI*D_d*D_d*endRadius*endRadius*criticalMasses[i];
				double startSlope = 2.0*CONST_PI*D_d*D_d*startRadius*criticalMasses[i-1];
				double endSlope = 2.0*CONST_PI*D_d*D_d*endRadius*criticalMasses[i];

				double height = endMass-startMass;
				double width = endRadius-startRadius;
				
				double scaledStartSlope = startSlope*width/height;
				double scaledEndSlope = endSlope*width/height;

				std::vector<double> coefficients(4);

				coefficients[0] = 0;
				coefficients[1] = scaledStartSlope;
				coefficients[2] = 3.0 - 2.0*scaledStartSlope-scaledEndSlope;
				coefficients[3] = -2.0 + scaledStartSlope + scaledEndSlope;
				
				grale::PolynomialPart polynomialPart(startRadius, startMass,
							             width, height,
								     endRadius,
							             coefficients);

				newLensParams.addPolynomialPart(polynomialPart);
			}
		
			// Constrained part

			{
					
				std::vector<double> coefficients(3);

				double width = endRadii[i]-startRadii[i];
				
				coefficients[0] = 0;
				coefficients[1] = 2.0*startRadii[i]/width;
				coefficients[2] = 1;
				
				grale::PolynomialPart polynomialPart(startRadii[i], 
							      CONST_PI*D_d*D_d*criticalMasses[i]*startRadii[i]*startRadii[i], 
							      width, 
							      CONST_PI*D_d*D_d*criticalMasses[i]*width*width,
							      endRadii[i],
							      coefficients);

				newLensParams.addPolynomialPart(polynomialPart);

			}
		}
	}

	if (numAnnuli > 0) // finish the profile
	{
		double startRadius = endRadii[numAnnuli-1];
		double endRadius = radius;
		double startMass = CONST_PI*D_d*D_d*startRadius*startRadius*criticalMasses[numAnnuli-1];
		double startSlope = 2.0*CONST_PI*D_d*D_d*startRadius*criticalMasses[numAnnuli-1];
		double endSlope = 0;

		double width = (endRadius-startRadius); 
		double height = 0.5*width*startSlope; 
				
		double scaledStartSlope = startSlope*width/height;

		std::vector<double> coefficients(3);

		coefficients[0] = 0;
		coefficients[1] = scaledStartSlope;
		coefficients[2] = -0.5*scaledStartSlope;
				
		grale::PolynomialPart polynomialPart(startRadius, startMass,
					             width, height,
						     endRadius,
					             coefficients);

		newLensParams.addPolynomialPart(polynomialPart);
	}

	grale::PolynomialMassProfileLens *pNewLens = new grale::PolynomialMassProfileLens();
	
	if (!pNewLens->init(D_d, &newLensParams))
	{
		setErrorString(std::string("Error initializing the polynomial profile lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}
	
	pInstance->SetLens(pNewLens);
	
	return true;
}

CmdLensGetLensPlane::CmdLensGetLensPlane(const std::string &cmdName) : ShellCommand(cmdName)
{
	setDescription("Creates a copy of the lens stored in the current lensplane and sets it as the active lens.");
}

CmdLensGetLensPlane::~CmdLensGetLensPlane()
{
}

bool CmdLensGetLensPlane::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::LensPlane *pLensPlane = pGSInst->GetLensPlane();

	if (pLensPlane == 0)
	{
		setErrorString("No lensplane has been created yet");
		return false;
	}

	const grale::GravitationalLens *pLens;

	if ((pLens = pLensPlane->getLens()) == 0)
	{
		setErrorString("Lensplane doesn't seem to contain a lens, this should never happen!");
		return false;
	}

	grale::GravitationalLens *pNewLens = pLens->createCopy();

	if (pNewLens == 0)
	{
		setErrorString(std::string("Couldn't create a copy of the lens in the lensplane instance: ") + pLens->getErrorString());
		return false;
	}

	pGSInst->SetLens(pNewLens);
	return true;
}

CmdLensGetImagePlane::CmdLensGetImagePlane(const std::string &cmdName) : ShellCommand(cmdName)
{
	setDescription("Creates a copy of the lens stored in the current image plane and sets it as the active lens.");
}

CmdLensGetImagePlane::~CmdLensGetImagePlane()
{
}

bool CmdLensGetImagePlane::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::ImagePlane *pImagePlane = pGSInst->GetImagePlane();

	if (pImagePlane == 0)
	{
		setErrorString("No lensplane has been created yet");
		return false;
	}

	const grale::GravitationalLens *pLens;

	if ((pLens = pImagePlane->getLens()) == 0)
	{
		setErrorString("Image plane doesn't seem to contain a lens, this should never happen!");
		return false;
	}

	grale::GravitationalLens *pNewLens = pLens->createCopy();

	if (pNewLens == 0)
	{
		setErrorString(std::string("Couldn't create a copy of the lens in the image plane instance: ") + pLens->getErrorString());
		return false;
	}

	pGSInst->SetLens(pNewLens);
	return true;
}

CmdLensMultipleWendland::CmdLensMultipleWendland(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pFileNameArg = new ShellCmdStringArg("filename", "*");
	m_pFileNameArg->setDescription("File name to read Wendland W7,3 info from or '*' to read from current input.");
	addArgument(m_pFileNameArg);
	setDescription("Create a lens based on multiple Wendland W7,3 distributions.");

	m_pFlagArg = new ShellCmdBoolArg("xyflag");
	m_pHeightArg = new ShellCmdRealArg("coeff");
	m_pAngularScaleArg = new ShellCmdRealArg("angscale");
	m_pAngularScaleArg->setUnitType(UNIT_ANGULAR);
	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	
	m_params.push_back(m_pFlagArg);
	m_params.push_back(m_pHeightArg);
	m_params.push_back(m_pAngularScaleArg);
	m_params.push_back(m_pXArg);
	m_params.push_back(m_pYArg);
}

CmdLensMultipleWendland::~CmdLensMultipleWendland()
{
	delete m_pFileNameArg;
	delete m_pFlagArg;
	delete m_pHeightArg;
	delete m_pAngularScaleArg;
	delete m_pXArg;
	delete m_pYArg;
}

bool CmdLensMultipleWendland::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	double D_d = GetD_d();
	std::string fileName = m_pFileNameArg->getValue();
	IOSystem *pIOSys;
	FileIOSystem fileIOSys;
	std::string prefix("Wendland info: ");
	
	if (fileName == std::string("*")) // use current input
		pIOSys = pGSInst->getIOSystem();
	else
	{
		if (!fileIOSys.init(fileName))
		{
			setErrorString(std::string("Can't open file '") + fileName + std::string("' to read Wendland data from: ") + fileIOSys.getErrorString());
			return false;
		}
		pIOSys = &fileIOSys;
	}
	
	pIOSys->writeOutputLine("Enter individual Wendland parameters below. Type 'end' to end input.");
	pIOSys->writeOutputLine("Argument order is: ");
	pIOSys->writeOutputLine(ShellCmdArg::getArgumentNames(m_params));
	
	grale::MultipleWendlandLensParams lensParams;
	
	bool done = false;
	bool error = false;
	while (!done)
	{
		std::string line;
		if (!pIOSys->readInputLine(prefix,line))
			done = true;
		else
		{
			if (pIOSys->echo())
				pIOSys->print("%s%s", prefix.c_str(), line.c_str());

			std::vector<std::string> args;

			IOSystem::splitLine(line, args);
			if (args.size() == 1)
			{
				std::string str = *(args.begin());

				if (str == std::string("end"))
					done = true;
			}

			if (!done)
			{
				std::string errstr;
				
				if (!ShellCmdArg::parseArgumentList(m_params, args, errstr))
				{
					error = true;
					done = true;
					setErrorString(std::string("Error parsing individual Wendland parameters: ") + errstr);
				}
				else
				{
					// ok, arguments are valid
					
					double height = m_pHeightArg->getValue();
					double angScale = m_pAngularScaleArg->getValue();
					double x = m_pXArg->getValue();
					double y = m_pYArg->getValue();

					if (m_pFlagArg->getValue())
						lensParams.addPhiXInfo(grale::WendlandLensInfo(height, angScale, grale::Vector2D<double>(x,y)));
					else
						lensParams.addPhiYInfo(grale::WendlandLensInfo(height, angScale, grale::Vector2D<double>(x,y)));
				}
			}
		}
	}

	if (error)
		return false;

	// parameters are filled in, create the lens
	
	grale::MultipleWendlandLens *pLens = new grale::MultipleWendlandLens();

	if (!pLens->init(D_d, &lensParams))
	{
		setErrorString(std::string("Error initializing the multiple Wendland lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	pGSInst->SetLens(pLens);

	return true;
}

CmdLensGetLensPlaneGrid::CmdLensGetLensPlaneGrid(const std::string &cmdName) : ShellCommand(cmdName)
{
	setDescription("Creates a lens based on the gridded deflection information in the current lensplane and sets it as the active lens.");
}

CmdLensGetLensPlaneGrid::~CmdLensGetLensPlaneGrid()
{
}

bool CmdLensGetLensPlaneGrid::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::LensPlane *pLensPlane = pGSInst->GetLensPlane();

	if (pLensPlane == 0)
	{
		setErrorString("No lensplane has been created yet");
		return false;
	}

	grale::GravitationalLens *pLens;

	if ((pLens = pLensPlane->createDeflectionGridLens()) == 0)
	{
		setErrorString(std::string("Couldn't create new deflection grid lens: ") + pLensPlane->getErrorString());
		return false;
	}

	pGSInst->SetLens(pLens);
	return true;
}

CmdLensTimeDelayAdjust::CmdLensTimeDelayAdjust(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pTimeDiffArg = new ShellCmdRealArg("timediff");
	m_pTimeDiffArg->setDescription("Time difference introduced at the center of the lens (in days).");
	addArgument(m_pTimeDiffArg);

	m_pLensZArg = new ShellCmdRealArg("z_lens");
	m_pLensZArg->setDescription("Redshift of the lens");
	m_pLensZArg->setMin(0);
	addArgument(m_pLensZArg);

	m_pThetaStartArg = new ShellCmdRealArg("theta1");
	m_pThetaStartArg->setDescription("Start radius. Below this radius the mass density will be zero.");
	m_pThetaStartArg->setUnitType(UNIT_ANGULAR);
	m_pThetaStartArg->setMin(0);
	addArgument(m_pThetaStartArg);

	m_pThetaStopArg = new ShellCmdRealArg("theta2");
	m_pThetaStopArg->setDescription("Stop radius. Beyond this radius the mass density will be zero.");
	m_pThetaStopArg->setUnitType(UNIT_ANGULAR);
	m_pThetaStopArg->setMin(0);
	addArgument(m_pThetaStopArg);

	setDescription("Create a lens which adds a specific time delay in the center of the lens.");
}

CmdLensTimeDelayAdjust::~CmdLensTimeDelayAdjust()
{
	delete m_pTimeDiffArg;
	delete m_pLensZArg;
	delete m_pThetaStopArg;
	delete m_pThetaStartArg;
}

bool CmdLensTimeDelayAdjust::execute()
{
	double theta1 = m_pThetaStartArg->getValue();
	double theta2 = m_pThetaStopArg->getValue();

	if (theta1 >= theta2)
	{
		setErrorString("The start radius must be smaller than the stop radius");
		return false;
	}

	double deltaT = m_pTimeDiffArg->getValue()*24.0*60.0*60.0; // convert days to seconds

	grale::PolynomialTimeDelayLensParams newLensParams(theta1, theta2, deltaT, m_pLensZArg->getValue());

	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::PolynomialMassProfileLens *pNewLens = new grale::PolynomialMassProfileLens();
	double D_d = GetD_d();

	if (!pNewLens->init(D_d, &newLensParams))
	{
		setErrorString(std::string("Error initializing the polynomial profile lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}
	
	pGSInst->SetLens(pNewLens);

	return true;
}

CmdLensMagnificationAdjust::CmdLensMagnificationAdjust(const std::string &cmdName) : ShellCommand(cmdName)
{
	m_pXArg = new ShellCmdRealArg("x");
	m_pXArg->setUnitType(UNIT_ANGULAR);
	m_pXArg->setDescription("X Coordinate of the point at which the magnification should be adjusted.");
	addArgument(m_pXArg);

	m_pYArg = new ShellCmdRealArg("y");
	m_pYArg->setUnitType(UNIT_ANGULAR);
	m_pYArg->setDescription("Y Coordinate of the point at which the magnification should be adjusted.");
	addArgument(m_pYArg);

	m_pTargetMagnificationArg = new ShellCmdRealArg("mag");
	m_pTargetMagnificationArg->setDescription("Target magnification at the specified coordinates.");
	addArgument(m_pTargetMagnificationArg);

	m_pMaxWidthArg = new ShellCmdRealArg("width");
	m_pMaxWidthArg->setUnitType(UNIT_ANGULAR);
	m_pMaxWidthArg->setMin(0);
	m_pMaxWidthArg->setDescription("Width of the added mass distribution. Beyond this radius, the mass density will not be altered.");
	addArgument(m_pMaxWidthArg);

	// TODO!!
	m_pPotentialScaleArg = new ShellCmdRealArg("scale[TODO]");
	m_pPotentialScaleArg->setDescription("TODO");
	addArgument(m_pPotentialScaleArg);

	m_pConstantTimeDelayArg = new ShellCmdBoolArg("keeptimedelay");
	m_pConstantTimeDelayArg->setDefault("yes");
	m_pConstantTimeDelayArg->setDescription("If set to 'yes' (default), the time delay surface outside the specified area will not be modified.");
	addArgument(m_pConstantTimeDelayArg);

	// TODO!
	setDescription("TODO");
}

CmdLensMagnificationAdjust::~CmdLensMagnificationAdjust()
{
	delete m_pXArg;
	delete m_pYArg;
	delete m_pTargetMagnificationArg;
	delete m_pMaxWidthArg;
	delete m_pPotentialScaleArg;
	delete m_pConstantTimeDelayArg;
}

bool CmdLensMagnificationAdjust::execute()
{
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();
	grale::GravitationalLens *pLens = pGSInst->GetLens();
	grale::SourcePlane *pSrcPlane = pGSInst->GetSourcePlane();

	if (pLens == 0)
	{
		setErrorString("No lens has been created yet");
		return false;
	}
	if (pSrcPlane == 0)
	{
		setErrorString("No source plane has been created yet");
		return false;
	}

	bool constTimeDelay = m_pConstantTimeDelayArg->getValue();

	grale::Vector2D<double> theta(m_pXArg->getValue(), m_pYArg->getValue());
	double axx,ayy,axy;

	if (!pLens->getAlphaVectorDerivatives(theta, axx, ayy, axy))
	{
		setErrorString(std::string("Unable to get deflection derivatives for the specified point and current lens: ") + pLens->getErrorString());
		return false;
	}

	double D_ds = pSrcPlane->getD_ds();
	double D_s = pSrcPlane->getD_s();
	double D_d = pLens->getLensDistance();
	double frac = D_ds/D_s;

	double Axx = frac*axx;
	double Ayy = frac*ayy;
	double Axy = frac*axy;

	double A = 1.0;
	double B = -(2.0-Axx-Ayy);

	double invMu0 = (1.0-Axx)*(1.0-Ayy)-Axy*Axy;
	double invMu = 1.0/m_pTargetMagnificationArg->getValue();
	double C = invMu0-invMu;

	double discriminant = B*B-4.0*A*C;
	if (discriminant < 0)
	{
		setErrorString("Couldn't find an appropriate value to adjust the magnification");
		return false;
	}

	double G1 = (-B + SQRT(discriminant))/(2.0*A);
	double G2 = (-B - SQRT(discriminant))/(2.0*A);
	double G = 0;

	if (ABS(G1) < ABS(G2))
		G = G1;
	else
		G = G2;

	G /= frac;

	double W = m_pMaxWidthArg->getValue();

	if (constTimeDelay)
		W /= 2.0;

	double H = m_pPotentialScaleArg->getValue()*24.0*60.0*60.0*SPEED_C/D_d;
	double w = G*W*W/H;

	grale::PolynomialMassProfileLensParams polyLensParams;

	{
		std::vector<double> a(6);

		a[0] = 0;
		a[1] = 0;
		a[2] = w;
		a[3] = 3.0*(10.0-3.0/2.0*w);
		a[4] = 4.0*(-15.0+3.0/2.0*w);
		a[5] = 5.0*(6.0-0.5*w);
		
		grale::PolynomialPart p1(0, 0, W, H*D_d*SPEED_C*SPEED_C/(4.0*CONST_G), W, a);

		polyLensParams.addPolynomialPart(p1);
	}

	if (constTimeDelay)
	{
		std::vector<double> b(6);
		double H2 = 1.0; // W/(2*W - W)

		b[0] = 0;
		b[1] = 0;
		b[2] = -30.0*H2;
		b[3] = -30.0+60*H2;
		b[4] = 60.0-30.0*H2;
		b[5] = -30.0;

		grale::PolynomialPart p2(W, 0, W, D_d*SPEED_C*SPEED_C*H/(4.0*CONST_G), 2.0*W, b);

		polyLensParams.addPolynomialPart(p2);
	}

	grale::PolynomialMassProfileLens *pNewLens = new grale::PolynomialMassProfileLens();

	if (!pNewLens->init(D_d, &polyLensParams))
	{
		setErrorString(std::string("Error initializing the polynomial profile lens: ") + pNewLens->getErrorString());
		delete pNewLens;
		return false;
	}
	
	pGSInst->SetLens(pNewLens);

	return true;
}

CmdLensTestLens::CmdLensTestLens(const std::string &cmdName) : CmdLens(cmdName)
{
	setDescription("Test lens");
}

CmdLensTestLens::~CmdLensTestLens()
{
}

bool CmdLensTestLens::execute()
{
	double D_d = GetD_d();
	
	grale::TestLens *pLens = new grale::TestLens();
	grale::TestLensParams params;

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the test lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	// Ok, lens created and initialized, now set it as the active lens
	
	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}

CmdLensNFW::CmdLensNFW(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDensScale = new ShellCmdRealArg("rho_s");
	m_pDensScale->setDescription("Density scale rho_s in the expression of an NFW profile, in units of kg m^-3.");
	m_pDensScale->setMin(0);
	addArgument(m_pDensScale);

	m_pAngularRadiusScale = new ShellCmdRealArg("theta_s");
	m_pAngularRadiusScale->setDescription("Scale radius of the NFW profile, expressed in angular units (theta_s = r_s/D_d).");
	m_pAngularRadiusScale->setMin(0);
	m_pAngularRadiusScale->setUnitType(UNIT_ANGULAR);
	addArgument(m_pAngularRadiusScale);

	setDescription("Create symmetric NFW lens.");
}

CmdLensNFW::~CmdLensNFW()
{
	delete m_pDensScale;
	delete m_pAngularRadiusScale;
}

bool CmdLensNFW::execute()
{
	double D_d = GetD_d();

	grale::NFWLens *pLens = new grale::NFWLens();
	grale::NFWLensParams params(m_pDensScale->getValue(), m_pAngularRadiusScale->getValue());

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the NFW lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}

CmdLensEllipticNFW::CmdLensEllipticNFW(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDensScale = new ShellCmdRealArg("rho_s");
	m_pDensScale->setDescription("Density scale rho_s in the expression of an NFW profile, in units of kg m^-3.");
	m_pDensScale->setMin(0);
	addArgument(m_pDensScale);

	m_pAngularRadiusScale = new ShellCmdRealArg("theta_s");
	m_pAngularRadiusScale->setDescription("Scale radius of the NFW profile, expressed in angular units (theta_s = r_s/D_d).");
	m_pAngularRadiusScale->setMin(0);
	m_pAngularRadiusScale->setUnitType(UNIT_ANGULAR);
	addArgument(m_pAngularRadiusScale);

	m_pEllipticityArg = new ShellCmdRealArg("q");
	m_pEllipticityArg->setDescription("Ellipticity parameter.");
	m_pEllipticityArg->setMin(0);
	m_pEllipticityArg->setMax(1);
	addArgument(m_pEllipticityArg);

	setDescription("Create an elliptical generalization of a symmetric NFW lens. Density is calculated as Sigma_ell(x,y) = Sigma_symm(R), in which R^2 = (q*x^2+y^2/q).");
}

CmdLensEllipticNFW::~CmdLensEllipticNFW()
{
	delete m_pDensScale;
	delete m_pAngularRadiusScale;
	delete m_pEllipticityArg;
}

bool CmdLensEllipticNFW::execute()
{
	double D_d = GetD_d();
	double q = m_pEllipticityArg->getValue();
	double rho_s = m_pDensScale->getValue();
	double theta_s = m_pAngularRadiusScale->getValue();

	double sqrtQ = SQRT(q);

	// Need to adjust the values somewhat, since in the library R = x^2 + (y/q)^2 is used instead of q*x^2+y^2/q
	rho_s *= sqrtQ;
	theta_s /= sqrtQ;

	grale::EllipticNFWLens *pLens = new grale::EllipticNFWLens();
	grale::EllipticNFWLensParams params(rho_s, theta_s, q);

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the elliptic NFW lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}


CmdLensSersic::CmdLensSersic(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDensScaleArg = new ShellCmdRealArg("sigma0");
	m_pDensScaleArg->setDescription("Central projected density of the profile.");
	m_pDensScaleArg->setMin(0);
	addArgument(m_pDensScaleArg);

	m_pAngularScaleArg = new ShellCmdRealArg("theta_s");
	m_pAngularScaleArg->setDescription("Scale radius of the profile, expressed in angular units.");
	m_pAngularScaleArg->setMin(0);
	m_pAngularScaleArg->setUnitType(UNIT_ANGULAR);
	addArgument(m_pAngularScaleArg);

	m_pSersicIndexArg = new ShellCmdRealArg("n");
	m_pSersicIndexArg->setDescription("Sersic index.");
	m_pSersicIndexArg->setMin(0);
	addArgument(m_pSersicIndexArg);

	setDescription("Create a symmetric lens with a Sersic density profile. The form assumed is sigma0*exp(-(theta/theta_s)^(1/n)).");
}

CmdLensSersic::~CmdLensSersic()
{
	delete m_pDensScaleArg;
	delete m_pAngularScaleArg;
	delete m_pSersicIndexArg;
}

bool CmdLensSersic::execute()
{
	double D_d = GetD_d();

	grale::SersicLens *pLens = new grale::SersicLens();
	grale::SersicLensParams params(m_pDensScaleArg->getValue(), m_pAngularScaleArg->getValue(), m_pSersicIndexArg->getValue());

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the Sersic lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}


CmdLensEllipticSersic::CmdLensEllipticSersic(const std::string &cmdName) : CmdLens(cmdName)
{
	m_pDensScaleArg = new ShellCmdRealArg("sigma0");
	m_pDensScaleArg->setDescription("Central projected density of the profile.");
	m_pDensScaleArg->setMin(0);
	addArgument(m_pDensScaleArg);

	m_pAngularScaleArg = new ShellCmdRealArg("theta_s");
	m_pAngularScaleArg->setDescription("Scale radius of the profile, expressed in angular units.");
	m_pAngularScaleArg->setMin(0);
	m_pAngularScaleArg->setUnitType(UNIT_ANGULAR);
	addArgument(m_pAngularScaleArg);

	m_pSersicIndexArg = new ShellCmdRealArg("n");
	m_pSersicIndexArg->setDescription("Sersic index.");
	m_pSersicIndexArg->setMin(0);
	addArgument(m_pSersicIndexArg);

	m_pEllipticityArg = new ShellCmdRealArg("q");
	m_pEllipticityArg->setDescription("Ellipticity parameter.");
	m_pEllipticityArg->setMin(0);
	m_pEllipticityArg->setMax(1);
	addArgument(m_pEllipticityArg);

	setDescription("Create an elliptical generalization of a symmetric Sersic lens. Density is calculated as Sigma_ell(x,y) = Sigma_symm(R), in which R^2 = (q*x^2+y^2/q).");
}

CmdLensEllipticSersic::~CmdLensEllipticSersic()
{
	delete m_pDensScaleArg;
	delete m_pAngularScaleArg;
	delete m_pSersicIndexArg;
	delete m_pEllipticityArg;
}

bool CmdLensEllipticSersic::execute()
{
	double D_d = GetD_d();
	double densScale = m_pDensScaleArg->getValue();
	double angScale = m_pAngularScaleArg->getValue();
	double n = m_pSersicIndexArg->getValue();
	double q = m_pEllipticityArg->getValue();

	// Need to adjust the value somewhat, since in the library R = x^2 + (y/q)^2 is used instead of q*x^2+y^2/q
	angScale /= SQRT(q);

	grale::EllipticSersicLens *pLens = new grale::EllipticSersicLens();
	grale::EllipticSersicLensParams params(densScale, angScale, n, q);

	if (!pLens->init(D_d, &params))
	{
		setErrorString(std::string("Error initializing the elliptic Sersic lens: ") + pLens->getErrorString());
		delete pLens;
		return false;
	}

	GraleShellInstance *pGSInst = (GraleShellInstance *)ShellInstance::getInstance();

	pGSInst->SetLens(pLens);

	return true;
}

CmdLensMass::CmdLensMass(const std::string &cmdName) : CmdLensPlotMassBase(cmdName)
{
	m_pNameArg = new ShellCmdStringArg("variablename");
	m_pNameArg->setDescription("Variable to store the result in (use default to avoid storing it), in units of the currently used mass unit.");
	m_pNameArg->setDefault("*");
	addArgument(m_pNameArg);
	m_pIgnoreOffsetArg = new ShellCmdBoolArg("ignoreconstoffset");
	m_pIgnoreOffsetArg->setDescription("Set to 'yes' if only the mass that differs from a constant offset (in the specified region) should be calculated.");
	m_pIgnoreOffsetArg->setDefault("no");
	addArgument(m_pIgnoreOffsetArg);
	m_pX1Arg = new ShellCmdRealArg("leftX");
	m_pX1Arg->setUnitType(UNIT_ANGULAR);
	m_pX1Arg->setDescription("X coordinate of the left side of the area to integrate.");
	addArgument(m_pX1Arg);
	m_pY1Arg = new ShellCmdRealArg("bottomY");
	m_pY1Arg->setUnitType(UNIT_ANGULAR);
	m_pY1Arg->setDescription("Y coordinate of the bottom of the area to integrate.");
	addArgument(m_pY1Arg);
	m_pX2Arg = new ShellCmdRealArg("rightX");
	m_pX2Arg->setUnitType(UNIT_ANGULAR);
	m_pX2Arg->setDescription("X coordinate of the right side of the area to integrate.");
	addArgument(m_pX2Arg);
	m_pY2Arg = new ShellCmdRealArg("topY");
	m_pY2Arg->setUnitType(UNIT_ANGULAR);
	m_pY2Arg->setDescription("Y coordinate of the left side of the area to integrate.");
	addArgument(m_pY2Arg);
	m_pNumXArg = new ShellCmdIntArg("numX");
	m_pNumXArg->setMin(16);
	m_pNumXArg->setDefault("256");
	m_pNumXArg->setDescription("Number of points to sample the area in the X direction.");
	addArgument(m_pNumXArg);
	m_pNumYArg = new ShellCmdIntArg("numY");
	m_pNumYArg->setMin(16);
	m_pNumYArg->setDefault("256");
	m_pNumYArg->setDescription("Number of points to sample the area in the Y direction.");
	addArgument(m_pNumYArg);

	setDescription("Estimate the total mass within a specified region, optionally storing the result in a variable. "
			       "If the 'ignoreconstoffset' is used, a constant density offset that is present in the area will "
				   "be ignored. This can be useful when using a mass sheet in an inversion: the estimated total mass "
				   "that can be specified in those inversion functions, does _not_ include the contribution of a mass "
				   "sheet. ");
}

CmdLensMass::~CmdLensMass()
{
	delete m_pNameArg;
	delete m_pIgnoreOffsetArg;
	delete m_pX1Arg;
	delete m_pX2Arg;
	delete m_pY1Arg;
	delete m_pY2Arg;
	delete m_pNumXArg;
	delete m_pNumYArg;
}

bool CmdLensMass::execute()
{
	GraleShellInstance *pInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	grale::GravitationalLens *lens = pInst->GetLens();
	
	if (lens == 0)
	{
		setErrorString("A lens must have been created first");
		return false;
	}

	const double x1 = m_pX1Arg->getValue();
	const double x2 = m_pX2Arg->getValue();
	const double y1 = m_pY1Arg->getValue();
	const double y2 = m_pY2Arg->getValue();
	const int numX = m_pNumXArg->getValue();
	const int numY = m_pNumYArg->getValue();

	if (x2 <= x1)
	{
		setErrorString("Left X-coordinate must be smaller than right one.");
		return false;
	}
	if (y2 <= y1)
	{
		setErrorString("Bottom Y-coordinate must be smaller than top one.");
		return false;
	}

	grale::Real2DFunction *pFunction = getPlotFunction(*lens);
	assert(pFunction != 0);

	grale::RectangularIntegrator integrator(grale::Vector2D<double>(x1, y1), 
			                                grale::Vector2D<double>(x2, y2), 
			                                numX-1, numY-1);
	
	double totalMass = integrator.integrate(*pFunction);
	double minVal = std::numeric_limits<double>::max();

	if (m_pIgnoreOffsetArg->getValue()) // ignore a constant offset
	{
		for (int Y = 0 ; Y < numY ; Y++)
		{
			double y = ((double)Y)*(y2-y1)/(numY-1) + y1;

			for (int X = 0 ; X < numX ; X++)
			{
				double x = ((double)X)*(x2-x1)/(numX-1) + x1;
				double val = (*pFunction)(grale::Vector2D<double>(x, y));

				if (val < minVal)
					minVal = val;
			}
		}

		if (minVal < 0)
			pIOSys->writeOutputLine("Warning: compensation for a constant offset was requested, but this is ignored because a negative value was detected.");
		else
			totalMass -= (x2-x1)*(y2-y1)*minVal;
	}

	double D_d = lens->getLensDistance();
	totalMass *= D_d*D_d;

	// Express in specified unit
	totalMass /= pInst->GetMassUnit();
	
	delete pFunction;

	pIOSys->print("Calculated mass is: %g (current units)", totalMass);

	std::string varName = m_pNameArg->getValue();
	if (varName != "*")
	{
		if (!pInst->storeVariable(varName, totalMass))
		{
			setErrorString("Unable to store result in variable: " + pInst->getErrorString());
			return false;
		}
	}

	return true;
}
