/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include <mpi.h>
#include "cmdlensesmpi.h"
#include "mainmpi.h"
#include "graleshellinstance.h"
#include <shellp/iosystem.h>
#include <grale/real2dlensfunction.h>
#include <grale/real2dfunctiongnuplotplotter.h>
#include <grale/real2dfunctionfitsplotter.h>
#include <grale/gridfunction.h>
#include <serut/memoryserializer.h>
#include <serut/dummyserializer.h>
#include <vector>

#include "debugnew.h"

using namespace shellp;
using namespace grale;
using namespace std;

class DiscreteGridFunction : public Real2DFunction
{
public:
	DiscreteGridFunction(double *pValues, Vector2D<double> bottomLeft, Vector2D<double> topRight, int numX, int numY);
	~DiscreteGridFunction();

	double operator()(Vector2D<double> v) const;
private:
	double *m_pValues;
	int m_numX;
	int m_numY;
	Vector2D<double> m_bottomLeft;
	Vector2D<double> m_topRight;
};

DiscreteGridFunction::DiscreteGridFunction(double *pValues, Vector2D<double> bottomLeft, Vector2D<double> topRight, int numX, int numY)
{
	m_pValues = pValues;
	m_bottomLeft = bottomLeft;
	m_topRight = topRight;
	m_numX = numX;
	m_numY = numY;
}

DiscreteGridFunction::~DiscreteGridFunction()
{
	delete m_pValues;
}

double DiscreteGridFunction::operator()(Vector2D<double> v) const
{
	double xFrac = (v.getX() - m_bottomLeft.getX())/(m_topRight.getX() - m_bottomLeft.getX());
	int xPos = (int)(xFrac * (m_numX-1) + 0.5); 

	if (xPos < 0)
		xPos = 0;
	if (xPos >= m_numX)
		xPos = m_numX-1;

	double yFrac = (v.getY() - m_bottomLeft.getY())/(m_topRight.getY() - m_bottomLeft.getY());
	int yPos = (int)(yFrac * (m_numY-1) + 0.5); 

	if (yPos < 0)
		yPos = 0;
	if (yPos >= m_numY)
		yPos = m_numY-1;

	return m_pValues[xPos + m_numX*yPos];
}

DiscreteGridFunction *startLensPlotMPIMaster(grale::GravitationalLens &lens, 
		                                    int numX, int numY, double x0, double x1, double y0, double y1,
											string &errStr)
{
	GraleShellInstance *pInst = (GraleShellInstance *)ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	int nProcs;
	MPI_Comm_size(MPI_COMM_WORLD, &nProcs);

	// Send the command and transmit the parameters to the other ranks
	int command = GRALESHELL_MPICOMMAND_DENSITY;
	MPI_Bcast(&command, 1, MPI_INT, 0, MPI_COMM_WORLD);

	// We're going to make it slightly larger
	double xStep = (x1-x0)/(numX-1);
	double yStep = (y1-y0)/(numY-1);
	/*
	x0 -= xStep;
	x1 += xStep;
	y0 -= yStep;
	y1 += yStep;

	numX += 2;
	numY += 2;
	*/

	int numTotal = numX*numY;
	double *pDensValues = new double[numTotal];

	// Get lens size
	
	int index = 0;
		
	serut::DummySerializer lengthCounter;
	lens.write(lengthCounter);
	
	vector<uint8_t> lensBuffer(lengthCounter.getBytesWritten());
	serut::MemorySerializer mSer(0, 0, &(lensBuffer[0]), lensBuffer.size());
	lens.write(mSer);

	// Make sure everyone gets the lens
	int lensSize = lensBuffer.size();
	MPI_Bcast(&lensSize, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&(lensBuffer[0]), lensBuffer.size(), MPI_BYTE, 0, MPI_COMM_WORLD);

	int linesRendered = 0;
	int lineToRender = 0;
	vector<int> nodeLine(nProcs); // we're not going to use 0, that process will coordinate

	for (int i = 0 ; i < nodeLine.size() ; i++)
		nodeLine[i] = -1;

	vector<double> buffer(numX);

	pIOSys->writeOutputLine("Building map");

	while (linesRendered < numY)
	{
		// Make sure all nodes have something to do
		for (int i = 1 ; i < nodeLine.size() ; i++)
		{
			if (lineToRender < numY && nodeLine[i] < 0) // available
			{
				int line = lineToRender;
				lineToRender++;

				nodeLine[i] = line;
				double dParams[] = { 
					x0, 
					x1, 
					y0 + yStep*line
				};
				
				int iParams[] = {
					line,
					numX
				};

				// cout << "Sending render command for line" << line << " to node " << i << endl;
				MPI_Send(iParams, 2, MPI_INT, i, 0, MPI_COMM_WORLD);
				MPI_Send(dParams, 3, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);	
			}
		}

		// Check for a received line

		MPI_Status status;
		MPI_Recv(&(buffer[0]), buffer.size(), MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

		int source = status.MPI_SOURCE;
		if (source < 1 || source >= nProcs)
		{
			errStr = "Internal error: received invalid source";
			return 0;
		}

		int y = nodeLine[source];
		if (y < 0 || y >= numY)
		{
			errStr = "Internal error: invalid y coordinate";
			return 0;
		}
		nodeLine[source] = -1;
		linesRendered++;

		//cout << "Read response for line " << y << ": " << buffer.size() << " double" << endl;

		// Copy the received result
		for (int x = 0 ; x < numX ; x++)
		{
			int idx = x+y*numX;

			pDensValues[idx] = buffer[x];
		}
		
		// Update progress

		int pct = (int)(((((double)linesRendered)/((double)(numY-1))))*100.0+0.5);
		pIOSys->setPercentageFeedback(pct);
	}

	// Signal helpers that we're done
	int doneParams[] = { -1, -1 };
	for (int i = 1 ; i < nodeLine.size() ; i++)
		MPI_Send(doneParams, 2, MPI_INT, i, 0, MPI_COMM_WORLD);

	pIOSys->writeOutputLine("Done");

	return new DiscreteGridFunction(pDensValues, Vector2D<double>(x0, y0), Vector2D<double>(x1, y1), numX, numY);
}

bool startLensPlotMPIHelper(string &errStr)
{
	// Get the lens for which the calculation is needed

	int lensSize = 0;
	MPI_Bcast(&lensSize, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if (lensSize <= 0)
	{
		errStr = "Invalid lens size";
		return false;
	}

	//cout << "Lens size is " << lensSize << endl;
	
	vector<uint8_t> lensBuffer(lensSize);
	MPI_Bcast(&(lensBuffer[0]), lensBuffer.size(), MPI_BYTE, 0, MPI_COMM_WORLD);
	
	serut::MemorySerializer mSer(&(lensBuffer[0]), lensBuffer.size(), 0, 0);
	GravitationalLens *pLens = 0;

	if (!GravitationalLens::read(mSer, &pLens, errStr))
		return false;

	Real2DLensFunction f(*pLens);
	vector<double> buffer;

	bool done = false;
	while (!done)
	{
		int iParams[2];
		MPI_Status status;

		// cout << "Reading command in node " << rank << endl;
		MPI_Recv(iParams, 2, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
		if (iParams[0] < 0 || iParams[1] <= 0)
			done = true;
		else
		{
			int line = iParams[0];
			int numx = iParams[1];

			double dParams[3];
			MPI_Recv(dParams, 3, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);

			double x0 = dParams[0];
			double x1 = dParams[1];
			double y = dParams[2];

			double xstep = (x1-x0)/((double)(numx-1));
			buffer.resize(numx);

			for (int i = 0 ; i < numx ; i++)
			{
				double x = ((double)i)*xstep + x0;
				Vector2D<double> theta(x, y);

				buffer[i] = f(theta);
			}

			MPI_Send(&(buffer[0]), buffer.size(), MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
		}
	}
	
	delete pLens;
	return true;
}

CmdLensPlotGnuplotMPI::CmdLensPlotGnuplotMPI(const string &cmdname) : CmdLensPlotGnuplot(cmdname)
{
	setDescription("Create a gnuplot file of the mass density of the lens, using MPI to speed up calculations.");
}

CmdLensPlotGnuplotMPI::~CmdLensPlotGnuplotMPI()
{
}

Real2DFunction *CmdLensPlotGnuplotMPI::getPlotFunction(grale::GravitationalLens &lens) const
{
	string errStr;
	Real2DFunction *pFunction = startLensPlotMPIMaster(lens, GetNumX(), GetNumY(), GetX1(), GetX2(), GetY1(), GetY2(), errStr);
	if (!pFunction)
		setErrorString(errStr);
	return pFunction;
}

CmdLensPlotFITSMPI::CmdLensPlotFITSMPI(const string &cmdname) : CmdLensPlotFITS(cmdname)
{
	setDescription("Create a FITS plot of the mass density of the lens, using MPI to speed up calculations.");
}

CmdLensPlotFITSMPI::~CmdLensPlotFITSMPI()
{
}

Real2DFunction *CmdLensPlotFITSMPI::getPlotFunction(grale::GravitationalLens &lens) const
{
	string errStr;
	Real2DFunction *pFunction = startLensPlotMPIMaster(lens, GetNumX(), GetNumY(), GetX1(), GetX2(), GetY1(), GetY2(), errStr);
	if (!pFunction)
		setErrorString(errStr);
	return pFunction;
}

CmdLensMassMPI::CmdLensMassMPI(const string &cmdName) : CmdLensMass(cmdName)
{
	string desc = getDescription();
	setDescription(desc + "This implementation uses MPI to speed up the calculation of the densities at the specified points.");
}

CmdLensMassMPI::~CmdLensMassMPI()
{
}

Real2DFunction *CmdLensMassMPI::getPlotFunction(grale::GravitationalLens &lens) const
{
	string errStr;
	Real2DFunction *pFunction = startLensPlotMPIMaster(lens, getNumX(), getNumY(), getX1(), getX2(), getY1(), getY2(), errStr);
	if (!pFunction)
		setErrorString(errStr);
	return pFunction;
}
