/*

  This file is a part of GRALESHELL, a program to facilitate the simulation
  and inversion of gravitational lenses, based on the GRALE library.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include <mpi.h>
#include "graleshellinstance.h"
#include "mainmpi.h"
#include "cmdinvertmpi.h"
#include "cmdlensplanempi.h"
#include "cmdimageplanempi.h"
#include "cmdlensplanempi.h"
#include "cmdlensesmpi.h"
#include <shellp/shellcommand.h>
#include <serut/memoryserializer.h>
#include <mogal/mpigeneticalgorithm.h>
#include <mogal/gamodule.h>
#include <mogal/gafactory.h>
#include <grale/lensplane.h>
#include <grale/imageplane.h>
#include <grale/openclkernel.h>
#include <iostream>

#include "debugnew.h"

std::list<shellp::ShellCommand *> registeredMPICommands;

void RegisterMPICommands()
{
	registeredMPICommands.push_back(new CmdInvertGridMPI("invert/grid/mpi"));
	registeredMPICommands.push_back(new CmdLensPlaneMPI("lensplane/new/mpi"));
	registeredMPICommands.push_back(new CmdImagePlaneMPI("imgplane/new/mpi"));
	registeredMPICommands.push_back(new CmdLensPlotGnuplotMPI("lens/plot/gnuplot/mpi"));
	registeredMPICommands.push_back(new CmdLensPlotFITSMPI("lens/plot/fits/mpi"));
	registeredMPICommands.push_back(new CmdLensMassMPI("lens/mass/mpi"));
}

void UnregisterMPICommands()
{
	while (!registeredMPICommands.empty())
	{
		delete *(registeredMPICommands.begin());
		registeredMPICommands.pop_front();
	}
}

void RegisterCommands();
void UnregisterCommands();

int main(int argc, char **argv)
{
#ifdef WIN32
	WSADATA dat;
	WSAStartup(MAKEWORD(2,2),&dat);
#endif // WIN32

	int worldSize, myRank;

	MPI_Init(&argc, &argv);
	MPI_Comm_size( MPI_COMM_WORLD, &worldSize);

	if (worldSize <= 1)
	{
		std::cerr << "Need more than one process to be able to work" << std::endl;
		MPI_Abort(MPI_COMM_WORLD, -1);
	}

	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);

	int returnStatus = 0;

	if (myRank == 0)
	{
		std::cout << "Starting MPI session with " << worldSize << " processes" << std::endl;

		GraleShellInstance instance;

#ifdef GRALECONFIG_SUPPORT_OPENCL // TODO: add a command to do this
		grale::OpenCLKernel *pCLKernel = new grale::OpenCLKernel();

		if (!pCLKernel->init())
			delete pCLKernel;
		else
			instance.SetOpenCLKernel(pCLKernel);
#endif // GRALECONFIG_SUPPORT_OPENCL

		RegisterCommands();
		RegisterMPICommands();

		int returnStatus = instance.run("GraleShell", "graleshell> ", argc, argv);

		UnregisterMPICommands();
		UnregisterCommands();

		int command = GRALESHELL_MPICOMMAND_EXIT;
		MPI_Bcast(&command, 1, MPI_INT, 0, MPI_COMM_WORLD);
	}
	else
	{
		bool done = false;

		while (!done)
		{
			int command = 0;
			MPI_Bcast(&command, 1, MPI_INT, 0, MPI_COMM_WORLD);

			if (command == GRALESHELL_MPICOMMAND_EXIT)
				done = true;
			else
			{
				if (command == GRALESHELL_MPICOMMAND_GA)
				{
					int paramsSize = 0;
					MPI_Bcast(&paramsSize, 1, MPI_INT, 0, MPI_COMM_WORLD);

					std::vector<uint8_t> paramBuffer(paramsSize);
					MPI_Bcast(&(paramBuffer[0]), paramsSize, MPI_BYTE, 0, MPI_COMM_WORLD);
					serut::MemorySerializer mSer(&(paramBuffer[0]), paramsSize, 0, 0);

					std::string baseDir, moduleName;
					
					if (!mSer.readString(baseDir))
					{
						std::cerr << "Internal error: can't read module base directory" << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}
					if (!mSer.readString(moduleName))
					{
						std::cerr << "Internal error: can't read module name" << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}
					
					mogal::GAModule module;

					if (!module.open(baseDir, moduleName))
					{
						std::cerr << "Internal error: can't open specified GA module: " << baseDir << "/" << moduleName << ": " << module.getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					mogal::GAFactory *pFactory = module.createFactoryInstance();

					if (pFactory == 0)
					{
						std::cerr << "Internal error: unable to create factory from module: " << module.getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					mogal::GAFactoryParams *pParams = pFactory->createParamsInstance();
					if (pParams == 0)
					{
						std::cerr << "Internal error: unable to create factory parameters instance: " << pFactory->getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					if (!pParams->read(mSer))
					{
						std::cerr << "Internal error: unable to read factory parameters: " << pParams->getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					if (!pFactory->init(pParams))
					{
						std::cerr << "Internal error: unable to init factory" << pFactory->getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					mogal::MPIGeneticAlgorithm gaHelper;

					if (!gaHelper.runMPIHelper(*pFactory))
					{
						std::cerr << "Internal error: error running MPI GA helper: " << gaHelper.getErrorString() << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}

					delete pParams;
					delete pFactory;
				}
				else if (command == GRALESHELL_MPICOMMAND_LENSPLANE)
				{
					std::string errStr;

					if (!grale::LensPlane::runMPIHelper(errStr))
					{
						std::cerr << "Internal error: error running MPI LensPlane helper: " << errStr << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}
				}
				else if (command == GRALESHELL_MPICOMMAND_IMAGEPLANE)
				{
					std::string errStr;

					if (!grale::ImagePlane::runMPIHelper(errStr))
					{
						std::cerr << "Internal error: error running MPI ImagePlane helper: " << errStr << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}
				}
				else if (command == GRALESHELL_MPICOMMAND_DENSITY)
				{
					std::string errStr;

					if (!startLensPlotMPIHelper(errStr))
					{
						std::cerr << "Internal error: error running MPI density plot helper: " << errStr << std::endl;
						MPI_Abort(MPI_COMM_WORLD, -1);
					}
				}
				else
				{
					std::cerr << "Received illegal MPI helper command number " << command << std::endl;
					MPI_Abort(MPI_COMM_WORLD, -1);
				}
			}
		}
	}
	
	MPI_Finalize();

#ifdef WIN32
	WSACleanup();
#endif
	return returnStatus;
}


